Confluence.Blueprint.setWizard('com.playsql.playsql-base-plugin:playsql-blueprint-webitem', function(wizard) {

	var defaultCase = null,
		mode = null,
		darkFeatures = [],
		listOfSpreadsheets = null;

	function updateVisibleTypes(mode) {
		if (mode == "READ_ONLY") {
			$(".bp-types .require-read-write").addClass("disabled");
			selectType($(".bp-types .selected"));
			var $addenum = $(".dialog-wizard-page-description p.addenum");
			if ($addenum.size() == 0) {
				$("<p>", {
					text: AJS.I18n.getText("com.playsql.playsql-base-plugin.create.about.desc.addenum"),
					"class": "addenum"
				}).appendTo(".dialog-wizard-page-description");
			} else {
				$addenum.removeClass("hidden");
			}
		} else {
			$(".bp-types .require-read-write").removeClass("disabled");
			$(".dialog-wizard-page-description p.addenum").removeClass("hidden");
		}
	}

	function selectType($template) {
		// If disabled, select the next possible
		var i = 0;
		while ($template.is(".disabled")) {
			if (i++ > 4) {
				$template.parent().children().removeClass("selected");
				break;
			}
			$template = $template.next(":visible:not(.disabled)").length != 0
				? $template.next(":visible:not(.disabled)")
				: $template.parent().children(":visible:not(.disabled)").first();
		}

		// Go ahead with the selection
		$template.addClass("selected").siblings().removeClass("selected");
		if ($template.hasClass("with-next")) {
        	$('.create-dialog-create-button:visible').text("Next");
		} else {
			$('.create-dialog-create-button:visible').text("Create");
		}
	}

	wizard.on('post-render.choose', function(e, state) {
		Confluence.PlaySQL.fetchContextProperties(false, state.wizardData.spaceKey, function(jqXHR, defaultCase1, mode1) {
			defaultCase = defaultCase1;
			mode = mode1;
			darkFeatures = jqXHR.getResponseHeader("X-DARK-FEATURES");
			darkFeatures = darkFeatures && darkFeatures.split(",") || [];
			var $types = $(".bp-types");
			if ($types.size() > 0) {
				$types.fadeIn(200);
				$types.attr('tabindex', 0).focus();
				updateVisibleTypes(mode);
				selectType($("#bp-create-spreadsheet"));
			}
			// Also fetch the list of tables, if the user creates a query
			Confluence.PlaySQL.fetchContextProperties(true, state.wizardData.spaceKey, function(jqXHR, defaultCase1, mode1) {
				var list;
				try {
					// List of [key, name, type] for every table
					list = $.parseJSON(jqXHR.responseText);
					if (!list && !list.length) return;
					listOfSpreadsheets = [{
						name: "",
						label: AJS.I18n.getText("com.playsql.playsql-base-plugin.create.query.select.none")
					}];
					for (var i in list) {
						if (list[i][2] == "SPREADSHEET")
							listOfSpreadsheets.push({name: list[i][0], label: list[i][1]});
					}
				} catch (err) {
				}

			});
		});

		$(".bp-types").on("click", ".template", function(){
			selectType($(this));
		}).on("dblclick", ".template", function(){
			$('.create-dialog-create-button:visible').click();
		}).keydown(function(e) {
			if (e.keyCode == 38 || e.keyCode == 40) {
				e.preventDefault();

				var $selected = $('.bp-types .template.selected'),
					$templates = $('.bp-types .template'),
					$newSelected;
				if (e.keyCode == 38) // up
					$newSelected = $selected.prev(":visible:not(.disabled)").length != 0
						? $selected.prev(":visible:not(.disabled)")
						: $selected.parent().children(":visible:not(.disabled)").last();
				else if (e.keyCode == 40) // down
					$newSelected = $selected.next(":visible:not(.disabled)").length != 0
						? $selected.next(":visible:not(.disabled)")
						: $selected.parent().children(":visible:not(.disabled)").first();
				selectType($newSelected);
			}
		}).hide().removeClass("hidden");
		//updateVisibleTypes(mode);
		//selectType($("#bp-create-spreadsheet"));
	});

	wizard.on('submit.choose', function(e, state) {
		var $selected = $(".bp-types .template.selected"),
			type = $selected.attr("id");
		if (type == "bp-create-spreadsheet") {
			state.nextPageId = "create-spreadsheet";
		} else if (type == "bp-create-query") {
			if (listOfSpreadsheets && listOfSpreadsheets.length > 0) {
				state.nextPageId = "create-query";
			} else {
				state.nextPageId = "create-redirect";
				state.wizardData.redirectUrl = AJS.contextPath() + "/playsql-base-plugin/query-editor.action?"
				+ $.param({
					key: state.wizardData.spaceKey,
					create: "true"
				});
			}
		} else if (type == "bp-create-joint-table") {
			state.nextPageId = "create-redirect";
			state.wizardData.redirectUrl = AJS.contextPath() + "/playsql-base-plugin/joint-table-editor.action?"
				+ $.param({
					key: state.wizardData.spaceKey,
					create: "true"
				});
		}
	});

	wizard.on('post-render.create-redirect', function(e, state) {
		// Submit immediately
		window.setTimeout(function(){
			$('.create-dialog-create-button:visible').click();
		}, 0);
	});

	wizard.on('submit.create-redirect', function(e, state) {
		if (!state.wizardData.redirectUrl) throw "A redirection url is expected when using the create-redirect wizard step";
		state.finalUrl = state.wizardData.redirectUrl;
	});

	wizard.on('post-render.create-spreadsheet', function(e, state) {
        state.$container.find("#spreadsheet-name, #spreadsheet-key").keyup(function(){
            var val = state.$container.find("#spreadsheet-name").val(),
                $key = state.$container.find("#spreadsheet-key"),
                keyVal = $key.val();
            if (keyVal === $key.data('generated-val') || keyVal == "") {
                // Generate the key from the name
                var generatedKey = Confluence.PlaySQL.Helpers.generateKey(val, defaultCase);
                keyVal != generatedKey && $key.val(generatedKey).data('generated-val', generatedKey);
            } else {
                // Autocorrect the key
                var correctKey = Confluence.PlaySQL.Helpers.generateKey($key.val(), defaultCase);
                if (correctKey != $key.val())
                    $key.val(correctKey);
            }
        });
	});

	wizard.on('submit.create-spreadsheet', function(e, state) {
		var $nameField = state.$container.find("#spreadsheet-name"),
			$keyField = state.$container.find("#spreadsheet-key"),
			$errorField = $keyField.parent().find(".error");
		var name = $nameField.val(),
			key = $keyField.val(),
			spaceKey = state.wizardData.spaceKey || AJS.Meta.get("space-key"),
			spreadsheet = {
				name: key,
				label: name
			};
		$errorField.empty();
		if (!name || !key || !spaceKey) {
			$errorField.text("All fields are required");
			return false;
		}
		// Submit the spreadsheet
		Confluence.PlaySQL.createSpreadsheet(spaceKey, spreadsheet, undefined, Confluence.PlaySQL.defaultCase, function(){
			//window.location.href=AJS.contextPath() + "/playsql-base-plugin/spreadsheet-editor.action?key=" + spaceKey + "&spreadsheetId=" + key;
			state.validationDeferred.resolve();
		}, function(data){
			var message = data.responseText;
			if (data && data.status == 409) {
				// Conflict, table already exists
			}
			if (message) {
				try {
					var data2 = $.parseJSON(data.responseText);
					if (data2.message) message = data2.message;
				} catch (err) {}
			}
			$errorField.text(message);
			state.validationDeferred.reject();
		});
		state.finalUrl = AJS.contextPath() + "/playsql-base-plugin/spreadsheet-editor.action?" +
						$.param({key: spaceKey, spreadsheetId: key});
		state.async = true;
	});

	wizard.on('pre-render.create-query', function(e, state) {
		state.soyRenderContext['spreadsheets'] = listOfSpreadsheets;
	});

	wizard.on('submit.create-query', function(e, state) {
		state.finalUrl = AJS.contextPath() + "/playsql-base-plugin/query-editor.action?"
				+ $.param({
					key: state.wizardData.spaceKey,
					create: "true",
					from: state.$container.find("#table-name").val() || "your_table"
				});
	});
});
