AJS.$(function($){
    Confluence.PlaySQL = Confluence.PlaySQL || {};
    var times = 0;
    function confirmDelete(name, type) {
        if (times == 0) {
            if (confirm("This will delete the " + type + " permanently. Pages which refer to this element " +
                "will display an empty table. Are you sure you want to delete the " + type + " \"" + name + "\"?")) {
                times++;
                return true;
            } else return false;
        } else if (times == 1) {
            if (confirm("Are you sure you want to delete this " + type + "? Note: Next time you try to " +
                "delete a query, no confirmation will be requested.")) {
                times++;
                return true;
            } else return false;
        } else return true;
    }

    $(".queryview").on("click", ".delete-query", function(){
        var $this = $(this),
            queryId = $this.data('query-id'),
            name = $this.closest("tr").find("td a").first().text();
            spaceKey = AJS.Meta.get("space-key"),
            url = AJS.contextPath() + "/rest/playsql/1.0/query/space:" + spaceKey + "/" + queryId;

        if (!confirmDelete(name, "query")) {
            return false;
        }
        $.ajax({
            url: url,
            type: 'DELETE',
            contentType: "application/json; charset=utf-8"
        }).done(function(data){
            $this.closest("tr").remove();
        }).fail(function(jqXHR, error, reason){
                alert("The query couldn't be deleted. You might have no permission to do this.");
            AJS.logError("Error while attempting to delete a query/spreadsheet. " + error + " - " + reason);
        });
        return false;
    });
    $(".jointtableview").on("click", ".delete-joint-table", function(){
        var $this = $(this),
            id = $this.data('joint-table-id'),
            name = $this.closest("tr").find("td a").first().text(),
            spaceKey = AJS.Meta.get("space-key"),
            url = AJS.contextPath() + "/rest/playsql-spreadsheets/1.0/space:" + spaceKey + "/" + id;

        if (!confirmDelete(name, "joint table")) {
            return false;
        }
        $.ajax({
            url: url,
            type: 'DELETE',
            contentType: "application/json; charset=utf-8"
        }).done(function(data){
            $this.closest("tr").remove();
        }).fail(function(jqXHR, error, reason){
            alert("The joint table couldn't be deleted. You might have no permission to do this.");
            AJS.logError("Error while attempting to delete a query/spreadsheet. " + error + " - " + reason);
        });
        return false;
    });
    $(".spreadsheetview").on("click", ".delete-spreadsheet", function(){
        var $this = $(this),
            spreadsheetId = $this.data('spreadsheet-id'),
            name = $this.closest("tr").find("td a").first().text();
        var spaceKey = AJS.Meta.get("space-key"),
            url = AJS.contextPath() + "/rest/playsql-spreadsheets/1.0/space:" + spaceKey + "/" + spreadsheetId;

        if (!confirmDelete(name, "spreadsheet and underlying table")) {
            return false;
        }
        $.ajax({
            url: url,
            type: 'DELETE',
            contentType: "application/json; charset=utf-8"
        }).done(function(data){
            $this.closest("tr").remove();
        }).fail(function(jqXHR, error, reason){
            alert("The spreadsheet couldn't be deleted. You might have no permission to do this.");
            AJS.logError("Error while attempting to delete a query/spreadsheet. " + error + " - " + reason);
        });
        return false;
    }).on("click", ".duplicate-spreadsheet", function(e){
        var $this = $(this),
            spreadsheetId = $this.data('spreadsheet-id');
        //    name = $this.closest("tr").find("td a").first().text();
        //var spaceKey = AJS.Meta.get("space-key"),
        //    url = AJS.contextPath() + "/rest/playsql-spreadsheets/1.0/space:" + spaceKey + "/" + spreadsheetId;

        e.preventDefault();
        createSpreadsheetShowDialog(spreadsheetId);
        return false;
    });


    function createDialog(copyId) {
        var $template = $(Confluence.Templates.PlaySQL.CreateExperience.createSpreadsheetForm({ copyId : copyId })),

            submitCallback = function () {
                $("#create-spreadsheet-dialog .button-panel-submit-button").disable().text("Submitting...");
                var $nameField = $("#spreadsheet-name"),
                    $keyField = $("#spreadsheet-key"),
                    $errorField = $keyField.parent().find(".error");
                var name = $nameField.val(),
                    key = $keyField.val(),
                    spaceKey = AJS.Meta.get("space-key"),
                    spreadsheet = {
                        name: key,
                        label: name
                    };
                $errorField.empty();
                if (!name || !key || !spaceKey) {
                    $errorField.text("All fields are required");
                    $("#create-spreadsheet-dialog .button-panel-submit-button").enable().text("Submit");
                    return false;
                }
                // Submit the spreadsheet
                Confluence.PlaySQL.createSpreadsheet(spaceKey, spreadsheet, copyId, Confluence.PlaySQL.defaultCase, function(){
                    window.location.href=AJS.contextPath() + "/playsql-base-plugin/spreadsheet-editor.action?key=" + spaceKey + "&spreadsheetId=" + key;
                }, function(data){
                    $("#create-spreadsheet-dialog .button-panel-submit-button").enable().text("Submit");
                    if (data && data.status == 409) {
                        // Conflict, table already exists
                        $errorField.text(data.responseText);
                    } else {
                        $errorField.text(data.responseText);
                    }
                });
            },

            cancelCallback = function () {
                dialog.hide();

                $nameField.val('');
                $keyField.val('');
                keyErrors.text('').addClass('hidden');
                nameErrors.text('').addClass('hidden');
                enableSubmit(false);
            },

            dialog = new AJS.ConfluenceDialog({
                width: 600,
                height: 320,
                id: 'create-spreadsheet-dialog',
                closeOnOutsideClick: true,
                onCancel: cancelCallback,
                onSubmit: submitCallback
            });

        dialog.addHeader("Create spreadsheet");
        dialog.addPanel("Panel 1", $template);

        dialog.addSubmit("Submit", submitCallback);
        dialog.addCancel("Cancel", cancelCallback);

        $submitButton = dialog.popup.element.find('.button-panel-submit-button');

        dialog.gotoPage(0);
        dialog.gotoPanel(0);
        return dialog;
    }

    /**
     * Shows the Create dialog for the spreadsheet
     * @param copyId if not undefined, the new spreadsheet will be a copy of the 'copyId'
     */
    function createSpreadsheetShowDialog(copyId) {
        createDialog(copyId).show();
        $("#spreadsheet-name, #spreadsheet-key").keyup(function(){
            var val = $(this).val(),
                $key = $("#spreadsheet-key"),
                keyVal = $key.val();
            if (keyVal === $key.data('generated-val') || keyVal == "") {
                // Generate the key from the name
                var generatedKey = Confluence.PlaySQL.Helpers.generateKey(val, Confluence.PlaySQL.defaultCase);
                keyVal != generatedKey && $key.val(generatedKey).data('generated-val', generatedKey);
            } else {
                // Autocorrect the key
                var correctKey = Confluence.PlaySQL.Helpers.generateKey($key.val(), Confluence.PlaySQL.defaultCase);
                if (correctKey != $key.val())
                    $key.val(correctKey);
            }
        });
    }

    $("#create-spreadsheet-button").click(function(e){
        e.preventDefault();
        alert("We've moved! Spreadsheets and Queries can be created using the Create wizard in the top bar.");
        //createSpreadsheetShowDialog();
        return false;
    });

    if (window.location.hash == "#create-spreadsheet") {
        createSpreadsheetShowDialog();
    }
    Confluence.PlaySQL.fetchContextProperties(false, undefined, function(data, defaultCase) {
        if (defaultCase != "LOWER" && defaultCase != "UPPER") {
            $("#create-query-button, #create-spreadsheet-button").hide();
            throw "Invalid default case: " + defaultCase;
        }
        Confluence.PlaySQL.defaultCase = defaultCase;
    });

    // Lalilarialadadaaah mellow web-item and space.vmd
    if ($(".space-query-index, .space-table-index").closest(".theme-documentation").size() != 0) {
        $("#tab-navigation .menu-item:contains('Tables')").addClass("active-tab")
    }
});