grammar Formulas;

value
    : cellReference                                                   # ValueCell
    | rangeReference                                                  # ValueRange
    | functionCall                                                    # ValueFunction
    | value ('-'|'+'|'OR'|'XOR') value                                # ValueAdd
    | value ('/'|'*'|'AND'|'='|'<>'|'!='|'<'|'>'|'<='|'>=') value     # ValueMul
    | 'NOT' value                                                     # ValueNot
    | LPAREN value RPAREN                                             # ValueParen
    | literalValue                                                    # ValueLiteral
    ;

functionCall
    : functionName=Identifier
      LPAREN value (',' value)* RPAREN
    ;

cellReference :
      columnName=columnIdentifier                   # CellSameRow
    | columnName=columnIdentifier ':' rowId         # CellRowID
    ;

rangeReference
    : columnName=columnIdentifier ':*'              # RangeAll
    | columnName=columnIdentifier ':[' condition ']'# RangeCondition
    ;

condition :
      columnIdentifier      # ConditionColumn
    | value                 # ConditionValue     // This is ambiguous, but the most used. Generally it's an RVBilateral '='. Best effort not guaranteed here.
    ;

literalValue
    : number            # LiteralValueNumber
    | ('TRUE'|'FALSE')  # LiteralConstant
    | StringLiteral     # LiteralValueString
    ;

rowId
    : Digits            # RowIdDigits ;

/* *** TOKENS *** */

StringLiteral             : '\'' StringCharacters? '\''
                            /*{   // This instruction makes that token.getText() returns the decoded string, not
                                // the escaped blurb
                                setText(
                                    com.playsql.core.formulas.parser.CharSupport.getStringFromGrammarStringLiteral(
                                        getText()
                                    )
                                );
                            }*/;
fragment StringCharacters : StringCharacter+;
fragment StringCharacter  : ~['\\] | EscapeSequence ;
fragment EscapeSequence   : '\\' . ;
/*		(	[nrft'"\\] // CR, LF, field, tab, quotes, antislash
		|	'u' (HEX_DIGIT (HEX_DIGIT (HEX_DIGIT HEX_DIGIT?)?)?)? // Unicode
		|	. // Useless antislash
		)
	;

fragment
UNICODE_ESC
    :   'u' (HEX_DIGIT (HEX_DIGIT (HEX_DIGIT HEX_DIGIT?)?)?)?
    ;
*/
fragment
HEX_DIGIT : [0-9a-fA-F]	;

columnIdentifier :
      columnName=Identifier
    | tableName=Identifier '.' columnName=Identifier
    | tableName=Identifier '|' tableAlias=Identifier '.' columnName=Identifier
    ;

Identifier : Letter LetterOrDigit* ;
fragment Letter : [a-zA-Z_];
fragment LetterOrDigit : [a-zA-Z0-9_];

// 'number' can't be a token. Becauuuse if you tokenize '-2',
// '2-2' isn't a valid operation.
number  : '-'? Digits ('.' Digits)? ;
Digits  : [0-9]+;

COMMA               : ',';
LPAREN              : '(';
RPAREN              : ')';

WS  :  [ \t\n\f\r\u00A0]+ -> channel(HIDDEN)
    ;

//COMMENT
//    :   '/*' .*? '*/' -> channel(COMMENT)
//    ;

//LINE_COMMENT
//    :   '//' ~[\r\n]* -> channel(COMMENT)
//    ;
