/**
 * This file loads the SQL browser, for any existing macro.
 * "SQL Browser" = textarea just before the div.playsql-placeholder
 * in which the user may enter SQL to update the placeholder and re-query the
 * results.
 */

var PlaySQLBrowser = PlaySQLBrowser || {}; 

(function($){

  var fieldAccessor = function(fieldNameInMacro, selector, overrides, isJSON, isSaved) {
    function getWiget(selector) {
      return PlaySQLBrowser.initialized.$queryForm.find(selector);
    };
    return $.extend({
      /** @param value string format of a value */
      setField : function(value) {
        if (getWiget(selector).size() == 0) {
          PlaySQLBrowser.initialized.showField2(fieldNameInMacro, value);
        }
        if (getWiget(selector).size() > 0) {
          getWiget(selector).val(value);
        }
      },
      /** @return value string format of a value */
      getField : function() {
        if (getWiget(selector))
          return getWiget(selector).val();
      },
      /** 
       * Sets the data() in string format
       * @param value string format of the value */
      set : function(value) {
        if (isJSON && typeof value == 'string') {
          value = $.parseJSON(value);
        } 
        PlaySQLBrowser.initialized.$placeholder.data(fieldNameInMacro, value);
      },
      /** 
       * Gets the value from the data parameter (data is in string format).
       * @return value the object format of the value */
      getObject : function() {
        var value = PlaySQLBrowser.initialized.$placeholder.data(fieldNameInMacro);
        if (isJSON && typeof value == "string") {
          value = $.parseJSON(value);
        }
        return value;
      },
      /** 
       * @return value the string format of the value */
      getString : function() {
        var value = PlaySQLBrowser.initialized.$placeholder.data(fieldNameInMacro);
        if (isJSON && typeof value == "object") {
          value = JSON.stringify(value);
        }
        return value;
      },
      isJSON : function() {
        return isJSON;
      },
      isSaved : function() {
        return isSaved;
      }
    }, overrides);
  };

  PlaySQLBrowser.fields = {
    'sql' : fieldAccessor('macro-sql', '.sql', undefined, undefined, true),
    'timeout': fieldAccessor('macro-timeout', '#timeout', undefined, undefined, true),
    'context': fieldAccessor('macro-context', "#context-selector", undefined, undefined, true),
    'readonly' : fieldAccessor('readonly', undefined, undefined, undefined, false),
    'open-browser': fieldAccessor('macro-open-browser', undefined, undefined, undefined, false),
    'groupby': fieldAccessor('macro-groupby', undefined, {}, true, true),
    'built-in-parameters': fieldAccessor('macro-built-in-parameters', "#built-in-parameters", {
        getField : function() {
            if ($("#built-in-parameters:visible").size() == 1) {
                return true;
            } else {
                return false;
            }
        },
        setField : function(value) {
            if (value === 'true' || value === true) {
                PlaySQLBrowser.initialized.showField2('macro-built-in-parameters', value);
            }
        }

    }, false, true),
    'user-parameters': fieldAccessor('macro-user-parameters', undefined, {
      setField : function(values) {
        var values = $.parseJSON(values);
        $('*[name="user-parameter"]').closest('.field-group').remove();
        for (var i in values) {
          PlaySQLBrowser.initialized.showField('user-parameter', values[i]);
        }
      },
      getField : function() {
        var values = [];
        $('*[name="user-parameter"]').each(function(){
          var $this = $(this),
              $fieldGroup = $this.closest('.field-group'),
              name = $this.val(),
              label = $fieldGroup.find('*[name="user-parameter-label"]').val(),
              defaultValue = $fieldGroup.find('*[name="user-parameter-default"]').val();
          if (name) {
            values.push({
              name: name,
              label: label,
              defaultValue: defaultValue
            });
          }
        });
        return values.length == 0 ? null : JSON.stringify(values);
      }
    }, true, true)
  };
  
  
/**
 * Initializes the browser features.
 * @param $queryForm the form#query-form in which all elements of the browser can be found. 
 */ 
PlaySQLBrowser.initialize = function($queryForm){
  if (PlaySQLBrowser.initialized) {
    throw 'The browser for Play SQL has already been initialized';
  }
  var $ = AJS.$,
      $sqlInput = $queryForm.find("textarea.sql"),
      $submitBtn = $queryForm.find("input.submit-sql"),
      $placeholder = $queryForm.next(".playsql-placeholder"),
      $autocomplete = $('#autocomplete-sql'),
      autocompleteEntries = $('<ul/>'),
      autocompleteActivated = false;
  PlaySQLBrowser.initialized = {
      $sqlInput : $sqlInput,
      $placeholder : $placeholder,
      $queryForm : $queryForm
  }

  $submitBtn.click(function(e) {
    e.preventDefault();
    PlaySQLBrowser.synchronizeValues();
    PlaySQLBrowser.query();
  });
  
  // According to the datasource in the combo box, update the macro placeholder
  var datasources,
      datasourceAvailableCallback,
      whenDatasourcesAvailable = function(callBack) {
        if (datasources) {
          callBack(datasources);
        } else {
          datasourceAvailableCallback = callBack;
        }
      };
      
  $.ajax(AJS.contextPath() + '/rest/playsql/1.0/playsql/contexts').done(function(data){
    if (data) {
      datasources = data;
      if (datasourceAvailableCallback) {
        datasourceAvailableCallback(data);
      }
    }
  });

  $sqlInput.keydown(function(e) {
    if (e.ctrlKey && (e.keyCode == 13 || e.keyCode == 10)) {
      $submitBtn.click();
    } else if (e.ctrlKey && (e.keyCode == 32)) {
      autocompleteActivated = true;
      window.setTimeout(autocomplete, 0);
    } else {
      window.setTimeout(function(){
        updateAutocomplete(getUserLastWord());
      }, 0);
    }
  }).blur(function(e) {
    $autocomplete.hide();
    autocompleteActivated = false;
  });

//  $queryForm.find(".hide-browser").click(function(e){
//    e.preventDefault();
//    $queryForm.remove();
//  });

  function autocomplete() {
    var lastWord = getUserLastWord();
    var beforeDot = lastWord.split(/\./)[0],
        afterDot = (lastWord.indexOf('.') && lastWord.split(/\./)[1]) || "",
        url = AJS.contextPath() + '/rest/playsql/1.0/playsql/tables?spaceKey=%s&schema=%s&startingWith=%s',
        spaceKey = $sqlInput.closest('.playsql-container').find('.playsql-placeholder').data('space-key');
    if (!spaceKey) return;
    if (!afterDot) {
      afterDot = beforeDot;
      beforeDot = "";
    }
    url = url.replace('%s', spaceKey);
    url = url.replace('%s', beforeDot);
    url = url.replace('%s', afterDot);
    $.ajax(url).done(function(data){
      var tables = data && data.tables;
      autocompleteEntries = [];
      for (var idx in tables) {
        autocompleteEntries.push({text: tables[idx].toUpperCase()});
      }
      if (autocompleteEntries == 0) {
        autocompleteEntries.push({text: 'No matches'});
        updateAutocomplete('');
      } else {
        updateAutocomplete(lastWord);
      }
      $autocomplete.removeClass('hidden').show();
    });
  }

  function getUserLastWord() {
    var selStart = $sqlInput[0].selectionStart,
        snippet, lastWord;
    
    if (!selStart || !$sqlInput.val()) {return;}
    snippet = $sqlInput.val().substring(0, selStart);
    if (!snippet) { return; }
    lastWord = snippet.substring(snippet.lastIndexOf(' ') + 1);
    return lastWord;
  }

  /* After an autocomplete, As the user types, reduces the matched items */ 
  function updateAutocomplete(startingWith) {
    if (autocompleteEntries.length==0 || !autocompleteActivated) return;
    startingWith = startingWith.toUpperCase();
    startingWith = startingWith.substring(startingWith.lastIndexOf('.')+1);
    
    var // Copy the values
        $newList = $('<ul/>'),
        $startingWith = $('<b/>').text(startingWith);
    
    for (var idx in autocompleteEntries) {
      var entry = autocompleteEntries[idx];
      if (entry.text) {
        if (idx == 15 && autocompleteEntries.length > 16) {
          $('<li/>').text('...').appendTo($newList);
          break;
        }
        if (entry.text.toLowerCase().indexOf(startingWith.toLowerCase()) == 0) {
          $('<li/>').text(entry.text.substring(startingWith.length)).prepend($startingWith.clone()).appendTo($newList);
        }
      }
    }
    $autocomplete.html($newList.children());
  }

  var dialog2 = AJS.InlineDialog(AJS.$('.display-more'), 188, function($defaultContents, trigger, showPopup) {
    $defaultContents.html("Some info: <ul><li>Press Ctrl + Space for autocompletion.</li>" +
    		                       "<li>Press Ctrl + Enter to submit the query.</li></ul>");
    showPopup();
  });
  
  
  var $addCombo = $("#addOption").click(function(){
    var $dropDown = $(this);
    $("option.unique", $dropDown).each(function(){
      var $option = $(this),
          name = $option.attr('value'),
          $fieldGroup = $("#" + name).parent();
      if ($fieldGroup.size() > 0 && $fieldGroup.is(":visible")) {
        $option.hide();
      } else {
        $option.show();
      }
    });
    
  }).change(function(){
    var name = $addCombo.val();
    if (name == 'initial') {
      return;
    }
    $(this).val("initial");
    if (!PlaySQLBrowser.initialized.showField(name)) {
      $("#addOptionError").removeClass('hidden');
      return;
    }
  });
  

  PlaySQLBrowser.initialized.showField2 = function(fieldNameInMacro, value) {
    // The first argument should be the name of the <option> tag
    // Translations between parameter name, name of the <option> and name of the #element
    var translations = {
        'macro-timeout' : 'timeout',
        'macro-context' : 'context-selector',
        'macro-built-in-parameters' : 'built-in-parameters',
        'macro-user-parameters' : 'user-parameter'
    },
        name = translations[fieldNameInMacro] || fieldNameInMacro;
    PlaySQLBrowser.initialized.showField(name, value);
  }
  
  /**
   * Shows a parameter of the form
   * @param name
   * @param value
   * @returns {Boolean} true if successful
   */
  PlaySQLBrowser.initialized.showField = function(name, value) {
    if (name == 'macro-context') name ='context-selector';
    var $option = $("option[value='" + name + "']", $addCombo),
        $fieldGroup = $("#" + name).parent();
    if ($option.size() == 0) {
      return false;
    }
    
    if (name == 'user-parameter' && !value) {
      value = {
          name : "$param1",
          label : "Enter the value",
          defaultValue : "My default value"
      }
    } else if (name == 'built-in-user-parameter' && value === undefined) {
        value = true;
    }
    
    // Add a new group
    var $newGroup = $(Confluence.Templates.PlaySQL.addParameter({name : name, value: value})).removeClass('hidden');
    $addCombo.parent().before($newGroup);

    if (name == 'context-selector') {
      whenDatasourcesAvailable(function(datasources){
        var context = $placeholder.data('macro-context'),
            contextHalves = context ? context.split(':') : [],
            spaceKey = contextHalves.length == 2 && contextHalves[0] == 'space' ? contextHalves[1] : undefined,
            currentSpaceKey = AJS.Meta.get('space-key'),
            contextName = contextHalves.length == 2 && contextHalves[0] == 'name' ? contextHalves[1] : undefined, 
            html = Confluence.Templates.PlaySQL.datasourceOptions({
              datasources: datasources, 
              spaceKey: spaceKey, 
              currentSpaceKey : currentSpaceKey,
              contextName: contextName
            }),
            $contextSelector = $("#context-selector");
        $contextSelector.html(html);
        if (value) {
          $contextSelector.val(value);
        }
        $contextSelector.change(function() {
          var value = $(this).val(),
              contextName = '',
              contextSpaceKey = '';
          if (!value) {
            // Leave empty strings
          } else if (value.indexOf('name:') == 0) {
            contextName = value.substring(5);
          } else if (value.indexOf('space:') == 0) {
            contextSpaceKey = value.substring(6);
          }
          $(".playsql-placeholder")
            .data('space-key', contextSpaceKey)
            .data('context-name', contextName);
        });
      });
    }
    return true;
  }
  
  $queryForm.delegate('.playsql-parameter-remove', 'click', function(e){
    e.preventDefault();
    var $button = $(this),
        $fieldGroup=$button.parent();
    if ($button.hasClass('remove-mode-delete')) {
      $fieldGroup.remove();
    } else if ($button.hasClass('remove-mode-hide')) {
      $fieldGroup.addClass('hidden');
    }
  });
  
  // Load the values from the placeholder into the fields
  PlaySQLBrowser.synchronizeValuesFromPlaceholder(PlaySQLBrowser.initialized.$placeholder);
}

PlaySQLBrowser.query = function() {
  PlaySQLMacro.load(PlaySQLBrowser.initialized.$placeholder);
}


// Sets the values passed as argument into the fields.
// Put all the values of the fields into the data of $placeholder,
// and returns them.
/**
 * Sets the values passed as argument into the fields.
 * Put all the values of the fields in to the data of $placeholder.
 * @param values the string format of the values
 * @return the data to save in the macro
 */
PlaySQLBrowser.synchronizeValues = function(values) {
  if (!PlaySQLBrowser.initialized) throw "The PlaySQL browser isn't open yet";
  var toReturn = {};
  values = values || {};
  /*values['groupby'] = JSON.stringify({
      fields : ['TABLE_SCHEMA'],
      aggregations : [{
        title: 'Total 1',
        expression: 'MAX(TABLE_NAME)'
      }],
      having : "TABLE_SCHEMA IS NOT NULL"
  };
  values['orderby'] = JSON.stringify({
      fields : [{
        field : 'TABLE_SCHEMA',
        ordering : 'ASC',
        nulls : 'LAST'
      }]
  };*/
  for (var fieldName in PlaySQLBrowser.fields) {
    var field = PlaySQLBrowser.fields[fieldName],
        value = values[fieldName];
    if (typeof value != 'undefined') {
      field.setField(value);
    } else {
      value = field.getField();
    }
    // Update all 'data' parameters of the placeholder
    if (typeof value != 'undefined') {
      field.set(value);
      if (field.isSaved()) toReturn[fieldName] = value;
    } else {
      if (field.isSaved()) toReturn[fieldName] = field.getString();
    }
  }
  return toReturn;
}

// Loads all the values from the placeholder into the fields
PlaySQLBrowser.synchronizeValuesFromPlaceholder = function($placeholder) {
  for (var fieldName in PlaySQLBrowser.fields) {
    var field = PlaySQLBrowser.fields[fieldName],
        value = field.getString();
    if (value)
      field.setField(value);
  }
}

})(AJS.$);