AJS.$(function($){
    var playsqlDialects = AJS.$.parseJSON($("meta[name=playsql-dialects]").attr("content"));
    var Connection = Backbone.Model.extend({
        defaults: {
            type: '',
            driver: '',
            dialect: '',
            url : '',
            username: '',
            password: '',
            mode: '',
            initSQL: '',
            schema: '',
            closeSQL: '',
            testSQL: '',
            jndiName: '',
            contextSpaceKey: '',
            contextName: '',
            deleted: false
        },
        toObject : function() {
            return {
                contextSpaceKey: this.get('contextSpaceKey'),
                contextName: this.get('contextName'),
                type: this.get('type'),
                driver: this.get('driver'),
                dialect: this.get('dialect'),
                url : this.get('url'),
                username: this.get('username'),
                password: this.get('password'),
                mode: this.get('mode'),
                initSQL: this.get('initSQL'),
                schema: this.get('schema'),
                closeSQL: this.get('closeSQL'),
                testSQL: this.get('testSQL'),
                jndiName: this.get('jndiName')
            }
        },
        parse : function(response) {
            return {
                contextSpaceKey : response.context.spaceKey,
                contextName : response.context.name,
                type : response.type,
                driver : response.driver,
                dialect : response.dialect,
                url : response.url,
                username : response.username,
                password : response.password,
                mode : response.mode,
                initSQL : response.initSQL,
                schema : response.schema,
                closeSQL : response.closeSQL,
                testSQL : response.testSQL,
                jndiName : response.jndiName
            };
        },
        toJSON : function() {
            return {
                context : {
                    name : this.get('contextName'),
                    spaceKey : this.get('contextSpaceKey')
                },
                type : this.get('type'),
                driver : this.get('driver'),
                dialect : this.get('dialect'),
                url : this.get('url'),
                username : this.get('username'),
                password : this.get('password'),
                mode : this.get('mode'),
                initSQL : this.get('initSQL'),
                schema : this.get('schema'),
                closeSQL : this.get('closeSQL'),
                testSQL : this.get('testSQL'),
                jndiName : this.get('jndiName')
            };
        }
    });

    var List = Backbone.Collection.extend({
        model: Connection,
        url: AJS.contextPath() + '/rest/playsql/1.0/datasources',
        comparator : function(a, b) {
            // Filter Factory Settings
            if (a.get('contextName') == 'Factory Settings')
                return -1;
            if (b.get('contextName') == 'Factory Settings')
                return 1;
            // Filter global connection
            if (!a.get('contextName') && !a.get('contextSpaceKey'))
                return -1;
            if (!b.get('contextName') && !b.get('contextSpaceKey'))
                return 1;

            // Compare named connections first
            if (a.get('contextName') != b.get('contextName'))
                return (a.get('contextName') || '').localeCompare(b.get('contextName'));

            // Space connections last
            return (a.get('contextSpaceKey') || '').localeCompare(b.get('contextSpaceKey'));
        }
    });

    var ConnectionView = Backbone.View.extend({
        tagName: 'tr',
        events : {
            "click .edit" : "edit",
            "click .remove" : "remove",
            "change .select.type" : "updateType",
            "change .select.dialect" : "updateDialect",
            "click .test-jdbc" : "test",
            "keypress" : "keypress",
            "click .submit" : "submit",
            "click .remove" : "remove"
        },
        initialize: function(){
            _.bindAll(this, 'render', 'edit', 'updateType', 'test', 'submit', 'remove', 'keypress', 'updateDialect');
            //this.model.bind('reset', this.render, this);
        },
        render: function() {
            var html = Confluence.Templates.PlaySQL.connectionView({connection: this.model.toObject(), mode: 'view', dialects: playsqlDialects});
            $(this.el).html(html);
            return this;
        },
        edit : function(e) {
            e && e.preventDefault();
            $('td.actions', this.$el).empty();
            var html = Confluence.Templates.PlaySQL.connectionView({connection: this.model.toObject(), mode: 'edit', dialects: playsqlDialects});
            this.$el.html(html);
            this.$el.addClass('editing');
            return this;
        },
        /** In edit mode, use a new type among INHERIT, DIRECT, DISABLE and JNDI */
        updateType : function(e) {
            var type = $(e.currentTarget).val(),
                that = this,
                connectionDetails = this.model.toObject();
            if (!type) return;
            connectionDetails.dialect = null;
            this.$el.find('.connection-details').html(Confluence.Templates.PlaySQL.connectionDetails({
                connection: connectionDetails,
                mode: 'edit',
                type: type,
                dialects: playsqlDialects
            }));
            if (type == 'JNDI') {
                $.ajax({
                    url : AJS.contextPath() + "/rest/playsql/1.0/admin/listJndiJdbc",
                    cache: false,
                    success : function(data) {
                        var $message = $('.jndi.field-group .description', that.$el),
                            count = 0;
                        $message.empty();
                        for ( var classLoaderIndex in data) {
                            for ( var name in data[classLoaderIndex]) {
                                if (name.indexOf("java:comp") == 0) {
                                    if (count++ > 0) $message.append('<br/>');
                                    $message.append(document.createTextNode('Example: ' + name + " (" + data[classLoaderIndex][name] + ")"));
                                }
                            }
                        }
                        if (count == 0) {
                            $message.text('Example: java:comp/env/jdbc/userdatasource');
                        }
                    }
                });
            }
        },
        updateDialect : function(e) {
            var dialect = $(e.currentTarget).val(),
                that = this,
                $driver = $('.driver', that.$el),
                $url = $('.url', that.$el),
                $dialectHint = $('.dialect-group .description', that.$el).empty(),
                $readWriteCombo = $(".select.mode"),
                $rwOption = $readWriteCombo.find("option[value='READ_WRITE']"),
                dialectDef = null;
            for (var i in playsqlDialects) {
                if (playsqlDialects[i].key == dialect) {
                    dialectDef = playsqlDialects[i];
                }
            }
            if (dialectDef == null) {
                dialectDef = {
                    key: dialect,
                    name: dialect,
                    driverName: "com.example.Driver",
                    exampleUrl: "jdbc:vendor://localhost:1111/database",
                    minimumDriverString: "always replace",
                    minimumUrlString: "always replace",
                    titleText: "Dialect not found",
                    titleLink: "",
                    readWrite: false
                }
            }

            if (dialectDef.titleText) {
                $dialectHint.text(dialectDef.titleText);
            }
            if (dialectDef.titleLink) {
                $dialectHint
                    .append(" ")
                    .append($("<a/>", {
                        text: "(Link)",
                        href: dialectDef.titleLink,
                        target: "_blank"
                    }));
            }
            if (!$driver.val() || $driver.val().toLowerCase().indexOf(dialectDef.minimumDriverString) == -1) {
                $driver.val(dialectDef.driverName);
            }
            if (!$url.val() || $url.val().toLowerCase().indexOf(dialectDef.minimumUrlString) == -1) {
                $url.val(dialectDef.exampleUrl);
            }
            if (dialectDef.readWrite) {
                // Check it's visible
                if ($rwOption.length == 0) {
                    $readWriteCombo.append($readWriteCombo.attr("data-hidden-option")).val("READ_WRITE");
                }
            } else {
                // Check it's hidden
                if ($rwOption.length == 1) {
                    var html = $("<p>").html($rwOption.detach()).html();
                    $readWriteCombo.attr("data-hidden-option", html).val("READ_ONLY");
                }
            }
        },
        keypress : function(e) {
            if (e.keyCode == 13) {
                this.test(e);
            }
        },
        test : function(e) {
            var that = this;
            e.preventDefault();
            $(".jdbc.is-valid", that.$el).hide();
            $(".jdbc.is-invalid", that.$el).hide();
            var type = this.$el.find('.select.type').val(),
                mode = this.$el.find('.mode', that.$el).val(),
                initSQL = this.$el.find('.initsql', that.$el).val(),
                schema = this.$el.find('.schema', that.$el).val(),
                closeSQL = this.$el.find('.closeSQL', that.$el).val(),
                testSQL = this.$el.find('.testsql', that.$el).val(),
                jndi = this.$el.find('.jndi-name', that.$el).val(),
                dialect = this.$el.find('.dialect', that.$el).val(),
                driver = this.$el.find('.driver', that.$el).val(),
                url = this.$el.find('.url', that.$el).val(),
                username = this.$el.find('.username', that.$el).val(),
                password = this.$el.find('.password', that.$el).val(),
                data;
            if (type == 'JNDI') {
                if (!jndi) {
                    $(".jdbc.is-invalid", that.$el).show();
                    $("#jdbc-message", that.$el).text("Missing jndi string");
                    return;
                }
                data = {
                    jndi : jndi,
                    mode : mode,
                    initSQL : initSQL,
                    schema : schema,
                    closeSQL : closeSQL,
                    testSQL : testSQL
                }
            } else if (type == 'DIRECT') {
                if (!driver || !url) {
                    $(".jdbc.is-invalid", that.$el).show();
                    $("#jdbc-message", that.$el).text("Missing driver/url");
                    return;
                }
                data = {
                    className : driver,
                    dialect: dialect,
                    url : url,
                    username : username,
                    password: password,
                    mode : mode,
                    initSQL : initSQL,
                    schema : schema,
                    closeSQL : closeSQL,
                    testSQL : testSQL
                }
            } else {
                $(".jdbc.is-invalid", that.$el).show();
                $("#jdbc-message", that.$el).text("Wrong type: " + type);
                return;
            }
            $.ajax({
                url : AJS.contextPath() + "/rest/playsql/1.0/admin/testJdbc",
                data : data,
                cache: false,
                success : function(data) {
                    if (data.status == "ok") {
                        $(".jdbc.is-valid", that.$el).show();
                        $("#jdbc-message", that.$el).text("Result: " + data.testSQL);
                    } else {
                        $(".jdbc.is-invalid", that.$el).show();
                        $("#jdbc-message", that.$el).text(data.exception);
                    }
                }
            });
        },
        submit : function(e) {
            var type = this.$el.find('.select.type').val(),
                mode = this.$el.find('#mode').val(),
                initSQL = this.$el.find('#initsql').val(),
                schema = this.$el.find('#schema').val(),
                closeSQL = this.$el.find('#closesql').val(),
                testSQL = this.$el.find('#testsql').val(),
                jndi = this.$el.find('#jndi-name').val(),
                driver = this.$el.find('#driver').val(),
                dialect = this.$el.find('#dialect').val(),
                url = this.$el.find('#url').val(),
                username = this.$el.find('#username').val(),
                password = this.$el.find('#password').val(),
                that = this;
            $('.submit.message', this.$el).empty();
            $('.submit', this.$el).hide();
            this.model.save({
                type : type,
                mode : mode,
                initSQL : initSQL,
                schema : schema,
                closeSQL : closeSQL,
                testSQL : testSQL,
                jndiName : jndi,
                driver : driver,
                dialect : dialect,
                url : url,
                username : username,
                password : password
            }, {
                success : function(model) {
                    that.render();
                },
                error : function(model, xhr) {
                    var error = "Error";
                    if (xhr.statusText) error += " - " + xhr.statusText;
                    if (xhr.responseText) {
                        try {
                            var response = JSON.parse(xhr.responseText);
                            if (response.message) error += " - " + response.message;
                            else error += response;
                        } catch (e) {
                            if (xhr.responseText) error += " - " + xhr.responseText;
                        }
                    }
                    $('.submit-message', that.$el).text(error);
                    $('.submit', that.$el).show();
                }
            });
            return false;
        },
        remove : function(e) {
            e.preventDefault();
            $('.remove', this.$el).hide();
            if (this.model.get('contextName') == 'Factory Settings') {
                throw "Can't delete the factory settings";
            }
            var type,
                that = this;
            if (this.model.get('contextName')) {
                type = 'DISABLED';
            } else {
                type = 'INHERIT';
            }
            this.model.save({
                type : type
            }, {
                success : function(model) {
                    //model.fetch();
                    that.model.set('deleted', true);
                    that.render();
                },
                error : function(model, xhr) {
                    var error = "Error";
                    if (xhr.statusText) error += " - " + xhr.statusText;
                    if (xhr.responseText) {
                        try {
                            var response = JSON.parse(xhr.responseText);
                            if (response.message) error += " - " + response.message;
                            else error += response;
                        } catch (e) {}
                    }
                    $('.remove', that.$el).show();
                    $('.submit-message', that.$el).text(error);
                }
            });
        }
    });


    var ListView = Backbone.View.extend({
        el: $('#datasource-list'),
        initialize: function(){
            _.bindAll(this, 'render', 'addItem', 'appendItem');
            this.collection = new List();
            this.collection.bind('add', this.appendItem);
            this.collection.bind('remove', this.removeItem);
            this.collection.fetch({add: true});
        },
        render: function(){
            var self = this;
            _(this.collection.models).each(function(item){
                self.appendItem(item);
            }, this);
            return this;
        },
        events: {
            'click #datasource-actions a': 'addItem'
        },
        addItem: function(){
            var this_collection = this.collection;
            Confluence.PlaySQL.SpacePicker.createDialog("Add a datasource", "", function(spaceKey, spaceName){
                var item = new Connection({
                    type: 'INHERIT',
                    contextSpaceKey: spaceKey
                });
                listView.collection.add(item);
                listView.editItem(item);
            }).show();
        },
        appendItem: function(item){
            var itemView = new ConnectionView({
                    model: item
                }),
                render = itemView.render();
            $(render.el).appendTo(this.el).data('backbone-view', itemView);
        },
        removeItem: function(item){
            $(this.el).each(function(){
                var bbView = $(this).data('backbone-view'),
                    contextSpaceKey = item.contextSpaceKey,
                    contextName = item.contextName;
                if (bbView && bbView.model
                    && bbView.get("contextSpaceKey") == contextSpaceKey
                    && bbView.get("contextName") == contextName) {
                    bbView.remove();
                    return;
                }
            });
        },
        editItem: function(item) {
            var contextSpaceKey = item.get("contextSpaceKey"),
                contextName = item.get("contextName");
            $(this.el).find("tr").each(function(){
                var bbView = $(this).data('backbone-view');
                if (bbView && bbView.model
                    && bbView.model.get("contextSpaceKey") == contextSpaceKey
                    && bbView.model.get("contextName") == contextName) {
                    bbView.edit();
                    return;
                }
            });
        }
    });
    var listView = new ListView();
    listView.render();
});