(function($){
    Confluence.PlaySQL = Confluence.PlaySQL || {};

    Confluence.PlaySQL.Helpers = Confluence.PlaySQL.Helpers || {};
    Confluence.PlaySQL.Helpers.generateKey = function (name, defaultCase) {
        var diacritics = {
            "A": /[\300-\306]/g,
            "a": /[\340-\346]/g,
            "E": /[\310-\313]/g,
            "e": /[\350-\353]/g,
            "I": /[\314-\317]/g,
            "i": /[\354-\357]/g,
            "O": /[\322-\330]/g,
            "o": /[\362-\370]/g,
            "U": /[\331-\334]/g,
            "u": /[\371-\374]/g,
            "N": /[\321]/g,
            "n": /[\361]/g,
            "C": /[\307]/g,
            "c": /[\347]/g
        };
        var result = name;
        for (var d in diacritics) {
            result = result.replace(diacritics[d], d);
        }
        if (defaultCase == "UPPER") {
            result = result.toUpperCase();
        } else if (defaultCase == "LOWER") {
            result = result.toLowerCase();
        } else {
            throw "Unknown case: " + defaultCase;
        }
        result = result
            .replace(/[- =+\/\\]/g, '')
            .replace(/[^a-zA-Z0-9_]/g, '')
            .replace(/^[_0-9]+|_$/g, '');

        return result;
    };

    AJS.$(function(){
        AJS.$(".playsql-query-placeholder table[data-query-editor-url] thead tr.playsql-col-headers th a").first().each(function(){
            var $this = $(this),
                href = $this.closest("table").attr("data-query-editor-url");
            if (href) $this.attr("href", href);
        });
    });

    Confluence.PlaySQL.Helpers.parseOrderByClause = function(clause) {
        var parsed = [],
            lengthParsed = 0,
            part, oPart;
        while (part = getNextExpression(clause.substr(lengthParsed), ',')) {
            lengthParsed += part.length;
            if (part.length == 0) break;
            oPart = {
                expression: part
            };
            chop(oPart, ',');
            chop(oPart, 'FIRST');
            chop(oPart, 'LAST');
            chop(oPart, 'NULLS');
            chop(oPart, 'ASC');
            chop(oPart, 'DESC');
            parsed.push({
                field: $.trim(oPart.expression),
                way: oPart.ASC ? "ASC" : oPart.DESC ? "DESC" : undefined,
                nulls: oPart.FIRST ? "FIRST" : oPart.LAST ? "LAST" : undefined
            });
        }
        return parsed;
    }

    Confluence.PlaySQL.Helpers.parseSelectClause = function(clause) {
        var result = ["SELECT"],
            expression,
            eaten = "SELECT".length;
        while ((expression = getNextExpression(clause.substr(eaten+1), ",")).length > 0) {
            eaten += expression.length;
            result.push({
                def: chop($.trim(expression), ",")
            });
        }
        return result;
    }

    var FROM_OPERATORS = ["LEFT JOIN", "RIGHT JOIN", "FULL JOIN",
        "LEFT OUTER JOIN", "RIGHT OUTER JOIN", "FULL OUTER JOIN",
        "INNER JOIN", "JOIN", "CROSS JOIN"];
    function getOperator(text) {
        for (var i = 0; i < FROM_OPERATORS.length; i++)
            if (text.indexOf(FROM_OPERATORS[i]) == 0)
                return FROM_OPERATORS[i];
        return undefined;
    }

    Confluence.PlaySQL.Helpers.parseFromClause = function(clause) {
        var parts = splitKeywords.apply(this, [clause.substr("FROM".length)].concat(FROM_OPERATORS)),
            expression;
        // Split FROM clauses when there are commas
        for (var i = 0; i < parts.length; i++) {
            var fromPart = parts[i],
                eaten = 0,
                result = [];
            if (fromPart.indexOf(",") != -1) {
                while ((expression = getNextExpression(fromPart.substr(eaten), ",")).length > 0) {
                    eaten += expression.length;
                    result.push($.trim(expression));
                }
                if (result.length > 1) {
                    parts.splice(i, 1);
                    for (var k = result.length - 1; k >= 0; k--) {
                        parts.splice(i, 0, result[k]);
                    }
                }
            }
        }
        for (var i = 0; i < parts.length; i++) {
            var part = $.trim(parts[i]),
                operator = getOperator(part),
                afterOperator = $.trim(part.substr(operator == undefined ? 0 : operator.length)),
                tableDef = getNextExpression(afterOperator, " "),
                aliasAndCriteria = tableDef.length < afterOperator.length ? $.trim(afterOperator.substr(tableDef.length)) : undefined,
                criteria = aliasAndCriteria != undefined ? splitKeywords(aliasAndCriteria, "ON") : undefined;
            parts[i] = {
                operator : operator,
                def: $.trim(tableDef),
                alias: criteria && $.trim(criteria[0]),
                criteria: criteria && criteria.length > 1 ? criteria.slice(1).join(" ") : undefined
            };
        }
        parts.unshift("FROM");
        return parts;
    }

    Confluence.PlaySQL.Helpers.buildOrderByClause = function(parsed) {
        var elements = parsed.map(function(p){
            var expression = p.field;
            if (p.way) expression+= ' ' + p.way;
            if (p.nulls) expression+= ' NULLS ' + p.nulls;
            return expression;
        });
        return elements.join(", ");
    }

    function chop(oPart, lastWord) {
        if (typeof oPart != "object") {
            oPart = {expression: oPart};
        }

        var trimmed = $.trim(oPart.expression);
        if (trimmed.substr(trimmed.length - lastWord.length).toLowerCase() == lastWord.toLowerCase()) {
            oPart.expression = $.trim(trimmed.substr(0, trimmed.length - lastWord.length));
            oPart[lastWord] = true;
        } else {
            oPart[lastWord] = false;
        }
        return oPart.expression;
    }

    /** Gets the next full expression, without entering into parenthesis or quotes */
    function getNextExpression(clause, delimiter) {
        var expression = '',
            next,
            delimiterNN = delimiter != null ? delimiter.charAt(0) : undefined;
        for (var i = 0 ; i < clause.length ; i++) {
            expression += clause.charAt(i);
            next = clause.charAt(i);
            switch (next) {
                // HERE we should check for comments "--.*\n"
                case "(":
                    expression += getNextExpression(clause.substr(i+1), ')');
                    i = expression.length - 1;
                    break;
                case "'":
                    expression += getString(clause.substr(i+1), "'");
                    i = expression.length - 1;
                    break;
                case "\"":
                    expression += getString(clause.substr(i+1), "\"");
                    i = expression.length - 1;
                    break;
                case delimiterNN:
                    if (clause.substr(i, delimiter.length) == delimiter) {
                        return expression;
                    }
                    break;
            }
        }
        return expression;
    }

    function getString(clause, end) {
        var escaped = false,
            expression = '',
            next;
        for (var i = 0; i < clause.length; i++) {
            next = clause.charAt(i);
            if (escaped) {
                escaped = false;
            } else if (clause.substr(i, end.length) == end) {
                expression += end;
                return expression;
            } else if (next == '\\') {
                escaped = true;
            }
            expression += next;
        }
        return expression;
    }

    function splitKeywords(text, keyword1, keywordn) {
        var keywords = Array.prototype.slice.call(arguments, 1),
            NON_ALPHANUM = /[^a-zA-Z0-9_]/;
        for (var i in keywords) {
            var keyword = keywords[i];
            // Find all keywords in the string. If any is valid, split the work in 2.
            var position = 0;
            while ((position = text.toLowerCase().indexOf(keyword.toLowerCase(), position+1)) > 0) {
                // If we're matching the first letters of the text, we might be matching a keyword that's already complete
                var isSubKeyword = false;
                for (var j = 0; j < i; j++) {
                    var positionOfSubKeyword = keywords[j].indexOf(keyword);
                    if (positionOfSubKeyword == position) {
                        // Bingo, our keyword is part of keyword j, and keyword j starts at the beginning of text
                        isSubKeyword = true;
                        break;
                    }
                }
                if (isSubKeyword) continue;

                // If it's a full word
                if (NON_ALPHANUM.test(text.charAt(position-1))
                    && (text.length <= (position + keyword.length)
                    || NON_ALPHANUM.test(text.charAt(position + keyword.length)))) {
                    var firstHalf = text.substr(0, position),
                        secondHalf = text.substr(position),
                        firstChops = splitKeywords.apply(this, [firstHalf].concat(keywords)),
                        secondChops = splitKeywords.apply(this, [secondHalf].concat(keywords));
                    // We've split the array recursively. Just return.
                    return firstChops.concat(secondChops);
                }
            }
        }
        // Keywords not found
        return [text];
    }

    function isBalanced(expression) {
        var test = expression + "))";
        return getNextExpression(test, ")").length != test.length;
    }

    function mergeIncompleteExpressions(parts) {
        var result = [];
        for (var i = 0 ; i < parts.length ; i++) {
            var expression = parts[i];
            while (!isBalanced(expression)) {
                // Merge one more element
                i++;
                if (i >= parts.length) {
                    break;
                }
                expression += parts[i];
            }
            result.push(expression);
        }
        return result;
    }

    Confluence.PlaySQL.Helpers.rebuildSQL = function(parsedSQL) {
        var result = "";
        for (var i = 0 ; i < parsedSQL.length ; i++) {
            var part = parsedSQL[i];
            if (part == undefined || part.length == 0) continue;
            if (part[0] == "SELECT") {
                result += "\n" + part[0];
                for (var j = 1; j < part.length; j++) {
                    var field = part[j];
                    if (j>1) result += ", ";
                    if (field.def != undefined) result += " " + field.def;
                    if (field.alias != undefined) result += " " + field.alias;
                }
            } else if (part[0] == "FROM") {
                result += "\n" + part[0];
                for (var j = 1; j < part.length; j++) {
                    var field = part[j];
                    if (j>1) result += "\n    ";
                    if (field.operator != undefined) result += " " + field.operator;
                    if (field.def != undefined) result += " " + field.def;
                    if (field.alias != undefined) result += " " + field.alias;
                    if (field.criteria != undefined) result += " " + field.criteria;
                }
            } else if (part.length > 1 && part[0] == "--" && part[1] == "Bilinear:") {
                // Do not insert that
            } else {
                result += "\n" + part[0];
                if (part.length > 0) {
                    result += " " + part.slice(1).join(" ");
                }
            }
        }
        return $.trim(result);
    }

    Confluence.PlaySQL.Helpers.isSQLEqual = function(sql1, sql2) {
        if (typeof sql1 == "object") {
            sql1 = Confluence.PlaySQL.Helpers.rebuildSQL(sql1);
        }
        if (typeof sql2 == "object") {
            sql2 = Confluence.PlaySQL.Helpers.rebuildSQL(sql2);
        }
        sql1 = $.trim(sql1.replace(/\s+/g, ' '));
        sql2 = $.trim(sql2.replace(/\s+/g, ' '));
        return sql1 == sql2;
    }

    var SQL_ROOT_KEYWORDS = ["SELECT", "FROM", "WHERE", "ORDER BY", "GROUP BY", "HAVING", "LIMIT", "OFFSET", "--"];
    var SQL_ROOT_KEYWORDS_AND_ROWSTARTERS = SQL_ROOT_KEYWORDS.concat(["--"]);

    Confluence.PlaySQL.Helpers.parseSQL = function(sql) {
        var parts = splitKeywords.apply(this, [sql].concat(SQL_ROOT_KEYWORDS_AND_ROWSTARTERS)),
            result = [];
        parts = mergeIncompleteExpressions(parts);
        for (var i in parts) {
            var part = parts[i],
                partUpper = part.toUpperCase();
            if (partUpper.indexOf("SELECT") == 0) {
                result.push(Confluence.PlaySQL.Helpers.parseSelectClause(part));
            } else if (partUpper.indexOf("FROM") == 0) {
                result.push(Confluence.PlaySQL.Helpers.parseFromClause(part));
            } else if (partUpper.indexOf("WHERE") == 0) {
                result.push([part.substr(0, "WHERE".length), part.substr("WHERE".length)]);
            } else if (partUpper.indexOf("ORDER BY") == 0) {
                result.push([part.substr(0, "ORDER BY".length), part.substr("ORDER BY".length)]);
            } else if (partUpper.indexOf("GROUP BY") == 0) {
                result.push([part.substr(0, "GROUP BY".length), part.substr("GROUP BY".length)]);
            } else {
                result.push([part]);
            }
        }

        try {
            var isBilinear = Confluence.PlaySQL.Helpers.isSQLEqual(sql, result) ? "true" : "false";
            isBilinear = isBilinear
            && result.length > 1
            && result[0][0] == "SELECT"
            && result[1][0] == "FROM";
            result.push(["--", "Bilinear:", isBilinear]);
            result.bilinear = isBilinear;
        } catch (err) {
            // Do nothing
        }

        /** Get or set the clause */
        result.clause = function(clauseName, value) {
            for (var i = 0; i < this.length; i++) {
                if (this[i].length > 1 && this[i][0] == clauseName) {
                    // If the value is set... and if it's not already the value.
                    if (value != undefined && value !== this[i]) {
                        this[i].splice(0, this[i].length, value);
                    }
                    return this[i];
                }
            }
            if (value != undefined) {
                // Find a spot for it
                var keywordId = SQL_ROOT_KEYWORDS.indexOf(clauseName);
                if (keywordId == -1) {
                    this.push(value);
                    return this[this.length-1];
                } else {
                    // Try putting it before any next keyword
                    while (++keywordId < SQL_ROOT_KEYWORDS.length) {
                        for (var i = 0; i < this.length; i++) {
                            if (this[i].length > 1 && this[i][0] == clauseName) {
                                this.splice(i, 0, value);
                                return this[i];
                            }
                        }
                    }
                    // Not found. Append at the end.
                    this.push(value);
                    return this[this.length-1];
                }
            }
            return undefined;
        }

        return result;
    }

    Confluence.PlaySQL.base64_decode = function (data) {
        // http://kevin.vanzonneveld.net
        // +   original by: Tyler Akins (http://rumkin.com)
        // +   improved by: Thunder.m
        // +      input by: Aman Gupta
        // +   improved by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
        // +   bugfixed by: Onno Marsman
        // +   bugfixed by: Pellentesque Malesuada
        // +   improved by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
        // +      input by: Brett Zamir (http://brett-zamir.me)
        // +   bugfixed by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
        // *     example 1: base64_decode('S2V2aW4gdmFuIFpvbm5ldmVsZA==');
        // *     returns 1: 'Kevin van Zonneveld'
        // mozilla has this native
        // - but breaks in 2.0.0.12!
        //if (typeof this.window['atob'] === 'function') {
        //    return atob(data);
        //}
        var b64 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
        var o1, o2, o3, h1, h2, h3, h4, bits, i = 0,
            ac = 0,
            dec = "",
            tmp_arr = [];

        if (!data) {
            return data;
        }

        data += '';

        do { // unpack four hexets into three octets using index points in b64
            h1 = b64.indexOf(data.charAt(i++));
            h2 = b64.indexOf(data.charAt(i++));
            h3 = b64.indexOf(data.charAt(i++));
            h4 = b64.indexOf(data.charAt(i++));

            bits = h1 << 18 | h2 << 12 | h3 << 6 | h4;

            o1 = bits >> 16 & 0xff;
            o2 = bits >> 8 & 0xff;
            o3 = bits & 0xff;

            if (h3 == 64) {
                tmp_arr[ac++] = String.fromCharCode(o1);
            } else if (h4 == 64) {
                tmp_arr[ac++] = String.fromCharCode(o1, o2);
            } else {
                tmp_arr[ac++] = String.fromCharCode(o1, o2, o3);
            }
        } while (i < data.length);

        dec = tmp_arr.join('');

        return dec;
    };


    /*
     Copyright Vassilis Petroulias [DRDigit]

     Licensed under the Apache License, Version 2.0 (the "License");
     you may not use this file except in compliance with the License.
     You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

     Unless required by applicable law or agreed to in writing, software
     distributed under the License is distributed on an "AS IS" BASIS,
     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     See the License for the specific language governing permissions and
     limitations under the License.
     */
    var B64 = {
        alphabet: 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=',
        lookup: null,
        ie: /MSIE /.test(navigator.userAgent),
        ieo: /MSIE [67]/.test(navigator.userAgent),
        decode: function (s) {
            var buffer = B64.fromUtf8(s),
                position = 0,
                len = buffer.length;
            if (B64.ieo) {
                result = [];
                while (position < len) {
                    if (buffer[position] < 128) result.push(String.fromCharCode(buffer[position++]));
                    else if (buffer[position] > 191 && buffer[position] < 224) result.push(String.fromCharCode(((buffer[position++] & 31) << 6) | (buffer[position++] & 63)));
                    else result.push(String.fromCharCode(((buffer[position++] & 15) << 12) | ((buffer[position++] & 63) << 6) | (buffer[position++] & 63)));
                }
                return result.join('').replace(/\0/g,"");
            } else {
                result = '';
                while (position < len) {
                    if (buffer[position] < 128) result += String.fromCharCode(buffer[position++]);
                    else if (buffer[position] > 191 && buffer[position] < 224) result += String.fromCharCode(((buffer[position++] & 31) << 6) | (buffer[position++] & 63));
                    else result += String.fromCharCode(((buffer[position++] & 15) << 12) | ((buffer[position++] & 63) << 6) | (buffer[position++] & 63));
                }
                return result.replace(/\0/g,"");
            }
        },
        fromUtf8: function (s) {
            var position = -1,
                len, buffer = [],
                enc = [undefined, undefined, undefined, undefined];
            if (!B64.lookup) {
                len = B64.alphabet.length;
                B64.lookup = {};
                while (++position < len)
                    B64.lookup[B64.alphabet[position]] = position;
                position = -1;
            }
            len = s.length;
            while (position < len) {
                enc[0] = B64.lookup[s.charAt(++position)];
                enc[1] = B64.lookup[s.charAt(++position)];
                buffer.push((enc[0] << 2) | (enc[1] >> 4));
                enc[2] = B64.lookup[s.charAt(++position)];
                if (enc[2] == 64) break;
                buffer.push(((enc[1] & 15) << 4) | (enc[2] >> 2));
                enc[3] = B64.lookup[s.charAt(++position)];
                if (enc[3] == 64) break;
                buffer.push(((enc[2] & 3) << 6) | enc[3]);
            }
            return buffer;
        }
    };
    Confluence.PlaySQL.B64 = B64;

    Confluence.PlaySQL.Totals = Confluence.PlaySQL.Totals || {};


    function getTotal(totalsData, index, value) {
        if (typeof index == "object") {
            value = index;
            index = undefined;
        }
        if (value != undefined) {
            if (index == undefined) {
                totalsData = value;
            } else {
                while (totalsData.length < index) {
                    totalsData.push({});
                }
                if (totalsData.length == index) {
                    totalsData.push(value);
                } else {
                    totalsData[index] = value;
                }
            }
        }
        if (index == undefined) {
            return totalsData;
        } else {
            return totalsData.length > index ? totalsData[index] : undefined;
        }
    }

    Confluence.PlaySQL.Totals.editTotal = function (totalsData, $cell) {
        $cell.attr('data-rendered', $cell.html());
        var index = $cell.index() - 1,
            value = getTotal(totalsData, index),
            label = value && value.label || "",
            calculation = value && value.calculation || "";
        $cell.addClass("editing selected focused")
            .html(Confluence.Templates.PlaySQL.editTotalsCell({
                label: label,
                calculation: calculation}));
        $cell.find("input:first").focus();
        return false;
    };
    Confluence.PlaySQL.Totals.addTotal = function(totals, $cell, position) {
        if (totals == undefined || totals.length == 0) return;
        var index = $cell.index() - 1,
            colspan = parseInt($cell.attr('colspan')),
            thisTotal = totals[index];
        if (colspan == undefined || colspan > 1) {
            if (colspan == undefined ) colspan = 2;
            colspan--;
            $cell.attr('colspan', colspan)
                .removeClass("selected focused");
            thisTotal.colspan = colspan;
            // This is inserting a copy of the total
            totals.splice(position < 0 ? index : index + 1, 0, thisTotal);
            // Insert cell on the side
            var $newCell = $("<td></td>");
            if (position < 0) $newCell.insertBefore($cell);
            else $newCell.insertAfter($cell);
            // Edit the cell
            Confluence.PlaySQL.Totals.editTotal(totals, $newCell);
        } else {
            // How come the menu was proposed for a non-spanned cell?
        }
    };
    Confluence.PlaySQL.Totals.removeTotal = function(totals, $cell, callbackSaveTotals) {
        var index = $cell.index() - 1,
            colspan = parseInt($cell.attr('colspan')),
            adjacentIndex = index > 0 ? index-1 : totals.length > 1 ? 1 : undefined,
            adjacentTotal = adjacentIndex != undefined ? totals[adjacentIndex] : undefined,
            $adjacentCell = adjacentIndex == undefined ? undefined : adjacentIndex > index ? $cell.next() : $cell.prev();
        if (adjacentIndex != undefined) {
            adjacentTotal.colspan = (parseInt($adjacentCell.attr('colspan')) || 1) + colspan;
            $adjacentCell.attr('colspan', adjacentTotal.colspan);
            $cell.remove();
        } else {
            // Just blank it out
            $cell.empty();
        }
        totals.splice(index, 1);
        if (callbackSaveTotals) callbackSaveTotals();
    };
    Confluence.PlaySQL.Totals.saveTotal = function(totals, $cell) {
        Confluence.PlaySQL.Selection.inlineDialog && Confluence.PlaySQL.Selection.inlineDialog.hide();
        var index = $cell.index() - 1,
            value = {
                label : $cell.find("input.totals-label").val(),
                calculation : $cell.find("input.totals-calculation").val(),
                colspan : $cell.attr('colspan'),
                name : "totals_" + index
            };
        $cell.empty().attr("data-label", value.label)
            .attr("data-calculation", value.calculation)
            .attr("colspan", value.colspan)
            .attr("data-name", value.name)
            .text(value.label)
            .removeClass("editing")
            .append($("<em></em>").text(" =" + value.calculation));
        totals[index] = value;
    };
    /** Update the span records of the totals and pagination */
    Confluence.PlaySQL.Totals.updateColspans = function(totals, $placeholder) {
        if (totals == null) return;
        // 1. The Totals
        var $cells = $placeholder.find("tfoot .totals-row td:not(.right-padding)");
        if (totals.length == 0) return;
        if (totals.length < $cells.length) {
            throw "Totals: A model is missing... $1 models and $2 cells."
                .replace("$1", totals.length)
                .replace("$2", $cells.length);
        } else if (totals.length > $cells.length) {
            totals.splice(0, totals.length - $cells.length);
        }
        var totalSpan = 0,
            totalHeaderSpan = $placeholder.find("thead tr.playsql-col-headers th:not(:first-child):not(.add-column)").size(),
            span;
        for (var i = 0 ; i < totals.length ; i++) {
            span = parseInt($cells.eq(i).attr("colspan")) || 1;
            totals[i].colspan = span;
            totalSpan += span;
        }
        var diff = totalSpan - totalHeaderSpan;
        if (totals.length > 0 && diff < 0) {
            totals[0].colspan += Math.abs(diff);
        } else if (diff > 0) {
            $cells.each(function(i) {
                var $cell = $(this),
                    span = parseInt($cell.attr("colspan")) || 1,
                    exchange;
                if (span >= 1 + diff) {
                    exchange = span - Math.max(1, span - diff);
                    span = span - exchange;
                    diff = diff - exchange;
                    // Update the real life
                    $cell.attr("colspan", span);
                    totals[i].colspan = span;
                }
            });
        }
        // 2. The pagination
        // Beware, there are two pagination rows
        var $paginations = $placeholder.find(".pagination-row td:not(.right-padding)");
        $paginations.attr("colspan", totalHeaderSpan);
    };

    var replaceBuiltInVariablesCache = undefined;
    Confluence.PlaySQL.Helpers.replaceBuiltInVariables = function(value) {
        if (value == null || value.indexOf("$") == -1) return value;
        if (!replaceBuiltInVariablesCache) {
            replaceBuiltInVariablesCache = {
                "$pageId": AJS.Meta.get("page-id"),
                "$spaceKey": AJS.Meta.get("space-key"),
                "$pageTitle": AJS.Meta.get("page-title"),
                "$userFullName": AJS.Meta.get("user-display-name"),
                "$user": AJS.Meta.get("remote-user")
            };
        }
        for (var key in replaceBuiltInVariablesCache) {
            value = value.replace(key, replaceBuiltInVariablesCache[key]);
        }
        return value;
    };

    Confluence.PlaySQL.Helpers.hasMinVersion = function(minVersion) {
        var version = AJS.Meta.get("version-number"),
            verT = version.split(/\./),
            minT = minVersion.split(/\./);

        if (verT.length < 3) return true;
        if (minT.length != 3) throw "Can't recognize version number: " + minVersion;

        for (var i = 0 ; i < 3 ; i++) {
            if (parseInt(verT[i]) < parseInt(minT[i])) return false;
            if (parseInt(verT[i]) > parseInt(minT[i])) return true;
        }
        return true;
    }

    Confluence.PlaySQL.Helpers.parsePerms = function(perms) {
        var result = [];
        if (!perms) return result;
        if (typeof perms === "string") {
            perms = perms.split(/[ ,]+/);
        }
        for (var i = 0 ; i < perms.length ; i++) {
            if (perms[i] == 'EDIT')
                perms.push('VIEW', 'ADD_ROWS', 'EDIT_CELLS', 'REMOVE_ROWS', 'EDIT_ROWS_METADATA', 'Q_VIEW_EDIT');
            if (perms[i] == 'ADMIN_COLS')
                perms.push('EDIT', 'EDIT_TOTALS', 'ADD_COLS', 'EDIT_COLS_FE', 'EDIT_COLS_BE', 'REMOVE_COLS');
            if (perms[i] == 'CONTEXT_ADMIN')
                perms.push('ADMIN_COLS', 'REMOVE_TABLE');
            if (perms[i] == 'Q_VIEW')
                perms.push('VIEW');
            if (perms[i] == 'Q_VIEW_EDIT')
                perms.push('Q_VIEW');
            if (perms[i] == 'Q_EDIT')
                perms.push('Q_VIEW_EDIT', 'EDIT_TOTALS', 'EDIT_COLS_FE');
            if (perms[i] == 'Q_SAVE')
                perms.push('Q_EDIT');
        }
        for (var i = 0 ; i < perms.length ; i++)
            if (result.indexOf(perms[i]) == -1)
                result.push(perms[i]);
        /**
         * Tells whether the user has a permission.
         * @param perm
         * @param column the column. Optional. Only provide the column if the permission is Q_VIEW_EDIT.
         * @returns {boolean}
         */
        result.hasPerm = function(perm, column) {
            var result = this.indexOf(perm) != -1;
            if (result && column
                && !column.formats["joint-id"]
                && !column.formats["joint-table"]) {
                // result = true, but this is not an editable column.
                return false;
            }
            return result;
        }
        return result;
    }

    Confluence.PlaySQL.Helpers.intersect = function(perms1, perms2) {
        var result = [];
        for (var i in perms1) {
            if (perms2.indexOf(perms1[i]) != -1) {
                result.push(perms1[i]);
            }
        }
        if (!perms1.hasPerm) throw "Confluence.PlaySQL.Helpers.intersect() requires authentic permissions as a first arg";
        result.hasPerm = perms1.hasPerm;
        return result;
    }

    Confluence.PlaySQL.Helpers.parsePagination = function(pagination) {
        var offset = pagination && pagination.offset || 0,
            limit = pagination && pagination.limit || 20,// Can't be zero, so can't div/0
            count = pagination && pagination.count || 0,
            pageEnd = Math.min(offset + limit - 1, count - 1),
            lastPage = Math.floor((count - 1) / limit) * limit;
        if (count < limit && offset == 0) return null;
        // Everything is 0-based in this object, but please make it 1-based for text you display
        return {
            offset: offset,
            limit: limit,
            count: count, // 1-based.
            pageEnd: pageEnd, // 1-based
            firstPage: offset == 0 ? undefined : 0,
            previousPage: offset == 0 ? undefined : Math.max(0, (Math.floor(offset/limit)-1) * limit),
            nextPage: offset >= lastPage ? undefined : Math.min(lastPage, (Math.floor(offset/limit)+1) * limit),
            lastPage: offset >= lastPage ? undefined : lastPage
        };
    }

    Confluence.PlaySQL.Helpers.escapeSqlString = function(string, dialect) {
        string = string.replace(/[\0\n\r\b\t\\'"\x1a]/g, function (s) {
            switch (s) {
                case "\0":
                    return "\\0";
                case "\n":
                    return "\\n";
                case "\r":
                    return "\\r";
                case "\b":
                    return "\\b";
                case "\t":
                    return "\\t";
                case "\x1a":
                    return "\\Z";
                case "'":
                    return "''";
                case '"':
                    return '""';
                default:
                    return "\\" + s;
            }
        });
        return string;
    }



    Confluence.PlaySQL.Helpers.checkNotNull = function(value, message) {
        if (value == undefined) {
            throw "Unexpectedly undefined value: " + value + "(" + message + ")";
        }
        if (value instanceof $) {
            if (value.size() == 0)
                throw "Unexpectedly empty jquery element: " + value + "(" + message + ")";
        }
        return value;
    }

    /** Returns an explanation about the selection in the inputBox.
     * @param inputBox the box, not jquery
     * @param range the range. Optional. Default is window.getSelection().getRangeAt(0).
     * */
    Confluence.PlaySQL.Helpers.caretSelection = function (inputBox, range) {
        // Here's why we check ![type="date"]: If it's an actual date field, we can't get the cursor's position.
        // However: If it's a date field and the AUI datePicker kicked in, it transformed the type to [text] and
        // we can get the position.
        if ($(inputBox).attr("type") == "date") {
            return {};
        }
        if (inputBox.nodeName == 'INPUT' || inputBox.nodeName == 'TEXTAREA') {
            if ('selectionStart' in inputBox) {
                // Standard-compliant browsers
                return {
                    position: inputBox.selectionStart,
                    length: inputBox.selectionEnd - inputBox.selectionStart
                }
            } else if (document.selection) {
                // IE
                var sel = document.selection.createRange();
                var selLen = document.selection.createRange().text.length;
                sel.moveStart('character', -inputBox.value.length);
                return {
                    position: sel.text.length - selLen,
                    length: selLength,
                    el: inputBox
                }
            }
        } else if (inputBox.contentEditable == "true" && window.getSelection) {
            if (!range) {
                range = window.getSelection && (window.getSelection().rangeCount > 0) && window.getSelection().getRangeAt(0);
            }
            if (!range) return {};
            var preSelectionRange = range.cloneRange();
            // Modify this copy of a range so we can count the number of selected characters and deduce the start position
            preSelectionRange.selectNodeContents(inputBox);
            preSelectionRange.setEnd(range.startContainer, range.startOffset);
            return {
                position: preSelectionRange.toString().length,
                length: range.toString().length,
                range: range,
                el: range.startContainer
            };
        }
    }

    /**
     * Sets the beginning or end of the range to the desired position,
     * even if the node is comprised of child nodes.
     *
     * It is generally used for a window.getSelection(), but strictly speaking only
     * ranges are involved here.
     *
     * @param methodName the method to set the range. Default is 'setEnd'.
     **/
    Confluence.PlaySQL.Helpers.setSelection = function (range, node, index, methodName) {
        methodName = methodName || 'setEnd';
        if (node.nodeType == node.TEXT_NODE) {
            if (index >= 0) {
                range[methodName](node, index);
                return true;
            } else {
                range[methodName](node, 0);
                return false;
            }
        } else {
            var childNodes = node.childNodes;
            for (var i = 0 ; i < childNodes.length ; i++) {
                if (index == 0) {
                    range[methodName](childNodes[i], 0);
                    return true;
                } else if (childNodes[i].textContent.length > index) {
                    return Confluence.PlaySQL.Helpers.setSelection(range, childNodes[i], index, methodName);
                } else {
                    index -= childNodes[i].textContent.length;
                }
            }
            // We didn't find the element. Selecting the max.
            var lastChild = childNodes[childNodes.length - 1];
            while (lastChild.childNodes.length > 0)
                lastChild = lastChild.childNodes[lastChild.childNodes.length - 1];
            range[methodName](lastChild, lastChild.textContent.length);
            return false;
        }
    }

    Confluence.PlaySQL.Helpers.isMac = navigator.platform.toUpperCase().indexOf('MAC') >= 0;

    /** Analyzes an event and returns a canonical human-understandable key */
    Confluence.PlaySQL.Helpers.getKeyCombinationDetail = function(e) {
        var code = e.keyCode,
            key = String.fromCharCode(code).toLowerCase(),
            onKeyDown = true;
        switch (code) {
            case 8: key = "backspace"; break;
            case 9: key = "tab"; break;
            case 13: key = "enter"; /*onKeyDown = false;*/ break;
            case 27: key = "esc"; break;
            case 37: key = "left"; break;
            case 38: key = "up"; break;
            case 39: key = "right"; break;
            case 40: key = "down"; break;
            case 46: key = "delete"; break;
            case 112: key = "f1"; break;
            case 113: key = "f2"; break;
        }
        var modifiers = [];
        if (e.altKey) modifiers.push("alt");
        if (e.shiftKey) modifiers.push("shift");
        if (Confluence.PlaySQL.Helpers.isMac) {
            if (e.metaKey) modifiers.push("control");
        } else {
            if (e.ctrlKey) modifiers.push("control");
        }

        return {
            code: code,
            key: key,
            modifiers: modifiers,
            onKeyDown: onKeyDown,
            canonical: (modifiers.length ? modifiers.join("+") + "+" : "") + key
        };
    }

    function strip(text, opening, closing, escape) {
        for (var i = 0 ; i < text.length ; i++) {
            if (text.charAt(i) == opening) {
                var j = i + 1;
                while (j < text.length) {
                    if (text.charAt(j) == closing) {
                        break;
                    } else if (closing != undefined && text.charAt(j) == escape && text.charAt(j+1) == closing) {
                        j += 2;
                    } else {
                        j++;
                    }
                }
                text = text.substr(0, i + 1) + text.substr(j);
                i++;
            }
        }
        return text;
    }
    /** Parse a formula and hopes to return the text being entered at the current
     * position, and the enclosing formula
     */
    Confluence.PlaySQL.Helpers.parseFunctionAtCaret = function(text) {
        text = strip(text, "'", "'", "\\");
        var currentExpression = text.match(/[a-zA-Z][\.a-zA-Z0-9]*$/),
            parentFunctions = text.match(/[a-zA-Z][a-zA-Z0-9]*\(/g),
            parentFunction, argList, argCount, currentTable;
        if (parentFunctions) {
            parentFunction = parentFunctions[parentFunctions.length-1];
            argList = text.substr(text.lastIndexOf(parentFunction));
            argCount = (argList + ",").match(/,/g).length - 1;
        }
        currentExpression = currentExpression && currentExpression[0];
        currentTable = currentExpression && currentExpression.indexOf(".") != -1 && currentExpression.substr(0, currentExpression.indexOf("."));
        parentFunction = parentFunction && parentFunction.substr(0, parentFunction.length - 1);
        return {
            currentTable:currentTable,
            current:currentExpression,
            parentFn:parentFunction,
            argCount:argCount
        };
    }

    var functionList = {};
    Confluence.PlaySQL.Helpers.getFunctionList = function(context, spreadsheetId, callback) {
        if (functionList[spreadsheetId]) return callback(functionList[spreadsheetId]);
        var url = AJS.contextPath() + "/rest/playsql/1.0/autocomplete/" + context + "/" + spreadsheetId + "/functionList";
        $.ajax({
            url: url,
            type: 'GET',
            contentType: "application/json; charset=utf-8"
        }).done(function(data){
            data.constants = [
                { name: "TRUE", type: "Constant" },
                { name: "FALSE", type: "Constant" }
            ];
            functionList[spreadsheetId] = data;
            callback(data);
        }).fail(function(jqXHR, error, reason){
            AJS.log("Can't get autocomplete for function names");
        });
    }

    // Fetch the source, for the default case-sensibility of the dialect
    Confluence.PlaySQL.fetchContextProperties = function(fetchTableList, spaceKey, callback) {
        spaceKey = spaceKey || AJS.Meta.get("space-key");
        var url = AJS.contextPath() + "/rest/playsql-spreadsheets/1.0/space:" + spaceKey + "?" + $.param({"empty":fetchTableList?"false":"true"});
        $.ajax({
            url: url,
            type: 'GET',
            contentType: "application/json; charset=utf-8"
        }).done(function (data, status, jqXHR) {
            var defaultCase = jqXHR.getResponseHeader("X-DIALECT_DEFAULT_CASE"),
                mode = jqXHR.getResponseHeader("X-MODE");
            callback && callback(jqXHR, defaultCase, mode);
        });
    }

    Confluence.PlaySQL.createSpreadsheet = function(spaceKey, spreadsheet, copyId, defaultCase, successCallback, failureCallback) {
        if (!spreadsheet) throw "Spreadsheet expected";
        if (!spreadsheet.name) throw "Spreadsheet key expected";
        if (!spreadsheet.label) throw "Spreadsheet label expected";
        if (!spreadsheet.columns) {
            spreadsheet.columns = [{
                    name: (defaultCase == "UPPER" ? "FIRSTCOLUMN" : "firstcolumn"),
                    label: "First column",
                    type: "TEXT"
                }, {
                    name: (defaultCase == "UPPER" ? "SECONDCOLUMN" : "secondcolumn"),
                    label: "Second column",
                    type: "TEXT"
                }];
        }
        spaceKey = spaceKey || AJS.Meta.get("space-key");
        if (spaceKey == undefined) {
            alert("The space where the spreadsheet must be created wasn't specified" );
            return;
        }
        var url = AJS.contextPath() + "/rest/playsql-spreadsheets/1.0/space:" + spaceKey + "/" + spreadsheet.name;
        if (typeof copyId != "undefined") {
            url += "?copyOf=" + copyId;
        }
        AJS.$.ajax({
            url: url,
            type: 'PUT',
            contentType: "application/json; charset=utf-8",
            dataType: "json",
            data: JSON.stringify(spreadsheet)
        }).done(function(data){
            var hasOpener = false;
            try {
                hasOpener = window.opener &&
                    !window.opener.closed &&
                    window.opener.AJS &&
                    window.opener.AJS.MacroBrowser &&
                    window.opener.AJS.MacroBrowser.fields &&
                    window.opener.AJS.MacroBrowser.fields.spreadsheetId &&
                    window.opener.AJS.MacroBrowser.fields.spreadsheetId.updateSpreadsheet;
            } catch (err) {
                // Surely a CORS exception with window.opener coming from localhost / baseUrl
                hasOpener = false;
            }
            if (hasOpener) {
                window.opener.AJS.MacroBrowser.fields.spreadsheetId.loadedSpreadsheetId = "please refresh anyway";
                window.opener.AJS.MacroBrowser.fields.spreadsheetId.setValue(key);
                window.opener.AJS.MacroBrowser.fields.spreadsheetId.updateSpreadsheet();
                // That's when people use Hsql: We shouldn't make the Macro Browser access the db at the same time as changing page...
                successCallback && window.setTimeout(function(){successCallback(data);}, 700);
            } else {
                successCallback && successCallback(data);
            }
        }).fail(failureCallback);
    }

})(AJS.$);