(function($){
    Confluence.PlaySQL = Confluence.PlaySQL || {};

    Confluence.PlaySQL.Selection = {
        forTable: function ($placeholder) {
            var that = $placeholder.data('PlaySQLSelection');
            if (that == undefined) {
                that = {};
                $placeholder.data('PlaySQLSelection', that);
                that.options = {
                    cellSecondClick: function ($cell) {
                        this.options.editCell.call(this, $cell);
                    },
                    rowSecondClick: undefined,
                    columnSecondClick: undefined,
                    baseSelector: ".playsql-spreadsheet-placeholder table",
                    // Could be optimized by not withholding the $placeholder inside the data() of thedit $placeholder
                    $placeholder: $placeholder,
                    colOf : function($cell) {
                        return {index: $cell.index()};
                    },
                    popupTrigger: undefined,
                    popupContentFn: undefined,
                    pasteHandler: undefined,
                    addRow: undefined,
                    /** The cssClassMap allows to operate on other CSS classes than "selection" and "focused".
                     - Undefined = It uses "selected" and "focused",
                     - Any other name = It uses the name, and no focus.
                     It is useful when the user is editing a formula and we're highlighting cells.
                     If this mode is activated, only selection/keymove is active and not secondClicks.
                     */
                    cssClassMap: undefined
                };
                that.setOptions = function (options) {
                    $.extend(that.options, options);
                    return that;
                };
                that.init = function (options) {
                    that.setOptions(options);
                    if (that.alreadyInitialized) {
                        // Do nothing
                    } else {
                        that.alreadyInitialized = true;
                        Confluence.PlaySQL.Selection.attach($placeholder, that, {
                            "click thead tr.playsql-col-headers th:not(:first-child)": columnClick,
                            "mousedown tbody > tr > th:first-child, tfoot > tr > th:first-child": rowClick,
                            "mousedown tbody > tr > td, tfoot > tr > td": cellClick,
                            "mouseenter thead > tr.playsql-col-headers > th:not(:first):not(:last)": headerMouseEnter,
                            "mouseleave thead > tr.playsql-col-headers  > th:not(:first):not(:last)": headerMouseLeave,
                            "mousedown .playsql-resize-handle": resizeHandleMousedown,
                            "click .playsql-resize-handle": dodgeClick,
                            "key enter": pressEnter,
                            "key esc": pressEsc,
                            "key down": moveDown,
                            "key up": moveUp,
                            "key left": moveLeft,
                            "key right": moveRight,
                            "key tab": nextTabOrder,
                            "key shift+tab": previousTabOrder,
                            "key control+a": selectAll
                        });
                        Confluence.PlaySQL.Selection.attach($("#main"), that, {
                            "click" : function($element, e){
                                // If the clicked element is still attached to the DOM (e.g. if a click handler
                                // came in between)
                                if (e.target.parent != null) {
                                    if (this.options.$placeholder.has(e.target).length == 0) {
                                        // The click was outside the spreadsheet
                                        unselectAll.call(this);
                                    }
                                }
                            }
                        });
                        if (that.options.popupTrigger) {
                            Confluence.PlaySQL.Selection.inlineDialog = AJS.InlineDialog(that.options.popupTrigger, "playsql-table-popup",
                                function($content, trigger, showPopup) {
                                    that.options.popupContentFn.call(that, $content, trigger, showPopup);
                                    return false;
                                },
                                {
                                    useLiveEvents:true,
                                    offsetY: 20
                                }
                            );
                        }
                        options.onInit && options.onInit();
                    }
                    return that;
                };
                that.unselectAll = function() {
                    unselectAll.call(that);
                };
                that.mapCss = mapCss;
            }

            return that;
        },

        /**
         * Quickly delegate events using jquery
         * @param $base the element to bind the handlers to
         * @param that the 'selection' object to apply the callback handlers to. Mandatory.
         * @param handlers an array in the form:{
         *  "event table > td": function,
         *  "key control+ctrl+meta+shift+alt+key": function // control = ctrl on windows, meta on mac
         * }
         */
        attach: function ($base, that, handlers) {
            var event, selector, cutPosition,
                keyboardBindings = [],
                hasKeyBindings = false;
            // Bind all handlers on the $placeholder
            for (var key in handlers) {
                cutPosition = key.indexOf(" ");
                event = cutPosition == -1 ? key : key.substring(0, cutPosition);
                selector = cutPosition == -1 ? undefined : key.substring(cutPosition + 1);
                if (event.indexOf("key") == 0) {
                    keyboardBindings.push({key:selector.split("+"), handler: handlers[key]});
                    hasKeyBindings = true;
                } else {
                    (function (event, selector, handler, that) {
                        $base.on(event, selector, function(e) {
                            return handler.call(that, $(this), e);
                        });
                    })(event, selector, handlers[key], that);
                }
            }

            if (hasKeyBindings) {
                $(window.document).on("keydown.playsql-key-bindings keypress.playsql-key-bindings", function(e) {
                    // Dismiss the event if no cell is selected on the spreadsheet
                    if ($base.find(that.mapCss(".focused")).size() == 0) return;
                    // Dismiss the event if a dialog is shown
                    var $popups = $(".aui-inline-dialog:visible, .aui-popup:visible, .aui-blanket:visible");
                    var isDialogShown = false;
                    $popups.each(function(){
                        var $this = $(this),
                            id = $this.attr("id") || ("." + $this.attr("class"));
                        // Don't dismiss the handler
                        if ($this.is("#inline-dialog-playsql-table-popup, .aui-datepicker-dialog")) return;
                        if ($this.is("#inline-dialog-formula-editor")) return;
                        if ($this.is("#inline-dialog-cell-properties")) return;
                        if ($this.is("#inline-dialog-selection-action-panel")) return;
                        if ($this.find(".ui-datepicker").size() > 0) return;
                        if ($this.attr("aria-hidden") == "true") return;
                        // Alright, dismiss the handler
                        AJS.log("Dismissing the key/click handler - A dialog is shown - " + id);
                        isDialogShown = true;
                        return false;
                    });
                    if (isDialogShown) return;
                    var $focus = $("*:focus"),
                        isFocusOutsidePlaceholder = $focus.size() == 1 && $base.has($focus).size() == 0;
                    if (isFocusOutsidePlaceholder) return;

                    var keyCombinationDetail = Confluence.PlaySQL.Helpers.getKeyCombinationDetail(e),
                        key = keyCombinationDetail.key,
                        onKeyDown = keyCombinationDetail.onKeyDown,
                        modifiers = keyCombinationDetail.modifiers;

                    //console.trace("Captured ", key, modifiers.join(","), e.type, $base.attr("data-macro-spreadsheet-id"));

                    // Only execute left, up, right, down and ctrl+a on keydown. The others on keypress.
                    if (onKeyDown && e.type != 'keydown') return;
                    if (!onKeyDown && e.type == 'keydown') return;

                    function trigger(key, handler) {
                        if (handler) {
                            handler.call(that, e);
                        }
                    }

                    function isCorrectCombination(key, modifiers, keyCombination) {
                        // Check the character matches
                        if (key != keyCombination[keyCombination.length - 1]) return false;
                        // Check the modifiers match
                        if (modifiers.length + 1 != keyCombination.length) return false;
                        for (var j = 0 ; j < keyCombination.length - 1 ; j++) {
                            if (modifiers.indexOf(keyCombination[j]) == -1) return false;
                        }
                        return true;
                    }

                    for (var i in keyboardBindings) {
                        if (isCorrectCombination(key, modifiers, keyboardBindings[i].key)) {
                            //console.trace("Trigger", keyboardBindings[i]);
                            trigger(keyboardBindings[i].key, keyboardBindings[i].handler);
                        }
                    }
                });
            }
        },
        bindMenu : function($element, items) {
            var $ul = $("<ul></ul>").appendTo($("#custom-menu").empty());
            function addElement(label, fn, args) {
                var $li = $("<li></li>").appendTo($ul),
                    $a = $("<a></a>", {
                        //href: '#',
                        text: label
                    }).appendTo($li);
                $a.click(function(e){
                    $element.removeClass("aui-dropdown2-trigger")
                        .removeAttr('aria-controls aria-haspopup aria-owns');
                    fn.apply(this, args.concat([$(this), e]));
                    //return false;
                });
            }
            for (var item in items) {
                addElement(items[item][0], items[item][1], items[item].slice(2));
            }
            $element
                .attr('aria-owns', 'custom-menu')
                .attr('aria-haspopup', 'true')
                .addClass('aui-dropdown2-trigger');
        },
        /**
         * Setup the copy-paste. Once for a page only. It uses byPlaceholder($0) to dispatch the event
         * @param options { pasteHandler: spreadsheet.function(contenteditableDiv) }
         */
        setupCopyHandler: function(options) {
            options = options || {};
            var that = this;
            $(document).on("copy", function() {
                // 1. Check whether we're in the spreadsheet
                var selection = window.getSelection(),
                    isConnexe = selection.rangeCount <= 1,
                    uniqueRange = selection.rangeCount == 1 ? selection.getRangeAt(0) : undefined,
                    hasPopups = $(".aui-inline-dialog:visible, .aui-dialog:visible").size() > 0,
                    hasEditingFocusedCell = $("table.playsql td.editing *:focus").size() > 0;
                if (hasPopups || hasEditingFocusedCell) return;
                if (!isConnexe) return;
                if (uniqueRange) {
                    if (!uniqueRange.collapsed) {
                        var didUserSelectSomething = !uniqueRange.collapsed;
                        if (didUserSelectSomething) {
                            var isPlaySqlZeroWidthSpan =
                                uniqueRange.startContainer == uniqueRange.endContainer
                                && uniqueRange.startOffset == 0
                                && uniqueRange.endOffset == 1;
                            isPlaySqlZeroWidthSpan = isPlaySqlZeroWidthSpan && (
                            (uniqueRange.startContainer.nodeName == "DIV" && uniqueRange.startContainer.className == "zws-playsql") ||
                            (uniqueRange.startContainer.nodeType == 3 && uniqueRange.startContainer.textContent == String.fromCharCode(8203)));
                            if (isPlaySqlZeroWidthSpan)
                                didUserSelectSomething = false;
                        }
                        if (didUserSelectSomething) {
                            return;
                        }
                    }
                }

                // 2. Since nothing is selected, select the cells/rows/col
                // Note: We use .selected directly without mapping (mapCss()) because the copyHandler is page-level
                // and we can't make specifics for each ss. Therefore it is always the .selected with is copied and
                // not the .cell-ref-active classes.
                var $selected = $(".playsql-spreadsheet-placeholder .selected, .playsql-query-placeholder .selected"),
                    $copyDiv = $(".tmp-clipboard-contents").empty();
                if ($copyDiv.size() == 0) {
                    $copyDiv = $("<div></div>", { "class" : "tmp-clipboard-contents" }).appendTo($("body").first());
                }
                if ($selected.size() == 0) return;
                var $placeholder = $selected.closest(".playsql-spreadsheet-placeholder, .playsql-query-placeholder").first();
                $placeholder.find("tr[data-row-metadata]").each(function(){
                    var $this = $(this),
                        rowMetadata = $this.data("row-metadata");
                    if (rowMetadata) {
                        $this.attr("data-row-metadata", JSON.stringify(rowMetadata));
                    } else {
                        $this.removeAttr("data-row-metadata");
                    }
                });
                $placeholder = $placeholder.clone();
                $placeholder.appendTo($copyDiv);

                // Remove header and footer
                $placeholder.find(".pagination-row, .playsql-resize-handle, .right-padding > *, .zws-playsql").remove();
                $placeholder.find("tfoot").remove();
                // Remove all rows which don't have any cell selected
                $placeholder.find("tbody > tr").each(function(){
                    if ($(this).find("td.selected").size() == 0) {
                        $(this).remove();
                    }
                });
                // Remove all columns which don't have any cell selected
                $($placeholder.find("thead tr.playsql-col-headers th").get().reverse()).each(function(){
                    var $this = $(this),
                        index = $this.index(),
                        $cells = $placeholder.find("tbody tr > *:nth-child(" + (index+1) + ")");
                    if ($(this).data("status-pk") == 'true' || $(this).data("status-position") == 'true') return;
                    if ($this.hasClass(".right-padding") || !$cells.hasClass("selected")) {
                        // Well, remove...
                        $this.remove();
                        $cells.remove();
                    }
                });
                // Update the anchor on the top left
                $placeholder.find("thead > tr.playsql-col-headers > th:first a").text("...");
                $placeholder.find("td, th").removeClass("selected focused invalid cell-ref-active cell-ref-inactive").removeAttr("data-rendered");
                $placeholder.find("tr").each(deserializeCellMetadataToTD);
                var table = $copyDiv.find("table.playsql")[0];

                // Move all attributes from the div placeholder to the TABLE html element,
                // because divs are stripped (in Chrome)
                var attrs = $placeholder[0].attributes;
                for (var i in attrs) {
                    var attr = attrs[i];
                    if (attr.name && attr.name.indexOf("data-") == 0)
                        table.setAttribute(attr.name, attr.value);
                }

                var range = document.createRange(),
                    selection = window.getSelection();
                range.setStart(table, 0);
                range.setEnd(table, table.childNodes.length);

                selection.removeAllRanges();
                selection.addRange(range);
            }).on("input keyup", ".zws-playsql", function(event){
                var $contenteditable = $(this);
                if ($.trim($contenteditable.html()).length == 0 || $contenteditable.html() == "&#x200B;" || $contenteditable.html() == "\u200B") return;
                window.setTimeout(function(){
                    // Yuughuu!
                    if (!options.pasteHandler) return;
                    // Find spreadsheet, if any
                    var $focused = $(".playsql-spreadsheet-placeholder .focused"),
                        $placeholder = $focused.closest(".playsql-spreadsheet-placeholder");
                    if ($placeholder.length == 0) return;
                    var spreadsheet = Confluence.PlaySQL.Spreadsheet.byPlaceholder($placeholder);
                    if (!spreadsheet) return;
                    options.pasteHandler.call(spreadsheet, $contenteditable);
                }, 0);
            });

            function deserializeCellMetadataToTD() {
                var $tr = $(this),
                    $tds = $tr.children(),
                    $ths = $tr.closest("table").find("thead tr").first().children();
                var rowMetadataText = $tr.data("row-metadata");
                var $td, colName, cellMetadata;
                if (!rowMetadataText) return;
                try {
                    var rowMetadata = typeof rowMetadataText == "string" ? $.parseJSON(rowMetadataText) : rowMetadataText;
                    if (!rowMetadata) return;
                } catch (err) {
                    return;
                }
                for (var i = 0 ; i < $tds.size() ; i++ ) {
                    colName = $ths.eq(i).attr("data-name");
                    // We may not find the colName for the padding column
                    if (!colName) continue;
                    cellMetadata = rowMetadata["c_" + colName];
                    $td = $tds.eq(i);
                    if (cellMetadata) {
                        for (var key in cellMetadata) {
                            $td.attr("data-cell-metadata-" + key, cellMetadata[key]);
                        }
                    }
                }
            }
        },
        /** This method should be called after every click/selection. It selects
         * a zws in a contenteditable div, so the user can copy, paste and type and
         * we'll catch these events.
         * @param that the playsql-selection object
         **/
        setNonCollapsedSelection : function (that, handlePaste) {
            // Check whether the user has consciously selected something
            var $placeholder = that.options.$placeholder,
                selection = window.getSelection(),
                uniqueRange = selection.rangeCount == 1 ? selection.getRangeAt(0) : undefined;

            // Abort if there's an alternate selection mode (formula editing)
            if (that.options.cssClassMap) return;
            // Abort if there's a textual selection
            if (uniqueRange && !uniqueRange.collapsed) return;
            // Abort if the user has selected any cell
            if ($(".playsql").find(that.mapCss(".selected")).length == 0) return;

            // Abort if the user is editing anything...
            // Go figure: When the caret is in the zws-playsql, IE9 declares it in the thead
            var $caret = $placeholder.find("*:focus:not(table.playsql > thead), " + that.mapCss(".editing.focused"));
            if ($caret.length > 0 && $caret.html().length > 0 && !$caret.html() == "\u200B" && !$caret.is(".zws-playsql *")) return;

            // Alright, make them select a zws
            var $th = $placeholder.find("th").first(),
                $zwsDiv = $th.find(".zws-playsql"),
                initialScroll = $(window).scrollLeft();
            if ($zwsDiv.size() == 0) {
                $zwsDiv = $("<div></div>", {
                    "class" : "zws-playsql",
                    "contenteditable": handlePaste ? "true" : "false",
                    html: "&#x200B;" // That's a zero-width space
                }).appendTo($th);

                // We want to bypass the keyboard shortcuts if the user is typing into zws-playsql
                $(document).on("keypress", function(e){
                    if (e && e.target && e.target.contentEditable == "true" && e.target.className) {
                        if ((e.target.className || "").indexOf("zws-playsql") != -1) {
                            e.stopImmediatePropagation();
                        }
                    }
                });

                // There are more than 70 'keypress' handlers on a Confluence page, and we want to be first to skip
                // them all. Note: We hope we don't bypass some useful key binding, apart from the keyboard shortcuts.
                var allEvents = $._data(document, "events").keypress;
                // Move our handler to the first position
                allEvents.unshift(allEvents.pop());

            } else {
                $zwsDiv.html("&#x200B;");
            }
            $zwsDiv.focus();
            var range = document.createRange(),
                el = $zwsDiv[0],
                selection = window.getSelection();
            range.setStart(el, 0);
            range.setEnd(el, el.childNodes.length);

            try {
                selection.removeAllRanges();
            } catch (err) {
                // Err is 800a025e and happens on removeAllRanges in IE (9 at least)
                document.selection && document.selection.clear && document.selection.clear();
            }
            selection.addRange(range);

            // Restore the left scroll
            $(window).scrollLeft(initialScroll);
        }
    };

    function checkNotNull(value, message) {
        if (value == undefined) {
            throw "Unexpectedly undefined value: " + value + "(" + message + ")";
        }
        if (value instanceof $) {
            if (value.size() == 0)
                throw "Unexpectedly empty jquery element: " + value + "(" + message + ")";
        }
        return value;
    }

    var pressEnter = function(e) {
        // We must prevent default, otherwise the form is submitted.
        e.preventDefault();
        var options = this.options,
            $placeholder = options.$placeholder,
            $editingCells = $placeholder.find(".editing"),
            hasEditingCells = $editingCells.size() != 0;
        if (hasEditingCells) {
            options.saveCells($editingCells);
            Confluence.PlaySQL.Selection.setNonCollapsedSelection(this, !!this.options.pasteHandler);
        } else {
            // Abort if there's an alternate selection mode (formula editing)
            if (this.options.cssClassMap) return false;
            $placeholder.find(this.mapCss(".focused")).each(function(){
                options.editCell($(this));
            });
        }
        return false;
    };

    var pressEsc = function(e) {
        var options = this.options,
            $placeholder = options.$placeholder,
            $beingEdited = $placeholder.find(".editing, .saving"),
            hasEdits = $beingEdited.size() != 0;
        if (hasEdits) {
            $beingEdited.each(function(){
                options.cancelEdit($(this));
            });
            Confluence.PlaySQL.Selection.setNonCollapsedSelection(this, !!this.options.pasteHandler);
        } else {
            // If there are open inline-dialogs, close them
            var spreadsheet = PlaySQL.Spreadsheet.byPlaceholder($placeholder);
            if (spreadsheet.inlinedialog != null && spreadsheet.inlinedialog) {
                spreadsheet.inlinedialog.close();
            } else {
                // Otherwise, deselect all
                unselectAll.call(this);
            }
            // Formulas in edition: Close them
            Confluence.PlaySQL.Spreadsheet.FormulaEditor.closeDialog(spreadsheet, true);
        }
        return false;
    };

    function shouldCaptureKey(selection, $placeholder, e, handlerName) {
        var $cell = checkNotNull($placeholder.find(selection.mapCss(".focused"))),
            $widget = $cell.find("input:focus, select:focus, textarea:focus, div[contenteditable]:focus");
        if ($widget.size() == 1) {
            /* Think about:
             * Input (numbers)
             * Textarea (multiline text)
             * Datepicker
             * Page picker
             */
            // Generally, don't capture, except if pressing up, down, or if caret is at the beginning or end.
            if ($widget.is("input, textarea, div")) {
                if ((handlerName == 'moveDown' || handlerName == 'moveUp') && $widget.is("input")) {
                    if ($widget.parent().find(".autocomplete li:visible").size() != 0) {
                        // We're in a page picker with the drop-down open
                        return false;
                    }
                    // We're in an input box
                    return true;
                }
                // We're an input box (left, right) or a textarea (all directions)
                var caret = Confluence.PlaySQL.Helpers.caretSelection($widget[0]),
                    length = ($widget.val() || $widget.text()).length;
                if (caret == undefined) return false;
                if ((handlerName == 'moveLeft' || handlerName == 'moveUp') && caret.position == 0 && caret.length == 0) return true;
                if ((handlerName == 'moveRight' || handlerName == 'moveDown') && caret.position == length) return true;
            } else if ($widget.is("select")) {
                if (handlerName == 'moveLeft') return true;
                if (handlerName == 'moveRight') return true;
            }
            return false;
        }
        return true;
    }

    var moveDown = function(e) {
        var $placeholder = checkNotNull(this.options.$placeholder),
            $cell = checkNotNull($placeholder.find(this.mapCss(".focused"))),
            wasEditing = $cell.hasClass("editing"),
            $row = $cell.parent(),
            $newRow = $row.next(),
            index = $cell.index(),
            $newCell;

        // Dismiss if the user is editing a cell
        if (!shouldCaptureKey(this, $placeholder, e, "moveDown")) return;

        if ($row.parent().is("thead")) {
            $newRow = $placeholder.find("tbody > tr:first-child");
        } else if ($row.is("tbody > tr:last-child") && this.options.canSelectFooter) {
            $newRow = $placeholder.find("tfoot > tr.totals-row");
            // Don't select the <th> or the right padding
            index = Math.max(index, 1);
            index = Math.min(index, $newRow.children(":not(.right-padding)").size() - 1);
        }

        $newCell = $newRow.children().eq(index);
        if ($newRow.size() != 0) {
            if (index == 0) {
                // Select row
                rowClick.call(this, $newCell);
                unbindDrag.call(this);
                e.preventDefault();
            } else {
                if ($newCell.size() == 1 && !$newCell.hasClass("right-padding")) {
                    e.preventDefault();
                    // Select cell
                    unselectAll.call(this);
                    $newCell.addClass(this.mapCss("selected focused"));
                    if (wasEditing) {
                        this.options.saveCells($cell);
                        this.options.editCell && this.options.editCell($newCell);
                    } else {
                        Confluence.PlaySQL.Selection.setNonCollapsedSelection(this, !!this.options.pasteHandler);
                    }
                }
            }
        }
    }

    var moveUp = function(e) {
        var $placeholder = checkNotNull(this.options.$placeholder),
            $cell = checkNotNull($placeholder.find(this.mapCss(".focused"))),
            wasEditing = $cell.hasClass("editing"),
            $row = $cell.parent(),
            $newRow = $row.prev(),
            index = $cell.index(),
            $newCell;

        // Dismiss if the user is editing a cell
        if (!shouldCaptureKey(this, $placeholder, e, "moveUp")) return;

        if ($row.is("tbody > tr:first-child")) {
            var $header = $placeholder.find("thead > tr.playsql-col-headers").children().eq(Math.max(index, 1));
            columnClick.call(this, checkNotNull($header, index), e);
            e.preventDefault();
            return;
        } else if ($row.is("tfoot > tr")) {
            $newRow = $placeholder.find("tbody > tr:last-child");
        }
        $newCell = $newRow.children().eq(index);

        if ($newRow.size() != 0) {
            if (index == 0) {
                // Select row
                rowClick.call(this, $newCell);
                unbindDrag.call(this);
                e.preventDefault();
                return;
            } else {
                e.preventDefault();
                // Just select a cell
                unselectAll.call(this);
                $newCell.addClass(this.mapCss("selected focused"));
                if (wasEditing) {
                    this.options.saveCells($cell);
                    this.options.editCell && this.options.editCell($newCell);
                } else {
                    Confluence.PlaySQL.Selection.setNonCollapsedSelection(this, !!this.options.pasteHandler);
                }
                return;
            }
        }
    }

    var moveLeft = function(e) {
        var $placeholder = checkNotNull(this.options.$placeholder),
            $cell = checkNotNull($placeholder.find(this.mapCss(".focused"))),
            wasEditing = $cell.hasClass("editing"),
            $newCell = $cell.prev();

        // Dismiss if the user is editing a cell
        if (!shouldCaptureKey(this, $placeholder, e, "moveLeft")) return;

        if ($newCell.is("thead > tr.playsql-col-headers > th")) {
            if ($newCell.index() == 0) {
                rowClick.call(this, checkNotNull($placeholder.find("tbody > tr > th").first()));
                unbindDrag.call(this);
                e.preventDefault();
            } else {
                columnClick.call(this, $newCell, e);
                e.preventDefault();
            }
        } else if ($newCell.is("tbody th")) {
            rowClick.call(this, $newCell);
            unbindDrag.call(this);
            e.preventDefault();
        } else if ($newCell.is("td")) {
            e.preventDefault();
            unselectAll.call(this);
            $newCell.addClass(this.mapCss("selected focused"));
            if (wasEditing) {
                this.options.saveCells($cell);
                this.options.editCell && this.options.editCell($newCell);
            } else {
                Confluence.PlaySQL.Selection.setNonCollapsedSelection(this, !!this.options.pasteHandler);
            }
        }
    }

    var moveRight = function(e) {
        var $placeholder = checkNotNull(this.options.$placeholder),
            $cell = checkNotNull($placeholder.find(this.mapCss(".focused"))),
            wasEditing = $cell.hasClass("editing"),
            $newCell = $cell.next(":not(.right-padding,.hidden,.nohighlight)");

        // Dismiss if the user is editing a cell
        if (!shouldCaptureKey(this, $placeholder, e, "moveRight")) return;

        if ($cell.is("thead > tr > th")) {
            if ($newCell.size() != 0)
                columnClick.call(this, $newCell, e);
        } else if ($newCell.size() != 0) {
            unselectAll.call(this);
            $newCell.addClass(this.mapCss("selected focused"));
            e.preventDefault();
            if (wasEditing) {
                this.options.saveCells($cell);
                this.options.editCell && this.options.editCell($newCell);
            } else {
                Confluence.PlaySQL.Selection.setNonCollapsedSelection(this, !!this.options.pasteHandler);
            }
        }
    }

    var nextTabOrder = function(e) {
        moveTabOrder.call(this, e, +1);
    }
    var previousTabOrder = function(e) {
        moveTabOrder.call(this, e, -1);
    }

    var moveTabOrder = function(e, increment) {
        function moveFocus($oldCell, $newCell) {
            $oldCell.removeClass(this.mapCss("focused"));
            $newCell.addClass(this.mapCss("focused"));

            if (!$newCell.hasClass(this.mapCss("selected"))) {
                // Moving inside a selection
                $oldCell.removeClass(this.mapCss("selected"));
                $newCell.addClass(this.mapCss("selected"));
            }

            if ($oldCell.hasClass("editing")) {
                // Save the previous cell
                this.options.saveCells($oldCell);
                // Make the cell editable
                this.options.editCell && this.options.editCell($newCell);
            }
        }
        var $placeholder = checkNotNull(this.options.$placeholder),
            $focused = $placeholder.find(this.mapCss(".focused")),
            $selected = $placeholder.find(this.mapCss(".selected")),
            $inputsInCell = $focused.find("input, div[contenteditable], textarea"),
            $focusedInput = $focused.find(":focus"),
            index = $selected.index($focused),
            newIndex = (index + increment) % $selected.size();

        if ($inputsInCell.size() > 1 && $focusedInput.size() == 1) {
            if (increment == 1 && $inputsInCell.last()[0] != $focusedInput[0]
                || increment == -1 && $inputsInCell.first()[0] != $focusedInput[0]) {
                // Let the natural tab operate
                AJS.log("Natural tab");
                return true;
            } else {
                // We're in the last/first input field of a cell. Normally we'd change cell,
                // except if we're editing a total
                if ($focusedInput.closest(".totals-row").size() > 0) {
                    AJS.log("Natural tab: Prevented");
                    e.preventDefault();
                    return false;
                }
            }
        }
        if ($focused.is("thead > tr.playsql-col-headers > th")) {
            // Just submit the column
            $focusedInput.blur();
            return false;
        }

        if (index != -1 && $selected.size() > 1) {
            moveFocus.call(this, $focused, $selected.eq(newIndex));
        } else if ($focused.size() != 0) {
            // Next cell in the row?
            var $newCell = increment > 0 ? $focused.next(":not(.right-padding)") : $focused.prev(":not(:first-child)");
            if ($newCell.size() != 1) {
                // Next row?
                $newCell = increment > 0
                    ? checkNotNull($focused.parent("tr")).next("tr").children("td").first()
                    : checkNotNull($focused.parent("tr")).prev("tr").children("td:not(.right-padding)").last();
            }
            if ($newCell.size() != 1) {
                // Top-left cell? Move around the spreadsheet
                if (increment < 0) {
                    $newCell = checkNotNull($placeholder.find("tbody td:not(.right-padding)").last());
                } else {
                    // Bottom-right cell? If edit -> add row, Otherwise go to top-left cell.
                    if ($focused.hasClass(this.mapCss("editing"))) {
                        this.options.addRow && this.options.addRow();
                    } else {
                        $newCell = checkNotNull($placeholder.find("tbody td").first());
                    }
                }
            }
            if ($newCell.size() == 1) {
                moveFocus.call(this, $focused, $newCell);
            }
        }

        e.preventDefault();
    }

    var selectAll = function() {
        var $cell = checkNotNull(this.options.$placeholder.find(this.mapCss(".focused") + ", .focused")),
            $widget = $cell.find("input:focus, select:focus, textarea:focus, div[contenteditable]:focus");

        // Abort if there's an alternate selection mode (formula editing)
        if (this.options.cssClassMap) return;
        // Abort if there is any cell in edition
        if ($widget.size() == 1) return;

        var $allCells = this.options.$placeholder.find("tbody tr > th, tbody tr > td:not(.right-padding)");
        $allCells.addClass(this.mapCss("selected")).removeClass(this.mapCss("focused"));
        $allCells.first().addClass(this.mapCss("focused"));
        Confluence.PlaySQL.Selection.setNonCollapsedSelection(this, !!this.options.pasteHandler);
    }

    /**
     * Unselect all cells and remove menus
     * @param keepSelectedCells true if we should just remove the focus and the menus, and keep the selected cells
     */
    var unselectAll = function(keepSelectedCells) {
        // It wasn't possible to have 2 spreadsheets on the same page, so I've changed from
        // unselecting within the $placeholder and unselecting on the whole page.
        $(".playsql-spreadsheet-placeholder,.playsql-query-container").find(this.mapCss(".selected,.focused"))
            .removeClass((keepSelectedCells ? "" : this.mapCss("selected")) + " " + this.mapCss("focused") + " aui-dropdown2-trigger")
            .removeAttr('aria-controls aria-haspopup aria-owns');
    }

    var columnClick = function($header, e) {
        var $placeholder = this.options.$placeholder,
            column = this.options.colOf($header);
        if (!column) return false;
        var index = column.index + 1;
        $cells = $placeholder.find("table.playsql > thead > tr.playsql-col-headers > *:nth-child(" + index + ")," +
        "table.playsql > tbody > tr:not(.playsql-message) > *:nth-child(" + index + ")");
        if ($cells.first().hasClass(this.mapCss("selected"))) {
            if (!$header.hasClass("editing")) {
                this.options.columnSecondClick && this.options.columnSecondClick($header);
            }
        } else if (!$cells.first().hasClass(this.mapCss("focused"))) {
            unselectAll.call(this);
            $cells.addClass(this.mapCss("selected"));
            $cells.first().addClass(this.mapCss("focused"));
            Confluence.PlaySQL.Selection.setNonCollapsedSelection(this, !!this.options.pasteHandler);
        }
        return false;
    };

    /** Select row */
    var rowClick = function($header) {
        var that = this;
        if (event.type != "keydown" && (event.which || event.button) != 1) return;
        if ($header.hasClass(this.mapCss("selected"))) {
            // Abort if there's an alternate selection mode (formula editing)
            if (this.options.cssClassMap) return;
            this.options.rowSecondClick && this.options.rowSecondClick($header);
        } else {
            if ($header.is("tfoot th")) return true;

            unselectAll.call(this);
            var $cells = $header.parent().children("th,td"),
                $placeholder = this.options.$placeholder;
            $cells.addClass(this.mapCss("selected"));
            $cells.first().addClass(this.mapCss("focused"));
            var selectDragRow = function($tr) {
                if (!$placeholder.has($tr))
                    return;
                var $allTrs = $header.closest("table").find("tbody > tr"),
                    rowId1 = $tr.attr("data-row-id"),
                    rowId2 = checkNotNull($header.parent().attr("data-row-id")),
                    highlighted = false;
                if (rowId1 == undefined)
                // We're hovering over a tr which isn't in the body
                    return;
                $allTrs.each(function(){
                    var $tr = $(this),
                        rowId = $tr.attr("data-row-id");
                    if (rowId == rowId1 || rowId == rowId2) {
                        highlighted = !highlighted;
                        $tr.find("td,th").addClass(that.mapCss("selected"));
                    } else if (highlighted) {
                        $tr.find("td,th").addClass(that.mapCss("selected"));
                    } else {
                        $tr.find("td,th").removeClass(that.mapCss("selected"));
                    }
                });
            };
            Confluence.PlaySQL.Selection.attach($(window.document), this, {
                "mouseover.select-drag tr" : selectDragRow,
                "mouseup.select-drag" : unbindDrag
            });
            $placeholder.addClass("no-select");
            Confluence.PlaySQL.Selection.setNonCollapsedSelection(this, !!this.options.pasteHandler);
        }
    };

    // I've introduced mouseupTriggered because "unbindDrag" must be attached as soon as mousedown is here in order
    // not to miss the mouseup, but it can't be treated until after the setTimeout(0) of cellClick is done.
    var mouseupTriggered = false;
    var unbindDrag = function() {
        mouseupTriggered = true;
        $(window.document)
            .unbind("mouseover.select-drag")
            .unbind("mouseup.select-drag");
        if (this.options.$placeholder.hasClass("no-select")) {
            this.options.$placeholder.removeClass("no-select");
            Confluence.PlaySQL.Selection.setNonCollapsedSelection(this, !!this.options.pasteHandler);
        }
    }

    var cellClick = function($cell, event) {
        if (event.type != "keydown" && (event.which || event.button) != 1) return;
        // We had to bind 'cellClick' (and all attachments) to 'mousedown' instead of 'click', because we wanted
        // to capture the drag event (mousedown/mouseup lapse).
        //
        // It caused cellClick to behave differently, because the 'click' would happen after mousedown, disturbing
        // the focus. Therefore we introduced window.setTimeout().
        //
        // Then we noticed events happened in this order in IE9: mousedown, mouseup, mousedown.timeout(0). So it would
        // appear like the cursor was dragging a sticky selection.
        // We solved that by checking the current state of the mouse button.
        var that = this;
        mouseupTriggered = false;
        Confluence.PlaySQL.Selection.attach($(window.document), that, {
            "mouseup.select-drag" : unbindDrag
        });
        window.setTimeout(function(){
            // Abort if the click is in the footer and it's not allowed to select it
            if ($cell.is("tfoot > tr > *") && !that.options.canSelectFooter) {
                return;
            }
            if ($cell.parent().hasClass("pagination-row")) {
                return;
            }
            if ($cell.hasClass("right-padding")) {
                return;
            }
            if ($cell.hasClass(that.mapCss("focused"))) {
                // Abort if there's an alternate selection mode (formula editing)
                if (that.options.cssClassMap) return;
                that.options.cellSecondClick && that.options.cellSecondClick.call(that, $cell);
                return;
            }
            // Special case to edit totals
            if ($cell.hasClass("edit-on-first-click")) {
                // Abort if there's an alternate selection mode (formula editing)
                if (that.options.cssClassMap) return;
                that.options.cellSecondClick && that.options.cellSecondClick.call(that, $cell);
                return;
            }
            if (!event || !(!Confluence.PlaySQL.Helpers.isMac && event.ctrlKey || Confluence.PlaySQL.Helpers.isMac && event.metaKey))
                unselectAll.call(that);
            else
                unselectAll.call(that, true);
            $cell.addClass(that.mapCss("selected focused"));

            if ($cell.hasClass(that.mapCss("editing"))) {
                return;
            }
            var $placeholder = that.options.$placeholder;
            var selectDragCell = function($cell) {
                if (!$placeholder.has($cell)) return;
                var $focused = $placeholder.find(that.mapCss(".focused")),
                    origCol = $focused.index(),
                    origRow = $focused.closest("tr").index(),
                    destCol = $cell.index(),
                    destRow = $cell.closest("tr").index(),
                    minCol = Math.min(origCol, destCol),
                    maxCol = Math.max(origCol, destCol),
                    minRow = Math.min(origRow, destRow),
                    maxRow = Math.max(origRow, destRow);
                if (!$cell.is($focused)) {
                    // We let the user select the text in one cell. If he moves over another
                    // cell, we disable selection.
                    $placeholder.addClass("no-select");
                }
                // Select all cells between that focused cell and $cell
                $placeholder.find("tbody tr > td").each(function(){
                    var $this = $(this),
                        col = $this.index(),
                        row = $this.closest("tr").index();
                    if (   col >= minCol && col <= maxCol
                        && row >= minRow && row <= maxRow) {
                        $this.addClass(that.mapCss("selected"));
                    } else {
                        $this.removeClass(that.mapCss("selected"));
                    }
                });
            };

            if (mouseupTriggered) {
                // Since the mouseup was already triggered, no need to attach the handler
            } else {
                Confluence.PlaySQL.Selection.attach($(window.document), that, {
                    "mouseover.select-drag td" : selectDragCell
                });
            }

            // Yeg, select something so the 'copy' event can be raised
            Confluence.PlaySQL.Selection.setNonCollapsedSelection(that, !!that.options.pasteHandler);
        }, 0);
    }

    var resizing = false,
        isHovering,
        minPageY, maxPageY;
    var headerMouseEnter = function($cell, e) {
        isHovering = true;
        if (resizing) return;
        var offset = $cell.offset(),
            width = $cell.outerWidth(),
            height = $cell.outerHeight(),
            left = offset.left + width - 10,
            top = offset.top,
            $placeholder = $cell.closest(".playsql-spreadsheet-placeholder"),
            $handle = $placeholder.find(".playsql-resize-handle");

        // Find the closest positioned element, that's the one which "position:relative" is relative to
        var $container = $cell;
        while ($container.css("position")=="static" && $container.parent().size() > 0) {
            $container = $container.parent();
        }
        var containerTop = $container.offset().top,
            containerLeft = $container.offset().left;

        // Determine the upper and lower allowed bounds of the cursor
        minPageY = top - 20;
        maxPageY = top + height + 20;

        $handle.remove();
        $handle = $("<div>", { "class": "playsql-resize-handle" });
        $cell.append($handle);
        $handle.css({
            top: (top - containerTop) + "px",
            left: (left - containerLeft) + "px",
            height: height + "px"
        }).removeClass("hidden");
    }
    var headerMouseLeave = function($cell, e) {
        if (!resizing) {
            $(".playsql-resize-handle").addClass("hidden");
        }
    }
    var resizeHandleMousedown = function($handle, e) {
        resizing = true;
        var pageX = e.pageX,
            $cell = $handle.closest("th"),
            originalWidth = $cell.width(),
        //$handle = $(".playsql-resize-handle"),
            originalLeft = parseInt($handle.css("left")),
            playsqlSelection = this;
        $handle.addClass("dragging");

        $(document).on('selectstart.playsql-resize', function (e) {
            e.preventDefault();
        });

        $(document).on("mousemove.playsql-resize", function(e) {
            var newPageY = e.pageY;
            isHovering = newPageY > minPageY && newPageY < maxPageY;
            if (!isHovering) {
                $handle.addClass("hidden");
                $cell.css("width", "");
                return;
            }
            var newPageX = e.pageX,
                newWidth = originalWidth - pageX + newPageX,
                newLeft = originalLeft - pageX + newPageX;
            $cell.width(newWidth);
            $handle.css("left", newLeft + "px").removeClass("hidden");
        });

        $(document).on("mouseup.playsql-resize", function(e) {
            e.preventDefault();
            $(document).off(".playsql-resize");
            $handle.removeClass("dragging");
            resizing = false;
            var width = isHovering ? $cell.width() : null;
            if (Math.abs(originalWidth - width) < 3) {
                // It's a header click. We let the fw decide whether it's a first or second click.
                $cell.click();
            } else {
                playsqlSelection.options.updateColumnWidth && playsqlSelection.options
                    .updateColumnWidth.call(playsqlSelection, $cell, width);
            }
        });
    };
    var dodgeClick = function($handle, e) {
        // Prevent the resize handle from feeling selected
        e.preventDefault();
        return false;
    }

    /** Depending on selection.options.cssClassMap, returns the equivalent CSS classes.
     * E.g.: mapClass("selected") returns "selected" by default.
     * When selection.options.cssClassMap == {"selected": "cell-ref-active"}, it returns "cell-ref-active"
     * @param cssClassName the name of the class. Several classes can be passed with a comma.
     * @return the mapped class names separated by a space
     */
    function mapCss(cssClassName) {
        if (!this.options.cssClassMap) return cssClassName;
        function isSpecialChar(a) {
            switch (a) {
                case '':
                case ' ':
                case ',':
                case '.':
                case '[':
                case ']':
                    return true;
                default:
                    return false;
            }
        }
        var parts = [""],
            state = false;
        // It loops on length+1 characters, because the last (virtual) one is considered
        // as a separator and it ensures the last class is transformed too.
        for (var i = 0 ; i <= cssClassName.length ; i++) {
            if (isSpecialChar(cssClassName.charAt(i))) {
                if (state) {
                    // The last string on "parts" is a CSS classname
                    var val = this.options.cssClassMap[parts[parts.length - 1]];
                    if (val) {
                        parts[parts.length - 1] = val;
                    }
                }
                state = false;
            } else {
                if (!state) parts.push("");
                state = true;
            }
            parts[parts.length - 1] += cssClassName.charAt(i);
        }
        var result = parts.join("");
        return result;
    }

})(AJS.$);