(function($){
    Confluence.PlaySQL.QueryEditor = Confluence.PlaySQL.QueryEditor || {};
    var wizards = Confluence.PlaySQL.QueryEditor.Wizards = [],
        ed = Confluence.PlaySQL.QueryEditor;

    /** Adds double-quotes to escape a table name or column name, if necessary*/
    function withQuotes(name) {
        if (name == undefined || name == "") return name;
        var dbCase = Confluence.PlaySQL.QueryEditor.dbCase;
        if (dbCase == undefined) {
            dbCase = AJS.Meta.get("datasource-case") || "";
            Confluence.PlaySQL.QueryEditor.dbCase = dbCase;
        }
        if (dbCase == "LOWER") {
            if (name == name.toLowerCase()) return name;
        } else if (dbCase == "UPPER") {
            if (name == name.toUpperCase()) return name;
        }
        return '"' + name + '"';
    }
    function escapeSqlString(s) {
        return s.replace(/'/g, "''").replace(/[\r]?\n[\r]?/g, "\\n");
    }

    wizards.push({
        name: "fromSpreadsheet",
        label: "Query from a spreadsheet",
        run : function() {
            if (wizards.fromSpreadsheetDialog) {
                wizards.fromSpreadsheetDialog.show();
                wizards.fromSpreadsheetDialog.gotoPage(0);
                return;
            }

            var spreadsheetList = [];
            for (var i in ed.databaseSchema.tableNames) {
                var table = ed.databaseSchema.tableNames[i],
                    tableDef = ed.databaseSchema.getTableDef(table);
                spreadsheetList.push({
                    name: table,
                    label: tableDef && tableDef.label || table
                });
            }

            var dialog,
                $template = $(Confluence.Templates.PlaySQL.wizardChooseSpreadsheet({
                    spreadsheets: spreadsheetList
                })),

                cancelCallback = function () {
                    dialog.hide();
                },
                gotoPage2Callback = function() {
                    var $panel = dialog.getCurrentPanel().body,
                        spreadsheetName = $panel.find("#table-name").val(),
                        $errorField = $panel.find(".error"),
                    //buttons = dialog.getCurrentPanel().page.buttonpanel;
                        $buttonNext = $(this),
                        $page2 = dialog.getPage(1).body;
                    if (spreadsheetName == undefined || spreadsheetName == "") {
                        $errorField.text("Please choose a spreadsheet");
                        return;
                    }
                    if (dialog.tableDef && dialog.tableDef.name == spreadsheetName) {
                        dialog.nextPage();
                        return;
                    }

                    ed.triggerSchemaUpdate(spreadsheetName, function(data) {
                        var tableDef = data.getTableDef(spreadsheetName),
                            $page2 = dialog.getPage(1).body,
                            $page3 = dialog.getPage(2).body,
                            $page4 = dialog.getPage(3).body;
                        dialog.tableDef = tableDef;
                        $page2.empty().html(Confluence.Templates.PlaySQL.wizardChooseSpreadsheetColumns({
                            tableDef: tableDef
                        }));
                        $page3.empty().html(Confluence.Templates.PlaySQL.wizardWhereClause({
                            tableDef: tableDef
                        }));
                        $page4.empty().html(Confluence.Templates.PlaySQL.wizardSortClause({
                            tableDef: tableDef
                        }));
                        dialog.getPage(1).body.height(dialog.getPage(0).body.height());
                        dialog.getPage(2).body.height(dialog.getPage(0).body.height());
                        dialog.getPage(3).body.height(dialog.getPage(0).body.height());
                        dialog.nextPage();
                    }, function() {
                        alert("Couldn't fetch database details for " + spreadsheetName);
                    });
                },
                submitCallback = function () {
                    $(this).disable().text("Submitting...");
                    var tablename = dialog.tableDef.name,
                        $allFields = dialog.getPage(1).body
                            .find("#wizard-choose-columns input[type='checkbox']"),
                        fields = $allFields.find(":not(:checked)").size() == 0
                            ? [{def:"*"}]
                            : dialog.getPage(1).body
                                .find("#wizard-choose-columns input[type='checkbox']:checked")
                                .map(function(){
                                    var name = $(this).data("colname"),
                                        label = $(this).text();
                                    if (label != name && !!label) {
                                        return {def:withQuotes(name) + " \"" + label + "\""};
                                    } else {
                                        return {def:withQuotes(name)};
                                    }
                                }).toArray(),
                        filter = dialog.getPage(2).body.find("table > tbody > tr").map(function(i) {
                            var $tr = $(this),
                                combination = i == 0 ? "" : $tr.find(".filter-combination").val(),
                                col = $tr.find(".filter-column").val(),
                                colType = $tr.find(".filter-column option[value='" + col + "']").data("type"),
                                op = $tr.find(".filter-operator.visible").val(),
                                value = $tr.find("input.filter-value").val(),
                                isNullValue = $tr.find("select.filter-is-null").val();
                            if (col == "") return undefined;
                            col = withQuotes(col);
                            if (op == "IS-NULL") {
                                op = "";
                                value = {"IS":"IS NULL", "NOT":"IS NOT NULL"}[isNullValue];
                            } else if (colType == 'VARCHAR') {
                                if (op == 'LIKE') {
                                    value = "UPPER('%" + escapeSqlString(value) + "%')";
                                    col = "UPPER(" + col + ")";
                                } else {
                                    value = "'" + escapeSqlString(value) + "'";
                                }
                            }
                            var toJoin = [combination, col, op, value];
                            for (var i in toJoin) {
                                if (toJoin[i] == undefined || toJoin[i] == "")
                                    toJoin.splice(i, 1);
                            }
                            return toJoin.join(" ");
                        }).toArray(),
                        sort = dialog.getPage(3).body.find("table > tbody > tr").map(function(i) {
                            var $tr = $(this),
                                col = $tr.find(".sort-column").val(),
                                asc = $tr.find(".sort-asc").val(),
                                nulls = $tr.find(".sort-nulls").val(),
                                result;
                            if (col == "") return undefined;
                            result = withQuotes(col);
                            if (asc) result += " " + asc;
                            if (nulls) result += " " + nulls;
                            return result;
                        }).toArray();

                    for (var i in filter) {
                        if (filter[i] == undefined || filter[i] == "")
                            filter.splice(i, 1);
                    }
                    for (var i in sort) {
                        if (sort[i] == undefined || sort[i] == "")
                            sort.splice(i, 1);
                    }

                    var query = [
                        ["SELECT"].concat(fields),
                        ["FROM", {def:withQuotes(tablename)} ],
                        filter.length > 0 ? ["WHERE"].concat(filter.join("\n    ")) : undefined,
                        sort.length > 0 ? ["ORDER BY"].concat(sort.join(", ")) : undefined
                    ];
                    var sql = Confluence.PlaySQL.Helpers.rebuildSQL(query);
                    ed.aceEditor.setValue(sql);

                    $(this).enable().text("Submit");
                    wizards.fromSpreadsheetDialog = undefined;
                    dialog.hide().remove();
                    Confluence.PlaySQL.QueryEditor.preview();
                };

            dialog = new AJS.ConfluenceDialog({
                width: 600,
                height: 450,
                id: 'wizard-select-dialog',
                closeOnOutsideClick: true,
                onCancel: cancelCallback,
                onSubmit: submitCallback
            });
            wizards.fromSpreadsheetDialog = dialog;

            dialog.addHeader("Create query from spreadsheet (1/4)");
            dialog.addPanel("Panel 1", $template);
            dialog.addButton("Next", gotoPage2Callback);
            dialog.addCancel("Cancel", cancelCallback);

            dialog.addPage("page-2");
            dialog.addHeader("Create query from spreadsheet (2/4)");
            dialog.addPanel("Panel 2", "<span/>");
            dialog.addButton("Previous", function() {
                dialog.prevPage();
            });
            dialog.addButton("Next", function() {
                dialog.nextPage();
            });
            dialog.addCancel("Cancel", cancelCallback);

            dialog.addPage("page-3");
            dialog.addHeader("Add filters (3/4)");
            dialog.addPanel("Panel 3", "<span/>");
            dialog.addButton("Previous", function() {
                dialog.prevPage();
            });
            dialog.addButton("Next", function() {
                dialog.page[2].body.find(".filter-operator").each(function(){
                    var $this = $(this);
                    $this.toggleClass("visible", $this.is(":visible"))
                });
                dialog.nextPage();
            });
            dialog.addCancel("Cancel", cancelCallback);
            dialog.page[2].body.on("change", "select", function(){
                var $select = $(this),
                    $tr = $select.closest("tr"),
                    type = $tr.find(".filter-column option:selected").attr("data-type"),
                    op = $tr.find(".filter-operator:visible option:selected").attr("value");
                $tr.attr("class", "type-" + type + " op-" + op);
            });


            dialog.addPage("page-4");
            dialog.addHeader("Sort the results (4/4)");
            dialog.addPanel("Panel 4", "<span/>");
            dialog.addButton("Previous", function() {
                dialog.prevPage();
            });

            // Last page
            dialog.addSubmit("Save", submitCallback);
            dialog.addCancel("Cancel", cancelCallback);

            dialog.gotoPage(0);
            dialog.gotoPanel(0);

            Confluence.PlaySQL.QueryEditor.databaseSchema.onUpdate.push(function(data) {
                var $select = dialog.getPage(0).body.find("#table-name"),
                    $option,
                    table;
                for (var i in data.tables["current schema"]) {
                    table = data.tables["current schema"][i];
                    $option = $select.find("option[value='" + table.name + "']");
                    if ($option.length == 0) {
                        $select.append($("<option></option>", {
                            text: table.label || table.name,
                            value: table.name
                        }));
                    } else {
                        $option.text(table.label || table.name);
                    }
                }
            });
            $("#wizard-select-dialog").on("click", "table .remove, table .add", function() {
                var $this = $(this),
                    remove = $this.hasClass("remove"),
                    $tr = $this.closest("tr");
                if (remove) {
                    $tr.remove();
                } else {
                    $tr.clone().insertAfter($tr);
                }
            });

            dialog.show();
        }
    });

    /** Wizard for subtracting two tables */
    wizards.push({
        name: "subtractTwoTables",
        label: "Subtract two tables",
        run: function() {
            if (wizards.subtractTwoTablesDialog) {
                wizards.subtractTwoTablesDialog.show();
                wizards.subtractTwoTablesDialog.gotoPage(0);
                return;
            }

            var spreadsheetList = [];
            for (var i in ed.databaseSchema.tableNames) {
                var table = ed.databaseSchema.tableNames[i],
                    tableDef = ed.databaseSchema.getTableDef(table);
                spreadsheetList.push({
                    name: table,
                    label: tableDef && tableDef.label || table
                });
            }

            var dialog,
                $template = $(Confluence.Templates.PlaySQL.wizardSubtractChooseSpreadsheet({
                    spreadsheets: spreadsheetList
                })),

                cancelCallback = function () {
                    dialog.hide();
                },
                submitCallback = function () {
                    $(this).disable().text("Submitting...");
                    var originalSql = ed.aceEditor.getValue(),
                        allMatchedCols = $("#wizard-subtract-dialog thead th.matched").map(function() { return $(this).data("colname"); }).toArray(),
                        matchingCols = $("#wizard-subtract-dialog tbody td.matched .match-column").map(function() { return $(this).val(); }).toArray(),
                        caseInsensitivity = $("#wizard-subtract-dialog tbody td.matched .case-sensitivity").map(function() { return $(this).is(":checked"); }).toArray(),
                        table1Name = dialog.table1.name,
                        table2Name = dialog.table2.name,
                        /** mainTableName must be double-quote escaped */ mainTableName,
                        parsedQuery,
                        existQuery, queryNotParsed,
                        whereClause,
                        sql;

                    if (table1Name != '') {
                        originalSql = "SELECT * FROM " + withQuotes(table1Name);
                        mainTableName = withQuotes(table1Name);
                    }
                    parsedQuery = Confluence.PlaySQL.Helpers.parseSQL(originalSql);
                    if (!parsedQuery.bilinear) {
                        queryNotParsed = "SELECT * FROM (" + originalSql + ") SUBQUERY";
                        mainTableName = "SUBQUERY";
                    } else {
                        //!\ whereClause is passed by ref, so we're modifying the query
                        whereClause = parsedQuery.clause("WHERE");
                        var firstTable = parsedQuery.clause("FROM")[1];
                        mainTableName = firstTable.alias;
                        if (!mainTableName) {
                            mainTableName = firstTable.def.split(" ");
                            mainTableName = mainTableName[mainTableName.length-1];
                        }
                        if (mainTableName == table2Name) {
                            mainTableName = "MAIN_TABLE";
                            firstTable.alias = mainTableName;
                        }
                    }

                    existQuery = "NOT EXISTS(SELECT * FROM " + withQuotes(table2Name) + " WHERE ";
                    for (var i in matchingCols) {
                        if (i > 0) existQuery += "\n\tAND ";
                        if (caseInsensitivity[i]) {
                            existQuery += "UPPER($table2.$col2) = UPPER($table1.$col1)";
                        } else {
                            existQuery += "$table2.$col2 = $table1.$col1";
                        }

                        existQuery = existQuery
                            .replace(/\$table1/, mainTableName)
                            .replace(/\$col1/, withQuotes(allMatchedCols[i]))
                            .replace(/\$table2/, withQuotes(table2Name))
                            .replace(/\$col2/, withQuotes(matchingCols[i]));
                    }
                    existQuery += ")";

                    if (queryNotParsed != undefined) {
                        // Working on an unparsed query
                        sql = queryNotParsed + "\nWHERE " + existQuery;
                    } else {
                        // Working on a parsed query
                        if (whereClause != undefined) {
                            whereClause.splice(1, 0, "(");
                            whereClause.push(") AND ");
                        } else {
                            whereClause = ["WHERE"];
                        }
                        whereClause.push(existQuery);
                        parsedQuery.clause("WHERE", whereClause);

                        sql = Confluence.PlaySQL.Helpers.rebuildSQL(parsedQuery);
                    }

                    ed.aceEditor.setValue(sql);
                    wizards.subtractTwoTablesDialog = undefined;

                    $(this).enable().text("Submit");
                    dialog.hide().remove();
                };

            dialog = new AJS.ConfluenceDialog({
                width: 600,
                height: 450,
                id: 'wizard-subtract-dialog',
                closeOnOutsideClick: true,
                onCancel: cancelCallback,
                onSubmit: submitCallback
            });
            wizards.subtractTwoTablesDialog = dialog;

            dialog.addHeader("Subtract two tables");
            dialog.addPanel("Panel 1", $template);
            dialog.addButton("Next", function() {
                var $this = $(this),
                    table1 = $("#wizard-subtract-choose-spreadsheet #table-name-main").val(),
                    table2 = $("#wizard-subtract-choose-spreadsheet #table-name-subtract").val(),
                    table1Name,
                    table2Name,
                    columns1,
                    columns2;

                if (table2 == undefined || table2 == "") {
                    $("#wizard-subtract-choose-spreadsheet .error").text("Please choose a table");
                } else {
                    $("#wizard-subtract-choose-spreadsheet .error").text("");
                }
                $this.disable().text("Fetching columns...");

                function callback() {
                    if (table1Name == undefined || table2Name == undefined) return;
                    var $page2 = dialog.getPage(1).body;
                    $page2.empty().html(Confluence.Templates.PlaySQL.wizardSubtractMatchingColumns({
                        columns1: columns1,
                        columns2: columns2,
                        table1Name: table1Name,
                        table2Name: table2Name
                    }));
                    dialog.getPage(1).body.height(dialog.getPage(0).body.height());

                    // Can we preselect options?
                    $page2.find("thead tr > th").each(function(){
                        var $this = $(this),
                            colname = $this.data("colname"),
                            $option = $page2.find("tbody > tr > td[headers='header-$colname'] option[value='$colname']".replace(/\$colname/g, colname)),
                            $tickMatch = $page2.find("tbody > tr > td[headers='header-$colname'] input.tickmatching".replace(/\$colname/g, colname));
                        if ($option.length == 1) {
                            // There's an column in the matching table with the exact column name. Preselect it.
                            $option.parent().val(colname);
                            $tickMatch.attr("checked", "on");
                            // Just match the first one.
                            return false;
                        }
                    });
                    $page2.find("tbody tr > td > .tickmatching").each(updateMatchedColumns);

                    $this.enable().text("Next");
                    dialog.nextPage();
                }

                ed.triggerSchemaUpdate(table2, function(data) {
                    var tableDef = data.getTableDef(table2);
                    if (tableDef == undefined) {
                        dialog.hide().remove();
                        alert("Couldn't retrieve information about the table " + table2 + ". We won't be able to help you for this query.");
                        return;
                    }
                    table2Name = tableDef.label || tableDef.name;
                    columns2 = tableDef.columns;
                    dialog.table2 = tableDef;
                    callback();
                }, function() {
                    alert("Could not fetch database details for " + table2);
                    $this.enable().text("Next");
                });

                if (table1 == "") {
                    // Take the columns from the query
                    var $previewTable = $(".playsql-query-container table.playsql"),
                        isBroken = $previewTable.hasClass("unavailable"),
                        columns = $previewTable.find("> thead > tr > th").map(function(){
                            var name = $(this).attr("title"),
                                label = $.trim($(this).text().replace(/\s+/, " ")) || name;
                            return {
                                name: name,
                                label: label
                            };
                        }).toArray();
                    if (isBroken || columns.length < 2) {
                        alert("If you choose to join records from the current query, please preview the query first.");
                        dialog.hide();
                    }
                    table1Name = "Current query";
                    columns1 = columns;
                    dialog.table1 = {
                        name: "",
                        label: table1Name,
                        columns: columns
                    };
                    callback();
                } else {
                    ed.triggerSchemaUpdate(table1, function(data) {
                        var tableDef = data.getTableDef(table1);
                        if (tableDef == undefined) {
                            dialog.hide().remove();
                            alert("Couldn't retrieve information about the table " + table1 + ". We won't be able to help you for this query.");
                            $this.enable().text("Next");
                            return;
                        }
                        table1Name = tableDef.label || tableDef.name;
                        columns1 = tableDef.columns;
                        dialog.table1 = tableDef;
                        callback();
                    }, function() {
                        alert("Could not fetch database details for " + table1);
                        $this.enable().text("Next");
                    });
                }
            });
            dialog.addCancel("Cancel", cancelCallback);

            dialog.addPage("page-2");
            dialog.addHeader("Select matching columns");
            dialog.addPanel("Panel 2", "<span></span>");
            dialog.addButton("Previous", function() {
                dialog.prevPage();
            });

            // Last page
            dialog.addSubmit("Save", submitCallback);
            dialog.addCancel("Cancel", cancelCallback);

            dialog.gotoPage(0);
            dialog.gotoPanel(0);

            Confluence.PlaySQL.QueryEditor.databaseSchema.onUpdate.push(function(data) {
                var $select = dialog.getPage(0).body.find("#table-name"),
                    $option,
                    table;
                for (var i in data.tables["current schema"]) {
                    table = data.tables["current schema"][i];
                    $option = $select.find("option[value='" + table.name + "']");
                    if ($option.length == 0) {
                        $select.append($("<option></option>", {
                            text: table.label || table.name,
                            value: table.name
                        }));
                    } else {
                        $option.text(table.label || table.name);
                    }
                }
            });
            function updateMatchedColumns($checkbox) {
                var $el = $checkbox instanceof $ ? $checkbox : $(this),
                    isMatched = $el.is(":checked"),
                    position = $el.closest("td").index() + 1,
                    $columnTd = $el.closest("table").find("tr > td:nth-child(" + position + "), tr > th:nth-child(" + position + ")");
                $columnTd.toggleClass("matched", isMatched).toggleClass("unmatched", !isMatched);
                updateSummary();
            }
            function updateSummary() {
                var allMatchedCols = $("#wizard-subtract-dialog thead th.matched").map(function() { return $(this).data("colname"); }).toArray(),
                    matchingCols = $("#wizard-subtract-dialog tbody td.matched .match-column").map(function() { return $(this).val(); }).toArray(),
                    caseSensitivity = $("#wizard-subtract-dialog tbody td.matched .case-sensitivity").map(function() { return $(this).val(); }).toArray(),
                    table1Name = dialog.table1.label || dialog.table1.name,
                    table2Name = dialog.table2.label || dialog.table2.name,
                    text = "No row will be removed from \"" + table1Name + "\"";
                if (allMatchedCols.length > 0) {
                    text = "Match rows from \"$table1\" and \"$table2\" where \"$columns\" are equal to \"$columns2\""
                        .replace(/\$columns/, allMatchedCols.join("\" and \""))
                        .replace(/\$columns2/, matchingCols.join("\" and \""))
                        .replace(/\$table1/, table1Name)
                        .replace(/\$table2/, table2Name);
                }
                $("#wizard-subtract-dialog #summary").text(text);
            }
            $("#wizard-subtract-dialog").on("change", ".tickmatching", updateMatchedColumns);

            dialog.show();
        }
    });

    /** Wizard for joining two tables */
    wizards.push({
        name: "joinTwoTables",
        label: "Join two tables",
        run: function() {
            if (wizards.joinTwoTablesDialog) {
                wizards.joinTwoTablesDialog.show();
                wizards.joinTwoTablesDialog.gotoPage(0);
                return;
            }

            var spreadsheetList = [];
            for (var i in ed.databaseSchema.tableNames) {
                var table = ed.databaseSchema.tableNames[i],
                    tableDef = ed.databaseSchema.getTableDef(table);
                spreadsheetList.push({
                    name: table,
                    label: tableDef && tableDef.label || table
                });
            }

            var dialog,
                $template = $(Confluence.Templates.PlaySQL.wizardJoinChooseSpreadsheet({
                    spreadsheets: spreadsheetList
                })),

                cancelCallback = function () {
                    dialog.hide();
                },
                submitCallback = function () {
                    $(this).disable().text("Submitting...");
                    var originalSql = ed.aceEditor.getValue(),
                        allMatchedCols = $("#wizard-join-dialog thead th.matched").map(function() { return $(this).data("colname"); }).toArray(),
                        matchingCols = $("#wizard-join-dialog tbody td.matched .match-column").map(function() { return $(this).val(); }).toArray(),
                        caseInsensitivity = $("#wizard-join-dialog tbody td.matched .case-sensitivity").map(function() { return $(this).is(":checked"); }).toArray(),
                        table1Name = dialog.table1.name,
                        table2Name = dialog.table2.name,
                        /** mainTableName must be double-quote escaped */ mainTableName,
                        secondTableAlias,
                        parsedQuery,
                        joinCriteria, queryNotParsed,
                        fromClause,
                        sql;

                    if (table1Name != '') {
                        originalSql = "SELECT * FROM " + withQuotes(table1Name);
                        mainTableName = withQuotes(table1Name);
                    }
                        parsedQuery = Confluence.PlaySQL.Helpers.parseSQL(originalSql);
                        if (!parsedQuery.bilinear) {
                            queryNotParsed = "SELECT * FROM (" + originalSql + ") SUBQUERY";
                            mainTableName = "SUBQUERY";
                        } else {
                            //!\ fromClause is passed by ref, so we're modifying the query
                            fromClause = parsedQuery.clause("FROM");
                            var firstTable = parsedQuery.clause("FROM")[1];
                            mainTableName = firstTable.alias;
                            if (!mainTableName) {
                                mainTableName = firstTable.def.split(" ");
                                mainTableName = mainTableName[mainTableName.length-1];
                            }
                            if (mainTableName == table2Name) {
                                secondTableAlias = "S";
                            }
                        }

                    joinCriteria = "ON ";
                    for (var i in matchingCols) {
                        if (i > 0) joinCriteria += "\n\tAND ";
                        if (caseInsensitivity[i]) {
                            joinCriteria += "UPPER($table2.$col2) = UPPER($table1.$col1)";
                        } else {
                            joinCriteria += "$table2.$col2 = $table1.$col1";
                        }
                        joinCriteria = joinCriteria
                            .replace(/\$table1/, mainTableName)
                            .replace(/\$col1/, withQuotes(allMatchedCols[i]))
                            .replace(/\$table2/, secondTableAlias != undefined ? secondTableAlias : withQuotes(table2Name))
                            .replace(/\$col2/, withQuotes(matchingCols[i]));
                    }

                    if (queryNotParsed != undefined) {
                        // Working on an unparsed query
                        sql = queryNotParsed + "\nLEFT JOIN " + withQuotes(table2Name);
                        if (secondTableAlias != undefined) sql += " " + secondTableAlias;
                        sql += " " + joinCriteria;
                    } else {
                        // Working on a parsed query
                        if (fromClause != undefined) {
                            fromClause.push({
                                operator: "LEFT JOIN",
                                def: withQuotes(table2Name),
                                alias: secondTableAlias,
                                criteria: joinCriteria
                            });
                        }
                        parsedQuery.clause("FROM", fromClause);

                        sql = Confluence.PlaySQL.Helpers.rebuildSQL(parsedQuery);
                    }

                    ed.aceEditor.setValue(sql);
                    wizards.joinTwoTablesDialog = undefined;

                    $(this).enable().text("Submit");
                    dialog.hide().remove();
                };

            dialog = new AJS.ConfluenceDialog({
                width: 600,
                height: 450,
                id: 'wizard-join-dialog',
                closeOnOutsideClick: true,
                onCancel: cancelCallback,
                onSubmit: submitCallback
            });
            wizards.joinTwoTablesDialog = dialog;

            dialog.addHeader("Join two tables");
            dialog.addPanel("Panel 1", $template);
            dialog.addButton("Next", function() {
                var $this = $(this),
                    table1 = $("#wizard-join-choose-spreadsheet #table-name-main").val(),
                    table2 = $("#wizard-join-choose-spreadsheet #table-name-join").val(),
                    table1Name,
                    table2Name,
                    columns1,
                    columns2;

                if (table2 == undefined || table2 == "") {
                    $("#wizard-join-choose-spreadsheet .error").text("Please choose a table");
                } else {
                    $("#wizard-join-choose-spreadsheet .error").text("");
                }
                $this.disable().text("Fetching columns...");

                function callback() {
                    if (table1Name == undefined || table2Name == undefined) return;
                    var $page2 = dialog.getPage(1).body;
                    $page2.empty().html(Confluence.Templates.PlaySQL.wizardJoinMatchingColumns({
                        columns1: columns1,
                        columns2: columns2,
                        table1Name: table1Name,
                        table2Name: table2Name
                    }));
                    dialog.getPage(1).body.height(dialog.getPage(0).body.height());

                    // Can we preselect options?
                    var preselected = 0;
                    $page2.find("thead tr > th").each(function(){
                        var $this = $(this),
                            colname = $this.data("colname"),
                            $option = $page2.find("tbody > tr > td[headers='header-$colname'] option[value='$colname']".replace(/\$colname/g, colname)),
                            $tickMatch = $page2.find("tbody > tr > td[headers='header-$colname'] input.tickmatching".replace(/\$colname/g, colname));
                        if ($option.length == 1) {
                            // There's a column in the matching table with the exact column name. Preselect it.
                            $option.parent().val(colname);
                            preselected++;
                            if (preselected == 1) $tickMatch.attr("checked", "on");
                        }
                    });
                    $page2.find("tbody tr > td > .tickmatching").each(updateMatchedColumns);

                    $this.enable().text("Next");
                    dialog.nextPage();
                }

                ed.triggerSchemaUpdate(table2, function(data) {
                    var tableDef = data.getTableDef(table2);
                        if (tableDef == undefined) {
                            dialog.hide().remove();
                            alert("Couldn't retrieve information about the table " + table2 + ". We won't be able to help you for this query.");
                            $this.enable().text("Next");
                            return;
                        }
                    table2Name = tableDef.label || tableDef.name;
                    columns2 = tableDef.columns;
                    dialog.table2 = tableDef;
                    callback();
                }, function() {
                    alert("Could not fetch database details for " + table2);
                    $this.enable().text("Next");
                });

                if (table1 == "") {
                    // Take the columns from the query
                    var $previewTable = $(".playsql-query-container table.playsql"),
                        isBroken = $previewTable.hasClass("unavailable"),
                        columns = $previewTable.find("> thead > tr > th").map(function(){
                            var name = $(this).attr("title"),
                                label = $.trim($(this).text().replace(/\s+/, " ")) || name;
                            return {
                                name: name,
                                label: label
                            };
                        }).toArray();
                    if (isBroken || columns.length < 2) {
                        alert("If you choose to join records from the current query, please preview the query first.");
                        dialog.hide();
                    }
                    table1Name = "Current query";
                    columns1 = columns;
                    dialog.table1 = {
                        name: "",
                        label: table1Name,
                        columns: columns
                    };
                    callback();
                } else {
                    ed.triggerSchemaUpdate(table1, function(data) {
                        var tableDef = data.getTableDef(table1);
                        if (tableDef == undefined) {
                            dialog.hide().remove();
                            alert("Couldn't retrieve information about the table " + table1 + ". We won't be able to help you for this query.");
                            $this.enable().text("Next");
                            return;
                        }
                        table1Name = tableDef.label || tableDef.name;
                        columns1 = tableDef.columns;
                        dialog.table1 = tableDef;
                        callback();
                    }, function() {
                        alert("Could not fetch database details for " + table1);
                        $this.enable().text("Next");
                    });
                }
            });
            dialog.addCancel("Cancel", cancelCallback);

            dialog.addPage("page-2");
            dialog.addHeader("Select matching columns");
            dialog.addPanel("Panel 2", "<span></span>");
            dialog.addButton("Previous", function() {
                dialog.prevPage();
            });

            // Last page
            dialog.addSubmit("Save", submitCallback);
            dialog.addCancel("Cancel", cancelCallback);

            dialog.gotoPage(0);
            dialog.gotoPanel(0);

            Confluence.PlaySQL.QueryEditor.databaseSchema.onUpdate.push(function(data) {
                var $select = dialog.getPage(0).body.find("#table-name"),
                    $option,
                    table;
                for (var i in data.tables["current schema"]) {
                    table = data.tables["current schema"][i];
                    $option = $select.find("option[value='" + table.name + "']");
                    if ($option.length == 0) {
                        $select.append($("<option></option>", {
                            text: table.label || table.name,
                            value: table.name
                        }));
                    } else {
                        $option.text(table.label || table.name);
                    }
                }
            });
            function updateMatchedColumns($checkbox) {
                var $el = $checkbox instanceof $ ? $checkbox : $(this),
                    isMatched = $el.is(":checked"),
                    position = $el.closest("td").index() + 1,
                    $columnTd = $el.closest("table").find("tr > td:nth-child(" + position + "), tr > th:nth-child(" + position + ")");
                $columnTd.toggleClass("matched", isMatched).toggleClass("unmatched", !isMatched);
                updateSummary();
            }
            function updateSummary() {
                var allMatchedCols = $("#wizard-join-dialog thead th.matched").map(function() { return $(this).data("colname"); }).toArray(),
                    matchingCols = $("#wizard-join-dialog tbody td.matched .match-column").map(function() { return $(this).val(); }).toArray(),
                    table1Name = dialog.table1.label || dialog.table1.name,
                    table2Name = dialog.table2.label || dialog.table2.name,
                    text = "No row will be removed from \"" + table1Name + "\"";
                if (allMatchedCols.length > 0) {
                    text = "Remove rows of \"$table1\" with the same \"$columns\" as in \"$table2\""
                        .replace(/\$columns/, allMatchedCols.join("\" and \""))
                        .replace(/\$columns2/, matchingCols.join("\" and \""))
                        .replace(/\$table1/, table1Name)
                        .replace(/\$table2/, table2Name);
                }
                $("#wizard-join-dialog #summary").text(text);
            }
            $("#wizard-join-dialog").on("change", ".tickmatching", updateMatchedColumns);

            dialog.show();
        }
    });

    wizards.push({
        name: "wizardRequest",
        label: "Oz, the magic wizard...",
        link: "mailto:support+wizardrequest9@play-sql.com?subject=New%20wizard%20request&body=G%27day,%20support!%0A%0ACan%20you%20help%20me%20to%20...%20(please%20include%20your%20table%20names%20and%20what%20you%20want%20to%20do)%20...%20%3F%0A%0A%0A%0AAlso%2C%20can%20you%20please%20create%20a%20wizard%20for%20...%20in%20a%20future%20version%20of%20Play%20SQL%20please%3F%0A%0ACheers%2C%0A(your%20name)",
        target: "_blank"
    });
})(AJS.$);