(function($){
    Confluence.PlaySQL = Confluence.PlaySQL || {};
    var ed = Confluence.PlaySQL.Queryeditor || {};
    Confluence.PlaySQL.QueryEditor = ed;
    Confluence.PlaySQL.Editors = Confluence.PlaySQL.Editors || {};
    Confluence.PlaySQL.Editors.ed = ed;

    /**
     * Initialize the form with the Query object passed in argument
     */
    ed.initialize = function() {
        var query = ed.queryModel;
        if (!query) {
            throw "The query wasn't loaded in the page";
        }

        ed.aceEditor.setValue(query.sql);
        $("#query-title").val(query.label).data("id", query.id);
        if (ed.permlevel != "Q_SAVE") {
            $("#btn-save").remove();
            if (ed.permlevel != "Q_EDIT") {
                $("#query-editor-bottom").addClass("disabled");
                ed.aceEditor.setReadOnly(true);
            }
        } else {
            if (query.id == undefined) {
                $("#btn-save").text("Create");
            }
        }

        ed.preview();
        if (ed.useAutocomplete) {
            ed.triggerSchemaUpdate();
        }
        Confluence.PlaySQL.Selection.setupCopyHandler();
    };

    ed.getValues = function() {
        var title = $("#query-title").val(),
            id = $("#query-title").data("id");

        ed.queryModel.label = title;
        if ($(".query-editor-options").size() > 0) {
            var $userParametersFields = $(".field-group.user-parameter");
            ed.queryModel.userParameters = [];
            $userParametersFields.each(function(i) {
                var $field = $(this),
                    value = {
                        name : $field.find("[name=user-parameter-value]").val(),
                        label : $field.find("[name=user-parameter-label]").val(),
                        defaultValue : $field.find("[name=user-parameter-default]").val()
                    };
                ed.queryModel.userParameters.push(value);
            });
        }
        if (ed.queryModel.sql != ed.aceEditor.getValue()) {
            ed.queryModel.sql = ed.aceEditor.getValue();
            ed.forcePreflight = true;
        }
        ed.queryModel.sql = ed.aceEditor.getValue();
        return ed.queryModel;
    }
    ed.save = function(e) {
        e && e.preventDefault();
        try {
            ed.getValues();
            Confluence.PlaySQL.Editors.saveBarMessage("Saving...", true);
            Confluence.PlaySQL.Totals.updateColspans($(".field-group.totals").data('totals'), $(".playsql-spreadsheet-placeholder"));
            var query = ed.getValues(),
                id = query.id != undefined ? "/" + query.id : "",
                spaceKey = AJS.Meta.get("space-key"),
                url = AJS.contextPath() + "/rest/playsql/1.0/query/space:" + spaceKey + id;
            $.ajax({
                url: url,
                type: 'POST',
                contentType: "application/json; charset=utf-8",
                dataType: "json",
                data: JSON.stringify(query)
            }).done(function(data){
                Confluence.PlaySQL.Editors.saveBarMessage("Saved!", false);
                var hasOpener = false;
                try {
                    hasOpener = window.opener &&
                        !window.opener.closed &&
                        window.opener.AJS &&
                        window.opener.AJS.MacroBrowser &&
                        window.opener.AJS.MacroBrowser.fields &&
                        window.opener.AJS.MacroBrowser.fields.queryId &&
                        window.opener.AJS.MacroBrowser.fields.queryId.updateQuery;
                } catch (err) {
                    // Surely a Cors exception, localhost vs baseUrl
                    hasOpener = false;
                }
                if (hasOpener) {
                    window.opener.AJS.MacroBrowser.fields.queryId.setValue(data.id);
                    window.opener.AJS.MacroBrowser.fields.queryId.loadedQueryId = "please refresh anyway";
                    window.opener.AJS.MacroBrowser.fields.queryId.updateQuery();
                }
                if (query.id != data.id) {
                    var location = window.location.href,
                        parts = location.split("?"),
                        page = parts[0],
                        queryParameters = parts[1].split("&"),
                        newQueryParameters = {};
                    for (var i in queryParameters) {
                        var param = queryParameters[i].split("=");
                        if (param[0] == "key")
                            newQueryParameters[param[0]] = param[1];
                    }
                    newQueryParameters["queryId"] = data.id;
                    window.location.href = page + "?" + $.param(newQueryParameters);
                }
            }).fail(function(data){
                Confluence.PlaySQL.Editors.saveBarMessage("Error while saving.", false);
            });
        }
        catch (err) {
            AJS.log(err);
        }
        return false;
    }

    ed.preview = function(e) {
        e && e.preventDefault();
        ed.getValues();
        var spreadsheet = Confluence.PlaySQL.Spreadsheet.byPlaceholder($(".playsql-spreadsheet-placeholder"));
        spreadsheet.reset();
        spreadsheet.orderbyClause = ed.queryModel.orderby;
        spreadsheet.options.onSelectionSet = function(sSelection) {};
        spreadsheet.options.onDisplay = function(spreadsheet, successStatus, data) {
            if (successStatus == "SUCCESS" || successStatus == "PARTIAL_SUCCESS" || successStatus == "PARTIAL_FAILURE") {
                // Update the system columns mode
                var mode = spreadsheet.systemColumnsMode,
                    $dropDown = $("#system-columns-mode-dropdown");
                ed.queryModel.systemColumnsMode = mode;
                if (mode && $dropDown.size() == 1) {
                    if ($dropDown.auiSelect2("val") != mode) {
                        $dropDown.auiSelect2("val", mode).closest(".field-group")
                            .find(".systemColumnsModeExplanation").removeClass("hidden");
                    } else {
                        $dropDown.closest(".field-group")
                            .find(".systemColumnsModeExplanation").addClass("hidden");
                    }
                }
            }
            // We only save the other info if the success is total
            if (successStatus == "SUCCESS") {
                // Save the DDL into the QueryModel
                var columns = data.tableDef.columns;
                ed.queryModel.ddlCols = columns;
                ed.queryModel.orderby = spreadsheet.orderbyClause;
                ed.forcePreflight = false;

                // Display the hint on how to add totals
                if (spreadsheet.perms.hasPerm("EDIT_TOTALS")) {
                    var $totals = $("tr.totals-row td:not(.right-padding)");
                    if ($totals.size() == 1 && AJS.trim($totals.text()).length == 0) {
                        $totals.html($("<span>", {
                            "class": "totals-hint",
                            text: AJS.I18n.getText("com.playsql.playsql-base-plugin.query.editor.totals-hint")
                        }));
                    }
                }
            }
        };
        spreadsheet.options.onChangeTotals = function(spreadsheet) {
            ed.queryModel.totals = spreadsheet.totals;
        }
        spreadsheet.options.saveColumn = function(spreadsheet, column, newColumn, changes,
            successCallback, failureCallback) {
            // "column" is still attached to "spreadsheet", but not to the query
            // Note: We add 1 to account for the POSITION column, which queryModel maintains and spreadsheet doesn't
            var queryColumn = ed.queryModel.ddlCols[column.index + 1];
            if (queryColumn.name != column.name) {
                queryColumn = null;
                for (var i = 0 ; i < ed.queryModel.ddlCols.length ; i++) {
                    if (ed.queryModel.ddlCols[i].name == column.name) {
                        queryColumn = ed.queryModel.ddlCols[i];
                    }
                }
                if (queryColumn == null) {
                    return failureCallback("The column name couldn't be found: " + column.name);
                }
            }
            if (changes.indexOf("RENAME") != -1) {
                column.name = newColumn.name;
                queryColumn.name = newColumn.name;
            }
            if (changes.indexOf("CHTYPE") != -1) {
                column.type = newColumn.type;
                queryColumn.type = newColumn.type;
            }
            if (changes.indexOf("CHLBL") != -1) {
                column.label = newColumn.label;
                queryColumn.label = newColumn.label;
            }
            if (changes.indexOf("CHFMA") != -1) {
                column.formula = newColumn.formula;
                queryColumn.formula = newColumn.formula;
            }
            if (changes.indexOf("CHRDR") != -1) {
                column.renderer = newColumn.renderer;
                queryColumn.renderer = newColumn.renderer;
                column.viewRenderer = newColumn.viewRenderer;
                queryColumn.viewRenderer = newColumn.viewRenderer;
            }
            if (changes.indexOf("CHFMT") != -1) {
                column.formats = newColumn.formats;
                queryColumn.formats = newColumn.formats;
            }
            if (changes.indexOf("WIDTH") != -1) {
                column.width = newColumn.width;
                queryColumn.width = newColumn.width;
            }

            successCallback && successCallback();
        };
        // We don't save the totals. We just write them up in the queryModel and preview the results.
        spreadsheet.options.saveTotals = function(spreadsheet, urlOptions, successCallback, failureCallback){
            ed.queryModel.totals = spreadsheet.totals;
            var url = spreadsheet.baseUrl("/data/preview", urlOptions);
            $.ajax({
                url: url,
                type: 'POST',
                contentType: "application/json; charset=utf-8",
                timeout: 180000,
                cache: false,
                contentType: "application/json; charset=utf-8",
                dataType: "json",
                data: JSON.stringify(spreadsheet.queryModelOverride)
            }).done(successCallback).fail(failureCallback);
        };
        spreadsheet.queryModelOverride = ed.queryModel;
        if (ed.forcePreflight) {
            spreadsheet.load({forcePreflight: true});
        } else {
            spreadsheet.load();
        }
        return false;
    }

    ed.refit = function() {
        var $area = $("#query-sql");
        $area.scrollTop(10000);
        var scrollTop = $area.scrollTop();
        if (scrollTop > 0) {
            var height = $area.height() + scrollTop + 10;
            $area.height(height);
        }
    };

    /**** Add/remove fields ***/
    ed.addOptionsPanel = function() {
        var $existing = $(".query-editor-options");
        if ($existing.size() == 1) {
            $existing.removeClass("hidden");
            $(".hide-options-panel").toggleClass("hidden");
            $(".add-options-panel").toggleClass("hidden");
            return;
        }
        var $template = $(Confluence.Templates.PlaySQL.optionsPanel({queryModel: ed.queryModel}));
        $template.appendTo($("#query-editor-bottom .aui-toolbar2-primary"));
        $template.find("#usebuiltinparameters").change(function(){
            ed.queryModel.builtInParameters = $(this).is(":checked");
        });
        $template.find("#usecache").change(function(){
            ed.queryModel.useCache = $(this).is(":checked");
        });
        $template.find("#useform").change(function(){
            ed.queryModel.useForm = $(this).is(":checked");
        });
        $template.find("#timeout").change(function(){
            var val = $(this).val(),
                timeout = parseInt(val);
            if (!isNaN(timeout)) {
                ed.queryModel.timeout = timeout;
            } else {
                ed.queryModel.timeout = undefined;
            }
        });
        $template.find("#customEmptyMessage").change(function(){
            var val = $(this).val();
            if (val != "") {
                ed.queryModel.customEmptyMessage = val;
            } else {
                ed.queryModel.customEmptyMessage = undefined;
            }
        });
        $template.find(".add-user-parameter").click(function(e){
            e.preventDefault();
            ed.addUserParameterField();
        });
        $template.find("#system-columns-mode-dropdown").auiSelect2({
            formatResult: function(state) {
                if (!state.id) return state.text;
                return "<div class=\"title\">" + $(state.element).data("label") + "</div><div class=\"description\">" + state.text + "</div>";
            },
            formatSelection: function(state) {
                if (!state.id) return state.text;
                return "<div class=\"title\">" + $(state.element).data("label") + "</div>";
            },
            escapeMarkup: function(m){return m;},
            dropdownAutowidth: false,
            dropdownCss: {width: "300px"},
            dropdownCssClass: "system-columns-mode-dropdown-contents",
            minimumResultsForSearch: -1 //deactivates the search box
        }).auiSelect2("val", ed.queryModel.systemColumnsMode).on("change", function(e){
            ed.queryModel.systemColumnsMode = e.val;
            ed.forcePreflight = true;
        });
        $(".hide-options-panel").toggleClass("hidden");
        $(".add-options-panel").toggleClass("hidden");
    }
    ed.removeOptionsPanel = function() {
        var $field = $(".query-editor-options");
        $field.animate({height: 0}, function(){
            $field.addClass("hidden").css("height", "");
        });
        $(".hide-options-panel").toggleClass("hidden");
        $(".add-options-panel").toggleClass("hidden");
    };

    ed.getUserParameterCount = function() {
        return $(".field-group.user-parameter").size();
    };
    ed.addUserParameterField = function(value) {
        if ($(".query-editor-options").size() == 0) {
            ed.addOptionsPanel();
        }
        var index = ed.getUserParameterCount();
        if (value == undefined) {
            value = {
                name: "$param" + (index+1),
                label: "Parameter " + (index+1) + ":",
                defaultValue: ""
            };
        }
        var template = Confluence.Templates.PlaySQL.userParameterField({
                index: index,
                value: value
            }),
            $template = $(template).appendTo($(".query-editor-options"));
        ed.queryModel.userParameters = ed.queryModel.userParameters || [];
        ed.queryModel.userParameters.push(value);
    };
    ed.removeUserParameterField = function($fieldGroup) {
        $fieldGroup.animate({height: 0}, function(){
            $fieldGroup.remove();
        });
        ed.queryModel.userParameters.splice(0, 1);
    };

    var updateAutocomplete = function() {
        var data = ed.databaseSchema,
            $panel = $(".playsql-contexthelp-panel"),
            $ul = $panel.children("ul"),
            tableName, table, $ul, $li, $columnList,
            atLeastOneTable = false;
        if ($ul.length == 0) {
            $ul = $("<ul></ul>").appendTo($panel);
        }
        for (var i in data.tableNames) {
            tableName = data.tableNames[i];
            $li = $ul.find("> li[data-table='" + tableName + "']");
            if ($li.length != 1) {
                if ($li.length > 1) $li.remove();
                $li = $("<li></li>", {
                    text: tableName,
                    "data-table": tableName,
                    "class": "table"
                }).appendTo($ul);
            } else {
                $li.empty().text(tableName);
            }

            for (var j in data.tables["current schema"]) {
                if (data.tables["current schema"][j].name == tableName) {
                    columns = data.tables["current schema"][j].columns;
                    $columnList = $("<ul></ul>").appendTo($li);
                    for (var k in columns) {
                        $("<li></li>", {
                            text: columns[k].name,
                            title: "Type: " + columns[k].type + (columns[k].length ? " (" + columns[k].length + ")" : "") +
                                    "\nRenderer: " + (columns[k].renderer != undefined ? columns[k].renderer : "-") +
                                    "\nLabel: " + columns[k].label
                        }).appendTo($columnList);
                    }
                    atLeastOneTable = true;
                    break;
                }
            }
        }
        var $lis = $ul.children();
        $lis.detach().sort(function(a, b){
            var aTextNode = a.childNodes[0],
                bTextNode = b.childNodes[0],
                aText = aTextNode.nodeType == 3 ? aTextNode.nodeValue : $(a).text(),
                bText = bTextNode.nodeType == 3 ? bTextNode.nodeValue : $(b).text();
            return aText.localeCompare(bText);
        });
        $ul.append($lis);
        // Re-update each time we update the model
        ed.databaseSchema.onUpdate.push(updateAutocomplete);
        // Display the panel + the wizards if that's the first time
        if ($panel.parent().hasClass("hidden") && atLeastOneTable) {
            $panel.parent().removeClass("hidden");
            $("#wizards-dropdown").removeClass("hidden");
        }
    }

    ed.databaseSchema = {
        tableNames: [],
        dialect: undefined,
        dialectDefaultCase: undefined,
        tables: {
            "current schema": []
        },
        // All callbacks after the schema is updated
        onUpdate : [updateAutocomplete],
        getTableDef : function(tableName) {
            for (var schema in ed.databaseSchema.tables) {
                for (var i in ed.databaseSchema.tables[schema]) {
                    if (ed.databaseSchema.tables[schema][i].name == tableName) {
                        return ed.databaseSchema.tables[schema][i];
                    }
                }
            }
            return undefined;
        }
    }

    ed.triggerSchemaUpdate = function(tablename, callback, onFail) {
        if (!ed.useAutocomplete) {
            onFail && onFail();
        }
        callback && ed.databaseSchema.onUpdate.push(callback);
        var spaceKey = AJS.Meta.get("space-key"),
            url = AJS.contextPath() + "/rest/playsql/1.0/autocomplete/space:" + spaceKey;
        if (tablename) url += "/" + tablename;
        $.ajax({
            url: url,
            type: 'GET',
            contentType: "application/json; charset=utf-8"
        }).done(function(data){
            data = $.extend(true, ed.databaseSchema, data);
            var callbacks = ed.databaseSchema.onUpdate.splice(0);
            for (var i in callbacks) {
                callbacks[i](ed.databaseSchema);
            }
        }).fail(function(data){
            //Say nothing. It's annoying for non-Postgres users.
            //Confluence.PlaySQL.Editors.saveBarMessage("Couldn't load autocompletion info.", false);
            onFail && onFail();
        });
    }
})(AJS.$);




/* ********************** Initialize the screen *****************/

AJS.$(function($){
    var ed = Confluence.PlaySQL.QueryEditor,
        $sqlPanel = $("#query-sql");

    if ($sqlPanel.size() == 0) return;

    $sqlPanel.unbind('.updateSize')
        .bind('keyup.updateSize', ed.refit)
        .bind('keydown.updateSize', ed.refit)
        .bind('change.updateSize', ed.refit)
        .bind('click.updateSize', ed.refit);

    $(".add-options-panel").click(function(e){
        e.preventDefault();
        ed.addOptionsPanel();
    });
    $(".hide-options-panel").click(function(e){
        e.preventDefault();
        ed.removeOptionsPanel();
    });
    $("#query-editor-bottom").on("click", ".remove", function() {
        var $this = $(this);
        if ($this.closest(".system-columns-mode").size() == 1) {
            ed.removeSystemColumnsModeField();
        } else {
            // It's a user parameter
            ed.removeUserParameterField($this.closest(".field-group.user-parameter"));
        }
        return false;
    });
    $("#btn-save").click(ed.save).tooltip({delayIn: 0, gravity: "ne"});
    $(".btn-preview").click(ed.preview);
    $("#query-title").keypress(function(e) {
        if (e.which == 13) {
            ed.save();
            ed.preview();
            return false;
        }
    });

    ed.aceEditor = ace.edit("query-sql");
    ed.aceEditor.setTheme("ace/theme/clouds");
    // editor.setTheme("ace/theme/ambiance");
    ed.aceEditor.getSession().setMode("ace/mode/sql");
    ed.aceEditor.commands.addCommand({
        name: 'preview',
        bindKey: {win: 'Ctrl-Enter',  mac: 'Command-Enter'},
        exec: function(editor) {
            ed.preview()
        },
        readOnly: true // false if this command should not apply in readOnly mode
    });
    ed.aceEditor.setShowPrintMargin(false);

    // Adjust editor height according to contents
    var heightUpdateFunction = function() {
        var contentHeight =
            ed.aceEditor.getSession().getScreenLength()
                * ed.aceEditor.renderer.lineHeight
                + ed.aceEditor.renderer.scrollBar.getWidth();
        // Must be more than 200px and less than 2/3 of the screen
        var maxHeight = ($(window).height() || 500) / 2,
            minHeight = 200,
            height = Math.max(minHeight, Math.min(maxHeight, contentHeight));

        $('#query-sql, .playsql-contexthelp-panel').height(height + "px");

        //$('#editor-section').height(height + "px");

        // This call is required for the editor to fix all of
        // its inner structure for adapting to a change in size
        ed.aceEditor.resize();
    };

    heightUpdateFunction();
    ed.aceEditor.getSession().on('change', heightUpdateFunction);

    // In the right sidebar, expand/collapse on click
    $(".playsql-contexthelp-panel").on("click", "li", function() {
        var $this = $(this);
        $this.toggleClass("expanded");
        if ($this.hasClass("table") && $this.children().length == 0) {
            ed.triggerSchemaUpdate($this.text());
        }
        return false;
    });

    var $wizardsDropdown = $("#dropdown-wizards ul");
    for (var i in Confluence.PlaySQL.QueryEditor.Wizards) {
        var wizardDef = Confluence.PlaySQL.QueryEditor.Wizards[i];
        $wizardsDropdown.append($("<li></li>", {
            html: $("<a></a>", {
                text: wizardDef.label,
                "data-wizard": wizardDef.name,
                href: wizardDef.link || "#",
                target: wizardDef.target || ""
            })
        }));
    }

    $wizardsDropdown.on("click", "a", function() {
        var wizardName = $(this).data("wizard");
        for (var i in Confluence.PlaySQL.QueryEditor.Wizards) {
            if (wizardName == Confluence.PlaySQL.QueryEditor.Wizards[i].name) {
                var wizard = Confluence.PlaySQL.QueryEditor.Wizards[i];
                if (wizard.run) {
                    wizard.run();
                }
                break;
            }
        }
    });

    // Initialize
    var $query = $("#query-form");
    ed.queryModel = $query.data("initial-data");
    ed.forcePreflight = false;
    ed.permlevel = $query.data("permlevel");
    ed.useAutocomplete = !!AJS.Meta.get("datasource-autocomplete");
    if (!ed.queryModel) {
        a = AJS.messages.error(".all-messages", {body: "An error occurred while loading this page."});
    } else {
        ed.initialize();
    }
});