/**
 * This file overrides the queryId field for the Macro Browser
 */
AJS.bind("init.rte", function() {
    var $ = AJS.$;
    function fillValues() {
        var spaceKey = AJS.Meta.get("space-key"),
            url = AJS.contextPath() + "/rest/playsql/1.0/query/space:" + spaceKey;
        $.ajax({
            url: url,
            type: 'GET',
            dataType: "json"
        }).done(function(data){
            var $select = $("#macro-param-queryId"),
                originalValue = $select.val();
            $select.empty();
            addValue($select, "undefined", "Select a value");
            for (var i in data) {
                addValue($select, i, data[i]);
            }
            setQueryId($select, originalValue);
            updateUserParameters();
        }).fail(function(data){
            AJS.logError("Can't fetch the list of queries for space:" + spaceKey)
        });
    }
    function addValue($select, value, text) {
        $select.append(AJS.$("<option/>", {
            value: value,
            text: text
        }));
    }
    function setQueryId($select, id) {
        $select.val(id);
        if (id != undefined && $select.val() !== id) {
            addValue($select, id, "Query");
            $select.val(id);
        }
    }
    var makeParameterDiv = AJS.MacroBrowser.makeParameterDiv;
    if (makeParameterDiv == null) {
        /**
         * This function existed until 5.6.0, so here's a copy for ulterior versions...
         */
        makeParameterDiv = function (metadata, param, jsOverride) {
            var t = AJS.MacroBrowser, field;
            var type = param.type.name;

            // Plugin point - other JS files can define more specific field-builders based on macro name, param name and
            // type.
            if (jsOverride) {
                var builder = jsOverride.fields && jsOverride.fields[type];
                if (builder && typeof builder != "function") {
                    // Types can be overridden for specific parameters - so the "type" object contains a "name" function.
                    builder = builder[param.name];
                }
                if (typeof builder == "function") {
                    field = builder.call(jsOverride, param);
                }
            }
            // If no override specific to the macro, look for general overrides specific to the parameter type.
            if (!field) {
                if (!(type in t.ParameterFields && typeof t.ParameterFields[type] == "function")) {
                    type = "string";
                }
                field = t.ParameterFields[type](param);
            }

            t.fields[param.name] = field;
            var paramDiv = field.paramDiv;
            var input = field.input;

            var paramId = "macro-param-" + param.name;
            paramDiv.attr("id", "macro-param-div-" + param.name);
            input.addClass("macro-param-input").attr("id", paramId);
            if (param.hidden) {
                paramDiv.hide();
            }

            // Use param label and desc or correct fallback.
            var pluginKey = metadata.pluginKey;
            if (param.displayName == AJS.MacroBrowser.Utils.makeDefaultKey(pluginKey, metadata.macroName, "param", param.name, "label")) {
                param.displayName = param.name;
            }
            if (param.description == AJS.MacroBrowser.Utils.makeDefaultKey(pluginKey, metadata.macroName, "param", param.name, "desc")) {
                param.description = "";
            }

            var labelText = param.displayName;
            if (param.required) {
                labelText += " *";
                paramDiv.addClass("required");  // set class against div, not input, to allow styling of label if nec
            }
            $("label", paramDiv).attr("for", paramId).text(labelText);

            if (param.description) {
                paramDiv.append(Confluence.Templates.MacroBrowser.macroParameterDesc({ description: param.description }));
            }
            return paramDiv;
        }
    }

    function updateUserParameters() {
        var queryIdField = AJS.MacroBrowser.fields["queryId"],
            queryId = queryIdField.getValue(),
            queryName = $("#macro-param-queryId option:selected").text(),
            oldQueryId = queryIdField.loadedQueryId,
            currentContext = AJS.MacroBrowser.fields['context'].getValue(),
            spaceKey = (currentContext.indexOf("space:") == 0 ? currentContext.substr(6) : AJS.Meta.get("space-key"));
        if (queryId === oldQueryId) {
            return;
        }
        queryIdField.loadedQueryId = queryId;
        // In IE8, make sure the MacroBrowser knows required fields have values
        AJS.MacroBrowser.paramChanged();

        AJS.MacroBrowser.fields["queryName"].setValue(queryName);

        for (var fieldName in AJS.MacroBrowser.fields) {
            if (fieldName.indexOf("user-parameter-") == 0) {
                AJS.MacroBrowser.fields[fieldName].remove();
            }
        }

        if (queryId == undefined || queryId == "undefined") {
            var $description = $("#macro-param-div-queryId .macro-param-desc"),
                $createLink = $("<a></a>", {
                        href : AJS.contextPath() + "/playsql-base-plugin/query-editor.action?key=" + spaceKey + "&create=true",
                        target : "_blank",
                        text : "or create a new query..."
                    });
            if ($description.size() == 0) {
                $description = $("#macro-param-div-queryId")
                    .append("<div class='macro-param-desc'></div>")
                    .find("div");
            }
            $description.unbind().empty().html($createLink);
        } else {
            var $description = $("#macro-param-div-queryId .macro-param-desc"),
                $editLink = $("<a></a>", {
                    href : AJS.contextPath() + "/playsql-base-plugin/query-editor.action?key=" + spaceKey + "&queryId=" + queryId,
                    text : "Edit the query...",
                    target : "_blank"
                });
            if ($description.size() == 0) {
                $description = $("#macro-param-div-queryId")
                    .append("<div class='macro-param-desc'></div>")
                    .find("div");
            }
            $description.unbind().empty().html($editLink);

            var url = AJS.contextPath() + "/rest/playsql/1.0/query/" + currentContext + "/" + queryId
            $.ajax({
                url: url,
                type: 'GET',
                dataType: "json"
            }).done(function(data){
                if (!data || !data.userParameters)
                {
                    return;
                }
                var userParamList = data.userParameters,
                    queryName = data.label,
                    $macroFieldArea = $("#macro-insert-container .macro-input-fields"),
                    $form = $macroFieldArea.find("form"),
                    metadata = AJS.MacroBrowser.getMacroMetadata("playsql-query"),
                    userParam, name, $paramDiv, $input, field, value;
                // Update the query name, if required
                $macroFieldArea.find("#macro-param-queryId option[value='" + queryId + "']").text(queryName);
                AJS.MacroBrowser.fields["queryName"].setValue(queryName);

                // Since Confluence 5.6.0, .macro-input-fields contains a form and it's required to use it
                if ($form.length > 0) $macroFieldArea = $form;

                for (var i in userParamList) {
                    userParam = userParamList[i];
                    name = "user-parameter-" + userParam.name.replace(/[^\w]/gi, '');
                    $paramDiv = makeParameterDiv.call(AJS.MacroBrowser, metadata, {name: name, displayName: userParam.label, type: {name: "string"}});
                    $input = $paramDiv.find("input");
                    value = AJS.MacroBrowser.unknownParams[name];

                    $macroFieldArea.append($paramDiv);
                    field = AJS.MacroBrowser.Field($paramDiv, $input, {});
                    field.remove = (function($paramDiv, name, field) {
                        return function() {
                            AJS.MacroBrowser.unknownParams[name] = field.getValue();
                            $paramDiv.remove();
                        }
                    })($paramDiv, name, field);
                    field.userParameterName = name;
                    if (value != undefined) {
                        field.setValue(value);
                    } else if (userParam.defaultValue != undefined) {
                        field.setValue(userParam.defaultValue);
                    }
                    AJS.MacroBrowser.fields[name] = field;
                }
                $("#macro-browser-preview-link").click();
            }).fail(function(data){
                AJS.logError("Can't fetch the query for space:" + spaceKey)
            });
        }
    }

    /** Creates a new query, displays it in a new tab, and report its ID into
     * the drop-down
     */
    function createNewQuery() {
        var currentUser = AJS.Meta.get("current-user-fullname"),
            query = {
                label : "New query" + (currentUser ? " by " + currentUser: ""),
                sql : ""
            },
            spaceKey = AJS.Meta.get("space-key"),
            url = AJS.contextPath() + "/rest/playsql/1.0/query/space:" + spaceKey,
            windowUrl = AJS.contextPath() + "/playsql-base-plugin/query-editor.action?key=" + spaceKey + "&queryId=",
            //newTab = window.open(windowUrl, 'newtab'),
            $link = $(this);
        $link.unbind().text("Edit the query...");
        $.ajax({
            url: url,
            type: 'POST',
            contentType: "application/json; charset=utf-8",
            dataType: "json",
            data: JSON.stringify(query)
        }).done(function(data){
            if (data.id != undefined) {
                var windowNewUrl = windowUrl + data.id
                // Update the link
                $link.attr('href', windowNewUrl).attr("target", "newtab");
                // Reset the tab's url
                //newTab.location.href = newTab.location.href.replace(windowUrl, windowNewUrl);
                // Update the $select
                var $select = $("#macro-param-queryId");
                addValue($select, "" + data.id, query.label);
                setQueryId($select, "" + data.id);
            }
        }).fail(function(jqXHR, error, reason){
            AJS.logError("Error while attempting to create a query. " + error + " - " + reason);
        });
        return false;
    }

    var jsOverrides = {
        "fields" : {
            "string" : function(params, options){
                options = options || {};
                if (params && params.name != "queryId") {
                    return AJS.MacroBrowser.ParameterFields["string"](params, options);
                }

                var $paramDiv = $(Confluence.Templates.MacroBrowser.macroParameterSelect());
                var $select = $("select", $paramDiv);

                options.setValue = function(value) {
                    setQueryId($("#macro-param-queryId"), value);
                }

                fillValues();

                options.onchange = updateUserParameters;

                var field = AJS.MacroBrowser.Field($paramDiv, $select, options);
                field.updateQuery = function(){
                    updateUserParameters();
                    $("#macro-browser-preview-link").click();
                }
                return field;

            }
        },
        "beforeParamsSet": function(paramMap, inserting) {
            if (!paramMap["context"]) {
                paramMap["context"] = "space:" + AJS.Meta.get("space-key");
            }
            // This is a migration from older versions. It happens when we open the
            // Macro Browser.
            if (paramMap["hideRedCell"]) {
                paramMap["output-rendering"] = "POOR";
                delete paramMap["hideRedCell"];
            }
            if (paramMap["hideHeaders"]) {
                paramMap["output-rendering"] = "TEXT";
                delete paramMap["hideHeaders"];
            }
            return paramMap;
        },
        "beforeParamsRetrieved" : function(paramMap, macro, sharedParamMap) {
            for (var fieldName in paramMap)
            {
                if (fieldName.indexOf('user-parameter') == 0) {
                    delete paramMap[fieldName];
                }
            }
            for (var fieldName in AJS.MacroBrowser.fields) {
                if (fieldName.indexOf("user-parameter-") == 0) {
                    var userField = AJS.MacroBrowser.fields[fieldName],
                        val = userField.getValue(),
                        name = userField.userParameterName;
                    paramMap[name] = val;
                }
            }
            return paramMap;
        }
    };
    AJS.MacroBrowser.setMacroJsOverride("playsql-query", jsOverrides);


    AJS.Confluence.PropertyPanel.Macro.registerButtonHandler("playsql-edit-query", function(e, macroNode) {
        var parameters = $(macroNode).attr("data-macro-parameters");
        if (typeof parameters != "string") return;
        parameters = parameters.split("|");
        var queryId,
            spaceKey = AJS.Meta.get("space-key"),
            context = "space:" + spaceKey;
        for (var i in parameters) {
            if (parameters[i].indexOf("queryId=") == 0) {
                queryId = parameters[i].substr("queryId=".length);
            } else if (parameters[i].indexOf("context") == 0) {
                context = parameters[i].substr("context=".length);
                if (context.indexOf("space:") == 0) {
                    spaceKey = context.substr("space:".length);
                }
            }
        }
        if (queryId != undefined) {
            window.open(AJS.contextPath() + "/playsql-base-plugin/query-editor.action?" + $.param({
                key : spaceKey,
                queryId : queryId
            }));
        }
    });
});