window.Confluence = window.Confluence || {};
window.Confluence.PlaySQL = window.Confluence.PlaySQL || {};
window.Confluence.PlaySQL.Spreadsheet = window.Confluence.PlaySQL.Spreadsheet || {};
window.Confluence.PlaySQL.Spreadsheet.CellProperties = window.Confluence.PlaySQL.Spreadsheet.CellProperties || {

    lastChosenColor : undefined,

    /**
     * Displays the formula editor for this cell.
     * If the cell was in edit mode, cancel it first.
     */
    display : function(spreadsheet, $cells) {
        $cells = $cells.filter("td:not(.right-padding)");
        var size = $cells.size(),
            formula = undefined;
        if (spreadsheet.inlinedialog) {
            if (spreadsheet.inlinedialog.is("#inline-dialog-cell-properties")) {
                // We just close the current cell properties - we don't open new ones
                spreadsheet.inlinedialog.close();
                return;
            }
            spreadsheet.inlinedialog.close();
        }
        if (size == 0) return;
        // If one cell, display the formula
        if (size == 1) {
            formula = spreadsheet.getCellMetadata($cells, "formula");
        } else {
            formula = $cells.filter(".formula").size() > 0;
        }

        if (typeof Confluence.PlaySQL.Spreadsheet.CellProperties.lastChosenColor === "undefined") {
            // Retrieve the _declared_ bgColor, not the inherited css bgColor.
            Confluence.PlaySQL.Spreadsheet.CellProperties.lastChosenColor = $cells.first()[0].style.backgroundColor;
        }

        // Create the inline dialog
        spreadsheet.inlinedialog && spreadsheet.inlinedialog.close && spreadsheet.inlinedialog.close();
        var inlinedialog = AJS.InlineDialog($cells, "cell-properties",
            function ($content, trigger, showPopup) {
                $content.css("width", "");
                $content.closest("#inline-dialog-cell-properties").addClass("small-inline");
                $content.html(Confluence.Templates.PlaySQL.cellProperties({
                    formula: formula,
                    cellCount: size,
                    lastChosenColor: Confluence.PlaySQL.Spreadsheet.CellProperties.lastChosenColor,
                    palette:
                    // Pastel palette
                    [
                    ["#202020", "#404040", "#808080", "#bfbfbf", "#f2f2f2", "transparent"],
                    ["#948a55", "#b3aa7c", "#c4bd9b", "#d5d1b5", "#e6e4d2", "#f0ece1"],
                    ["#23518b", "#2a69a8", "#3173c4", "#4785d2", "#72a1dc", "#1f497e"],
                    ["#622524", "#953736", "#d9a094", "#e6bab8", "#f5dcdb", "#c0514d"],
                    ["#4e6326", "#72933d", "#c4d69b", "#d8e4bd", "#eff1de", "#9cbb59"],
                    ["#253f61", "#346092", "#95b3d8", "#b9cee5", "#dfe6f4", "#4f82bd"],
                    ["#b38808", "#f7c121", "#fcf9af", "#feffea", "#fefaec", "#fada7b"],
                    ["#901929", "#d9253e", "#f0a7b1", "#f5c4cb", "#fae2e5", "#e66c7d"],
                    ["#403154", "#654a7b", "#b4a2c7", "#ccc1db", "#e6e4ec", "#8065a2"],
                    ["#215b68", "#32859c", "#93cddf", "#b7dfe8", "#dbeff4", "#4aacc7"],
                    ["#994808", "#e16c0a", "#fab990", "#fdd5b5", "#feeada", "#f79646"]
                    ]
                    // AUI Palette
                    /*[["transparent"],
                        ["#205081", "#4a6785", "#3572b0", "#84bbc6", "#59afe1", "#ebf2f9"],
                        ["#14892c", "#67ab49", "#8eb021"],
                        ["#654982", "#ac707a"],
                        ["#d04437", "#f15c75", "#f691b2"],
                        ["#333333", "#999999", "#cccccc", "#707070", "#f5f5f5"],
                        ["#815b3a", "#f79232", "#d39c3f", "#f6c342", "#f1a257"]]*/
                    // Confluence palette:
                    // [["#000000","#800000","#ff0000","#ff00ff","#ff99cc"],["#993300","#ff6600","#ff9900","#ffcc00","#ffcc99"],["#333300","#808000","#99cc00","#ffff00","#ffff99"],["#003300","#008000","#339966","#00ff00","#ccffcc"],["#003366","#008080","#33cccc","#00ffff","#ccffff"],["#000080","#0000ff","#3366ff","#00ccff","#99ccff"],["#333399","#666699","#800080","#993366","#cc99ff"],["#333333","#808080","#999999","#c0c0c0","#ffffff"]]
                }));
                showPopup();
                if (spreadsheet.perms.hasPerm('EDIT_ROWS_METADATA')) {
                    $content.find(".color-item, .aui-button-split-main").click(function (e) {
                        e.preventDefault();
                        var $this = $(this).is(".color-item") ? $(this) : $(this).find("span"),
                            color = $this[0].style.backgroundColor;
                        if (color == "transparent" || /[^0-9]0\)$/.test(color)) color = null;
                        Confluence.PlaySQL.Spreadsheet.CellProperties.lastChosenColor = color;
                        $cells.each(function () {
                            var $cell = $(this);
                            if ($cell.is("td")) {
                                //$cell.css("background-color", color);
                                $content.find(".aui-button-split-main span").css("background-color", color || "");
                                spreadsheet.saveCellMetadata($cell, "background-color", color);
                            }
                        });
                        spreadsheet.sprayRequestPipe();
                    });
                }
                $content.find(".formula-recalculate").click(function(e){
                    e.preventDefault();
                    spreadsheet.recalculateFormulas($cells);
                    spreadsheet.inlinedialog && spreadsheet.inlinedialog.close();
                });
                $content.find("#playsql-color-picker-background .aui-button-split-more").click(function(){
                    var $arrow = $(this);
                    // Ensure it happens after the dialog was displayed
                    window.setTimeout(function(){
                        var $el = $("#playsql-color-picker-dropdown:visible");
                        if ($el.size() == 0 || !$arrow.is(":visible")) return;
                        if ($el.offset().top > $arrow.offset().top + 50) {
                            $el.css("top", (parseInt($el.css("top")) - window.scrollY) + "px");
                        }
                    }, 0);
                });

                return false;
            }, {
                noBind: true,
                persistent: true,
                hideCallback: function () {
                    spreadsheet.inlinedialog && spreadsheet.inlinedialog.close();
                },
                hideDelay: 10000000
            }
        );
        spreadsheet.inlinedialog = inlinedialog;
        inlinedialog.close = function(){
            inlinedialog.remove();
            $(".tipsy.tipsy-n").remove();
            $("body, .playsql-spreadsheet-placeholder th").off(".playsql-inlinedialog");
            if (spreadsheet.inlinedialog == this) {
                spreadsheet.inlinedialog = undefined;
            }
        }
        var onClickOutsideInlineDialog = function (e) {
            if (!spreadsheet.inlinedialog || !spreadsheet.inlinedialog.is(":visible")) return;
            var target = e.target || this;
            var $this = $(target);
            if ($this.closest("#inline-dialog-cell-properties").size() > 0) return;
            var selectionIsCorrect = $cells.size() == $(".selected").size();
            $cells.each(function() {
                if (!$(this).hasClass("selected"))
                    selectionIsCorrect = false;
                return selectionIsCorrect;
            });
            if (selectionIsCorrect)
                return;
            spreadsheet.inlinedialog.close();
            // If we don't prevent bubbling, each of the parents will receive the same click
            return false;
        };
        $("body").on("click.playsql-inlinedialog", "*", onClickOutsideInlineDialog);
        inlinedialog.show();
        return inlinedialog;
    }
};