(function($){
    AJS.bind("playsql-spreadsheet-init", function(event, api) {

        Confluence.PlaySQL.Spreadsheet.ColumnToolbar = {
            autoCompleteFilterSQL: function(column, term, data){
                // SQL?
                if (/[a-z =<>]/.test(term)) {
                    if (/(AND|OR|\(|\))/g.test(term.toUpperCase())) {
                        data.addEntry({
                            label: term.toUpperCase(),
                            tip: "keyword",
                            type: "strict",
                            sql: " " + term.toUpperCase() + " "
                        });
                    } else if (/^(IS\s*|IS\s+NULL|NULL)$/.test(term.toUpperCase())) {
                        data.addEntry({
                            label: "IS NULL",
                            tip: "keyword",
                            sql: "$col IS NULL",
                            replaceColName: "$col"
                        });
                    } else {
                        // And always display the raw user's entry
                        if (/\$col\b/.test(term)) {
                            data.addEntry({
                                label: AJS.escapeHtml(term.replace(/\$col\b/g, column.name)),
                                tip: "sql",
                                sql: term,
                                replaceColName: "$col"
                            });
                        }
                        if (term.indexOf("=") == 0) {
                            if ($.trim(term) != "=") {
                                data.addEntry({
                                    label: AJS.escapeHtml("$col " + term),
                                    tip: "sql",
                                    sql: "$col " + term,
                                    replaceColName: "$col"
                                });
                            }
                        } else {
                            /* Don't suggest a sql literal. Rather improve the filter.
                             data.addEntry({
                             label: AJS.escapeHtml(term),
                             tip: "sql literal",
                             sql: term
                             });*/
                        }
                    }
                } else if (term.length == 0) {
                    data.addEntry({
                        label: "AND",
                        tip: "keyword",
                        type: "strict",
                        sql: " AND "
                    });
                    data.addEntry({
                        label: "OR",
                        tip: "keyword",
                        type: "strict",
                        sql: " OR "
                    });
                }
            },
            autocompleteFilterNumber: function(column, term, data){
                var charAt0 = /^[=<>]/.test(term) && term.charAt(0),
                    potentialNumber = charAt0 && term.substr(1) || term,
                    number = /[.,][0-9]/.test(potentialNumber) ? parseFloat(potentialNumber) : parseInt(potentialNumber);
                if (/[ -.]*[0-9].*-.*/.test(term)) { // Two numbers?
                    // Dodge the first minus, in the case of "-3.14--3.13"
                    var position = term.indexOf("-", 1),
                        exprA = term.substr(0, position),
                        exprB = term.substr(position + 1),
                        numberA = /[.,][0-9]/.test(exprA) ? parseFloat(exprA) : parseInt(exprA),
                        numberB = /[.,][0-9]/.test(exprB) ? parseFloat(exprB) : parseInt(exprB);
                    if (!isNaN(numberA)) {
                        if (isNaN(numberB)) {
                            numberB = numberA + 1;
                        }
                        data.addEntry({
                            label: numberA + "&ndash;" + numberB,
                            sql: "$col BETWEEN ? AND ?",
                            tip: "between",
                            replaceColName: "$col",
                            arguments: [numberA, numberB]
                        });
                    }
                } else if (!isNaN(number)) { // One number?
                    if (!charAt0 || charAt0 == "=")
                        data.addEntry({
                            label: "= " + number,
                            sql: "$col = ?",
                            tip: "equals",
                            replaceColName: "$col",
                            arguments: [number]
                        });

                    if (!charAt0 || charAt0 == ">")
                        data.addEntry({
                            label: "&gt; " + number,
                            sql: "$col > ?",
                            tip: "min",
                            replaceColName: "$col",
                            arguments: [number]
                        });

                    if (!charAt0 || charAt0 == "<")
                        data.addEntry({
                            label: "&lt; " + number,
                            sql: "$col < ?",
                            tip: "max",
                            replaceColName: "$col",
                            arguments: [number]
                        });
                    data.addEntry({
                        label: number + "—" + (number+1), // endash: – emdash: —
                        sql: "$col BETWEEN ? AND ?",
                        tip: "between",
                        replaceColName: "$col",
                        arguments: [number, number+1]
                    });
                } else {
                    if (!term || charAt0 == "=")
                        data.addEntry({
                            label: "= 10",
                            sql: "$col = 10",
                            tip: "equals"
                        });
                    if (!term || charAt0 == "<")
                        data.addEntry({
                            label: "< 10",
                            sql: "$col < 10",
                            replaceColName: "$col",
                            tip: "max"
                        });
                    if (!term || charAt0 == ">")
                        data.addEntry({
                            label: "> 10",
                            sql: "$col > 10",
                            replaceColName: "$col",
                            tip: "min"
                        });
                    if (!term)
                        data.addEntry({
                            label: "10&ndash;20",
                            sql: "$col between 10 and 20",
                            replaceColName: "$col",
                            tip: "range"
                        });
                }
            },
            autoCompleteFilterString: function(column, term, data){
                if ($.trim(term).length > 0) {
                    data.addEntry({
                        label: "= " + AJS.escapeHtml(term),
                        tip: "equals",
                        sql: "UPPER($col) = UPPER(?)",
                        replaceColName: "$col",
                        arguments: [term]
                    });
                    if (term.indexOf("%") != -1) {
                        data.addEntry({
                            label: "LIKE '" + AJS.escapeHtml(term) + "'",
                            sql: "$col LIKE ?",
                            tip: "sql",
                            replaceColName: "$col",
                            arguments: [term]
                        });
                    } else {
                        data.addEntry({
                            label: "%" + AJS.escapeHtml(term) + "%",
                            tip: "contains",
                            sql: "$col LIKE '%' || ? || '%'",
                            replaceColName: "$col",
                            arguments: [term]
                        });
                    }
                }
            },
            autoCompleteFilterDate: function(column, term, data){
                var spreadsheet = Confluence.PlaySQL.Spreadsheet.byPlaceholder(column.$placeholder);
                if (spreadsheet.dialect != 'com.playsql.jdbc.dialect.PostgreSQLDDLDialect') {
                    data.addEntry({
                        label: "Dates are only supported for Postgres",
                        tip: "information",
                        sql: "",
                        type: "strict"
                    });
                    return;
                }
                var operator = /^[<>=]/.test(term) && term.charAt(0),
                    potentialValue = operator && term.substr(1) || term,
                    date, yyyymmdd;

                if ($.trim(potentialValue).length > 0) {
                    date = new Date(potentialValue);
                    if (!date || isNaN(date.getTime())) {
                        date = new Date();
                    }
                } else {
                    date = new Date();
                }
                var yyyy = date.getFullYear().toString();
                var mm = (date.getMonth()+1).toString();
                var dd  = date.getDate().toString();
                yyyymmdd = yyyy + "-" + (mm[1]?mm:"0"+mm[0]) + "-" + (dd[1]?dd:"0"+dd[0]);
                if (!operator || operator == "<")
                    data.addEntry({
                        label: "< " + date.toDateString(),
                        tip: "max",
                        sql: "$col <= ?",
                        replaceColName: "$col",
                        arguments: [yyyymmdd]
                    });
                if (!operator || operator == ">")
                    data.addEntry({
                        label: "> " + date.toDateString(),
                        tip: "min",
                        sql: "$col >= ?",
                        replaceColName: "$col",
                        arguments: [yyyymmdd]
                    });
                if (!operator || operator == "=")
                    data.addEntry({
                        label: "= " + date.toDateString(),
                        tip: "equals",
                        sql: "$col = ?",
                        replaceColName: "$col",
                        arguments: [yyyymmdd]
                    });
            },
            autoCompleteFilterList: function(column, term, data, list){
                for (var i in list) {
                    var word = list[i];
                    if (word.indexOf(term) != -1) {
                        data.addEntry({
                            label: AJS.escapeHtml(word),
                            tip: "value",
                            sql: "$col = ?",
                            replaceColName: "$col",
                            arguments: [word]
                        });
                    }
                }
            },
            buildWhereClause : function (whereClause, whereClauseColumnFilters, dialect) {
                var whereClause2 = "";
                if (whereClauseColumnFilters) {
                    function quoteColumnName(dialect, columnName) {
                        if (!!columnName) {
                            if (columnName.toLowerCase() != columnName) {
                                return '"' + columnName + '"';
                            }
                        }
                        return columnName;
                    }
                    for(var columnName in whereClauseColumnFilters) {
                        var filters = whereClauseColumnFilters[columnName],
                            quotedColumnName = quoteColumnName(dialect, columnName),
                            columnSQL = "",
                        // if strict, the next operand won't try to insert a conjunction like OR/AND
                            strict = true,
                            multiple = false;
                        if (filters) {
                            if (filters.length > 1) multiple = true;
                            for (var i in filters) {
                                var detail = filters[i],
                                    sql = detail.sql;
                                if (detail.arguments) {
                                    for(var j in detail.arguments) {
                                        var value = detail.arguments[j];
                                        if (typeof value != "number") {
                                            value = "'" + Confluence.PlaySQL.Helpers.escapeSqlString(value) + "'";
                                        }
                                        sql = sql.replace("?", value);
                                    }
                                }
                                if (detail.replaceColName) {
                                    sql = sql.replace(/\$col\b/g, quotedColumnName);
                                }
                                if (!strict && detail.type != "strict") {
                                    columnSQL += " OR ";
                                }
                                columnSQL += sql;
                                strict = detail.type == "strict";
                            }
                        }
                        // If there's at least one condition, append it to the whereClause2 with AND
                        if (columnSQL != "") {
                            if (multiple) {
                                whereClause2 += " AND (" + columnSQL + ")";
                            } else {
                                whereClause2 += " AND " + columnSQL;
                            }
                        }
                    }
                }
                if (whereClause && whereClause2.indexOf(" AND ") == 0) {
                    return "(" + whereClause + ")" + whereClause2;
                } else if (whereClause) {
                    return whereClause;
                } else {
                    return whereClause2.substr(4);
                }
            },
            create: function (spreadsheet, column, $header) {
                // Check if there's already an inline dialog for this column
                if (spreadsheet.inlinedialog) {
                    if (spreadsheet.inlinedialog.column == column) {
                        return spreadsheet.inlinedialog;
                    }
                }

                var orderbyClause = Confluence.PlaySQL.Helpers.parseOrderByClause(spreadsheet.orderbyClause),
                    orderbyDef = {},
                    renderer = Confluence.PlaySQL.Spreadsheet.renderers[column.renderer],
                    viewRenderer = Confluence.PlaySQL.Spreadsheet.renderers[column.viewRenderer || column.renderer];

                for (var i in orderbyClause) {
                    if (orderbyClause[i].field == '"' + column.name + '"') {
                        orderbyDef = orderbyClause[i];
                        break;
                    }
                }
                column.rendererName = renderer ? renderer.getI18nName() : "?";
                column.viewRendererName = viewRenderer ? viewRenderer.getI18nName() : "?";

                var inlinedialog = AJS.InlineDialog($header, "col-properties",
                    function ($content, trigger, showPopup) {
                        $content.closest("#inline-dialog-col-properties").addClass("small-inline");
                        $content.css("width", "");
                        $content.html(Confluence.Templates.PlaySQL.columnPropertiesInlineDialog({
                            column: column,
                            orderbyDef: orderbyDef,
                            hasColFEPerm: spreadsheet.perms.hasPerm('EDIT_COLS_FE'),
                            hasColBEPerm: spreadsheet.perms.hasPerm('EDIT_COLS_BE')
                        }));
                        showPopup();
                        $content.find(".sort-none, .sort-desc, .sort-asc").click(function (e) {
                            e.preventDefault();
                            var $this = $(this),
                                way = $this.hasClass("sort-desc") ? "DESC"
                                    : $this.hasClass("sort-asc") ? "ASC"
                                    : "NO-SORTING",
                                nulls = orderbyDef.nulls;
                            spreadsheet.sortCol($header, way, nulls);
                            spreadsheet.inlinedialog && spreadsheet.inlinedialog.close();
                            return false;
                        });
                        $content.find(".nulls-default, .nulls-first, .nulls-last").click(function (e) {
                            e.preventDefault();
                            var $this = $(this),
                                way = orderbyDef.way,
                                nulls = $this.hasClass("nulls-default") ? undefined
                                    : $this.hasClass("nulls-first") ? "FIRST"
                                    : "LAST";
                            spreadsheet.sortCol($header, way, nulls);
                            spreadsheet.inlinedialog && spreadsheet.inlinedialog.close();
                            return false;
                        });
                        $content.on("click", ".btn-properties", function (e) {
                            e.preventDefault();
                            Confluence.PlaySQL.Spreadsheet.ColumnToolbar.displayProperties(spreadsheet, column, inlinedialog);
                        });/*
                         $content.on("click", ".btn-properties1", function (e) {
                         e.preventDefault();
                         spreadsheet.inlinedialog.close();
                         displayColProps(spreadsheet, $header);
                         });*/

                        var allFilterAutocompleteData = [],
                            initialValue = [];
                        // Initialize with the spreadsheet's values
                        if (spreadsheet.whereClauseColumnFilters && spreadsheet.whereClauseColumnFilters[column.name]) {
                            var filters = spreadsheet.whereClauseColumnFilters[column.name];
                            for(var i in filters) {
                                var filter = filters[i];
                                allFilterAutocompleteData.push(filter);
                                initialValue.push(i);
                            }
                        }

                        function formatResult(state) {
                            if (!state.id) return state.text; // optgroup
                            var complement = allFilterAutocompleteData[state.id].tip;
                            if (complement) {
                                return allFilterAutocompleteData[state.id].label + '<span class="playsql-filter-tip">' + complement + '</span>';
                            } else {
                                return allFilterAutocompleteData[state.id].label;
                            }
                        }

                        function formatSelection(state) {
                            if (!state.id) return state.text; // optgroup
                            return allFilterAutocompleteData[state.id].label;
                        }

                        $content.find("#playsql-inline-filter5").auiSelect2({
                            multiple: true,
                            placeholder: "Filter",
                            formatResult: formatResult,
                            formatSelection: formatSelection,
                            escapeMarkup: function (m) {
                                return m;
                            },
                            formatNoMatches: function(term) {
                                return "Enter a filter"
                            },
                            query: function (query) {
                                var term = query.term,
                                    data = {
                                        results: [],
                                        addEntry: function (json) {
                                            allFilterAutocompleteData.push(json);
                                            this.results.push({id: allFilterAutocompleteData.length - 1, text: json.label});
                                            AJS.log("Added entry: ", allFilterAutocompleteData[allFilterAutocompleteData.length - 1]);
                                        }
                                    };
                                if (viewRenderer && viewRenderer.autoCompleteFilter) {
                                    viewRenderer && viewRenderer.autoCompleteFilter(spreadsheet, column, term, data, function () {
                                        query.callback(data);
                                    });
                                } else {
                                    data.addEntry({
                                        label: "No filter available for " + column.viewRendererName,
                                        sql: undefined
                                    });
                                }

                                query.callback(data);
                            },
                            initSelection: function($element, callback){
                                var data = [];
                                $($element.val().split(",")).each(function(){
                                    if (this != "" && typeof this != "undefined") {
                                        data.push({id: this, text: "-loading #" + this});
                                    }
                                });
                                if (data.length > 0) {
                                    $element.parent().addClass("hovered");
                                }
                                callback(data);
                            }
                        }).auiSelect2("val", initialValue)
                            .on("change", function(){
                                var values = $content.find("#playsql-inline-filter5").auiSelect2("val").map(function(id){
                                    return allFilterAutocompleteData[id];
                                });
                                $("#inline-dialog-col-properties .error").remove();
                                spreadsheet.whereClauseColumnFilters = spreadsheet.whereClauseColumnFilters || {};
                                var initialFilter = spreadsheet.whereClauseColumnFilters[column.name];
                                spreadsheet.whereClauseColumnFilters[column.name] = values;
                                spreadsheet.load(undefined, {failureCallback: function(jqXHR){
                                    if (Math.round(jqXHR.status/100) * 100 == 400) {
                                        try {
                                            // If the BAD_REQUEST is specially formatted by Play SQL
                                            var data = $.parseJSON(jqXHR.responseText);
                                            var message;
                                            if (data.type == "SQL_SYNTAX") {
                                                message = data.message || "Error";// + "\n" + data.sql;
                                            } else {
                                                message = data.message || "Error";
                                            }
                                            $("#inline-dialog-col-properties .aui-buttons").last().after("<br><span class=\"error\"></span><br>&nbsp;");
                                            $("#inline-dialog-col-properties .error").text(message);
                                            return false;
                                        } catch (err) {
                                            // Swallow and proceed to the next section
                                            AJS.log(err);
                                        }
                                    }
                                }});
                            });


                        $content.find(".filter-input5").mouseenter(function () {
                            $(this).parent().addClass("hovered");
                            /*$content.find(".inline-filter5-on-hover-display").removeClass("hidden");
                             // Allow for the time for the fade-in effect
                             window.setTimeout(function(){$content.find(".inline-filter5-on-hover-display").addClass("hovered");}, 0);*/
                        });
                        $content.find("#apply-inline-filter").click(function(e){
                            e.preventDefault();
                            var values = $content.find("#playsql-inline-filter5").auiSelect2("val").map(function(id){
                                return allFilterAutocompleteData[id];
                            });
                            spreadsheet.whereClauseColumnFilters = spreadsheet.whereClauseColumnFilters || {};
                            spreadsheet.whereClauseColumnFilters[column.name] = values;
                            spreadsheet.inlinedialog.close();
                            spreadsheet.load();
                        });
                        $content.on("click", ".btn-cancel", function (e) {
                            e.preventDefault();
                            spreadsheet.inlinedialog.close();
                        });
                        var $formulaField = $content.find(".column-formula");
                        $formulaField.blur(function () {
                            window.setTimeout(function () {
                                $formulaField.removeClass("aui-dropdown2-trigger")
                                    .removeAttr('aria-controls aria-haspopup aria-owns');
                            }, 100);
                        }).keyup(function (e) {
                            window.setTimeout(function () {
                                if ($formulaField.val() == "=") {
                                    if (!$formulaField.attr("aria-owns")) {
                                        $formulaField
                                            .attr('aria-owns', 'dropdown-wizards')
                                            .attr('aria-haspopup', 'true')
                                            .addClass('aui-dropdown2-trigger')
                                            .trigger("aui-button-invoke");
                                    }
                                } else {
                                    if ($("#dropdown-wizards:visible").length > 0) {
                                        $formulaField.trigger("aui-button-invoke");
                                    }
                                    $formulaField
                                        .attr('aria-owns', '')
                                        .attr('aria-haspopup', '')
                                        .removeClass('aui-dropdown2-trigger');
                                }
                            }, 0);
                        });

                        return false;
                    }, {
                        noBind: true,
                        persistent: true,
                        hideCallback: function () {
                            spreadsheet.inlinedialog && spreadsheet.inlinedialog.close();
                        },
                        hideDelay: 10000000/*,
                         gravity: "n",
                         onTop: true*/
                    }
                );
                inlinedialog.column = column;
                inlinedialog.close = function(){
                    inlinedialog.remove();
                    $(".tipsy.tipsy-n, #null-sorting-dropdown").remove();
                    $("body, .playsql-spreadsheet-placeholder th").off(".playsql-inlinedialog");
                    if (spreadsheet.inlinedialog == this) {
                        spreadsheet.inlinedialog = undefined;
                    }
                }
                var onClickOutsideInlineDialog = function (e) {
                    if (!spreadsheet.inlinedialog || !spreadsheet.inlinedialog.is(":visible")) return;
                    var target = e.target || this;
                    var $this = $(target);
                    if ($this.closest("#inline-dialog-col-properties, " +
                        "#null-sorting-dropdown, " +
                        "#select2-drop, #select2-drop-mask, #select2-drop-mask").size() > 0) return;
                    // If it's not in the DOM, it was detached, therefore it's probably the Configure button
                    if (!$.contains(document.documentElement, target)) return;
                    spreadsheet.inlinedialog.close();
                };
                $("body").on("click.playsql-inlinedialog", "*", onClickOutsideInlineDialog);
                // For elements which answer 'false' to clicks:
                $(".playsql-spreadsheet-placeholder th").on("click.playsql-inlinedialog", onClickOutsideInlineDialog);
                return inlinedialog;
            }, // End of the 'create' function
            displayProperties : function(spreadsheet, column, inlinedialog){
                $(".tipsy.tipsy-n").remove();
                // We brutally replace the inline dialog's contents with the new form
                var template = Confluence.Templates.PlaySQL.columnPropertiesDetails({
                    column: column,
                    hasColFEPerm: spreadsheet.perms.hasPerm('EDIT_COLS_FE'),
                    hasColBEPerm: spreadsheet.perms.hasPerm('EDIT_COLS_BE')
                });
                var $content = $("#inline-dialog-col-properties .contents").empty().html(template),
                    $header = column.$placeholder.find("th:nth-child(" + (column.index + 1) + ")");
                $(".btn-advanced").click(function(e){
                    e.preventDefault();
                    $(".remove-when-show-advanced-details").remove();
                    Confluence.PlaySQL.Spreadsheet.ColumnToolbar.displayAdvancedProperties(spreadsheet, column, inlinedialog);
                });

                // Select the default quick renderer, if any
                var bestGuess = getQuickRendererBestGuess(column);
                if (bestGuess != undefined) {
                    $content.find("[data-quick-renderer='" + bestGuess + "']").addClass("active");
                }

                if (!spreadsheet.perms.hasPerm('EDIT_COLS_FE')) {
                    $content.find(".quick-renderer").remove();
                } else {
                    $content.find(".quick-renderer").each(function (i) {
                        var $this = $(this),
                            action = $this.attr("data-quick-renderer"),
                            proposition = getProposedRenderer(action, column),
                            renderer = Confluence.PlaySQL.Spreadsheet.renderers[proposition.renderer];
                        if (!renderer) {
                            $this.remove();
                        }
                        if (!renderer.isSupportedType(column.type)) {
                            if (!spreadsheet.perms.hasPerm('EDIT_COLS_BE')) {
                                $this.remove();
                            }
                        }
                    });
                }
                // Remove intermediary groups of button, to avoid the empty-space effect
                $content.find(".aui-buttons:empty").remove();
                if (!Confluence.PlaySQL.Helpers.hasMinVersion("5.5.0")) $content.find(".aui-iconfont-calendar").parent().text("D");
                $content.on("click", ".quick-renderer", function (e) {
                    e.preventDefault();
                    var $this = $(this),
                        $parent = $this.closest(".aui-toolbar2");
                    $parent.find(".quick-renderer").removeClass("active");
                    $this.addClass("active enforce");
                });
                function saveColumnFormula(formula, viewRenderer) {
                    var changes = [],
                        newColumn = {
                            name: column.name,
                            label: column.label,
                            type: column.type,
                            formula: column.formula,
                            renderer: column.renderer,
                            viewRenderer: column.viewRenderer,
                            length: column.length,
                            width: column.width,
                            formats: column.formats
                        };
                    if (formula != newColumn.formula) {
                        changes.push("CHFMA");
                        newColumn.formula = formula;
                    }
                    if (viewRenderer != newColumn.viewRenderer) {
                        changes.push("CHRDR");
                        newColumn.viewRenderer = viewRenderer;
                    }
                    if (changes.length == 0) return;

                    spreadsheet.saveColumn($header, column, newColumn, changes, function () {
                        spreadsheet.load();
                    }, function (message) {
                        Confluence.PlaySQL.Messages.addMessage("error", "The wizard couldn't update the formula. The spreadsheet will be reloaded", "wizard-formula", message);
                        spreadsheet.load();
                    });
                }

                function saveColumnFormulaFailure(message) {
                    Confluence.PlaySQL.Messages.addMessage("error", "The wizard couldn't update the formula.", "wizard-formula", message);
                }

                function submitColumnPopupCallback() {
                    var $label = $content.find(".column-label"),
                        $name = $content.find(".column-name"),
                        $formula = $content.find(".column-formula"),
                        $type = $content.find("#column-type:visible"),
                        $renderer = $content.find("#column-renderer:visible"),
                        $viewRenderer = $content.find("#view-renderer:visible"),
                        $quickRenderer = $content.find(".quick-renderer.active.enforce").first(),
                        $length = $content.find("#column-length:visible"),
                        changes = [],
                        hasFormula = $formula.val(),
                        newColumn = {
                            name: column.name,
                            label: column.label,
                            type: column.type,
                            formula: column.formula,
                            renderer: column.renderer,
                            length: column.length,
                            width: column.width,
                            formats: undefined
                        };
                    if ($label.size() > 0 && $label.val() != column.label) {
                        newColumn.label = $label.val();
                        if ($.trim(newColumn.label).length == 0) {
                            $("#inline-dialog-col-properties .error").last().text("A label is required.");
                            return;
                        }
                        changes.push("CHLBL");
                    }
                    if ($name.size() > 0 && $name.val() != column.name) {
                        newColumn.name = $.trim($name.val());
                        if (newColumn.name.length == 0) {
                            $("#inline-dialog-col-properties .error").last().text("A name is required. This is the key of the column.");
                            return;
                        }
                        if (spreadsheet.colOf(newColumn.name) != undefined) {
                            $("#inline-dialog-col-properties .error").last().text("Another column already uses this name.");
                            return;
                        }
                        changes.push("RENAME");
                    }
                    if ($formula.val() != (column.formula || "")) {
                        changes.push("CHFMA");
                        newColumn.formula = $formula.val();
                    }
                    if ($quickRenderer.size() == 1) {
                        // When we update with a quick renderer, we reuse the existing format details.
                        // We only reset format details if the "Advanced" panel was shown and quick renderers removed.
                        newColumn.formats = column.formats;
                        var action = $quickRenderer.attr("data-quick-renderer"),
                            proposition = getProposedRenderer(action, column);
                        var renderer = Confluence.PlaySQL.Spreadsheet.renderers[proposition.renderer];
                        // We don't change the backend type:
                        // - If there's a formula. It's not a problem since we'll detect a potential type mismatch soon enough.
                        // - If the current type is compatible with the proposed renderer (unless the proposition requires enforcing the type)
                        if (hasFormula) {
                            // We change the renderer of the formula
                            changes.push("CHRDR");
                            newColumn.viewRenderer = proposition.renderer;
                            if (proposition.formats) {
                                newColumn.formats = $.extend({}, newColumn.formats, proposition.formats);
                                changes.push("CHFMT");
                            }
                        } else if (renderer.isSupportedType(column.type) && (!proposition.updateType || !spreadsheet.perms.hasPerm('EDIT_COLS_BE'))) {
                            // We just change the renderer of the column
                            changes.push("CHRDR");
                            newColumn.renderer = proposition.renderer;
                            if (proposition.formats) {
                                newColumn.formats = $.extend({}, newColumn.formats, proposition.formats);
                                changes.push("CHFMT");
                            }
                        } else if (spreadsheet.perms.hasPerm('EDIT_COLS_BE')) {
                            // We change the back-end type of the column
                            changes.push("CHTYPE");
                            changes.push("CHRDR");
                            newColumn.type = renderer.getDefaultType();
                            newColumn.length = undefined;
                            newColumn.renderer = proposition.renderer;
                            if (proposition.formats) {
                                newColumn.formats = $.extend({}, newColumn.formats, proposition.formats);
                                changes.push("CHFMT");
                            }
                        } else {
                            // The quick renderer shouldn't have been displayed anyway
                            message = "Unsupported type: " + column.type + " and can't change type to " + renderer.getDefaultType();
                            Confluence.PlaySQL.Helpers.checkNotNull($("#inline-dialog-col-properties .error").last()).text(message);
                            return;
                        }
                    } else {
                        // We only look at the type/renderer/length/formatDetail fields if no Quick Renderer
                        // was selected
                        if ($type.size() > 0 && $type.val() != newColumn.type) {
                            changes.push("CHTYPE");
                            newColumn.type = $type.val();
                        }
                        if ($renderer.size() > 0 && $renderer.val() != newColumn.renderer) {
                            changes.push("CHRDR");
                            newColumn.renderer = $renderer.val();
                        }
                        if ($viewRenderer.size() > 0) {
                            newColumn.viewRenderer = $viewRenderer.val();
                            if (newColumn.viewRenderer == newColumn.renderer) newColumn.viewRenderer = undefined;
                            if (newColumn.viewRenderer == "") newColumn.viewRenderer = undefined;
                            if (newColumn.viewRenderer != column.viewRenderer) {
                                changes.push("CHRDR");
                                newColumn.viewRenderer = $viewRenderer.val();
                            }
                        }
                        if ($length.size() > 0 && $length.val() != newColumn.length) {
                            newColumn.length = parseInt($length.val());
                            if (isNaN(newColumn.length)) newColumn.length = undefined;
                            if (newColumn.length == undefined) {
                                $content.find(".for-length .error").last().text("The length is mandatory.");
                                return;
                            }
                            if (newColumn.length != column.length) changes.push("CHTYPE");
                        }
                        var newRenderer = Confluence.PlaySQL.Spreadsheet.renderers[newColumn.viewRenderer || newColumn.renderer];
                        var oldFormats = column.formats,
                            newFormats = newRenderer.getFormatValues($content, newColumn);

                        function isEqual(newFormats, oldFormats) {
                            if ((newFormats == undefined) ^ (oldFormats == undefined)) {
                                return false;
                            } else if (newFormats != undefined && oldFormats != undefined) {
                                if (Object.keys(newFormats).length != Object.keys(oldFormats).length) {
                                    return false;
                                } else {
                                    for (var key in newFormats) {
                                        if (newFormats[key] != oldFormats[key]) {
                                            return false;
                                        }
                                    }
                                }
                            }
                            return true;
                        }

                        if (!isEqual(newFormats, oldFormats)) {
                            changes.push("CHFMT");
                        }
                        newColumn.formats = newFormats;
                    }

                    if (changes.length == 0) {
                        spreadsheet.inlinedialog && spreadsheet.inlinedialog.close();
                        return;
                    }

                    spreadsheet.saveColumn($header, column, newColumn, changes, function () {
                        // Remove cache. Sorry for breaking encapsulation.
                        column.statusMappings = undefined;
                        spreadsheet.renderAllColumn(column);
                        spreadsheet.inlinedialog && spreadsheet.inlinedialog.close();
                    }, function (message) {
                        Confluence.PlaySQL.Helpers.checkNotNull($("#inline-dialog-col-properties .error").last()).text(message);
                    });
                    return;
                }

                $content.find(".wizardGMaps").click(function () {
                    Confluence.PlaySQL.Spreadsheet.ColumnToolbar.wizardGMaps(spreadsheet, $header, column, saveColumnFormula, saveColumnFormulaFailure);
                });
                $content.find(".wizardForeignKey").click(function () {
                    Confluence.PlaySQL.Spreadsheet.ColumnToolbar.wizardForeignKey(spreadsheet, $header, column, saveColumnFormula, saveColumnFormulaFailure);
                });
                $content.on("click", ".btn-delete", function (e) {
                    e.preventDefault();
                    if (!spreadsheet.perms.hasPerm('EDIT_COLS_BE')) {
                        alert("Permissions are required to delete columns");
                        return;
                    }
                    spreadsheet.inlinedialog.close();
                    spreadsheet.deleteColumn($header);
                });
                $content.on("click", ".btn-submit", submitColumnPopupCallback);
                $content.on("keypress", function (e) {
                    if (e.which == 13) {
                        submitColumnPopupCallback();
                        return false;
                    } else if (e.which == 27) {
                        spreadsheet.inlinedialog && spreadsheet.inlinedialog.close();
                        return false;
                    }
                });
                $content.find(".playsql-tipsy").tipsy();
            },
            displayAdvancedProperties : function(spreadsheet, column, inlinedialog) {
                var hasFormula = !!column.formula,
                    renderers = getRenderers(column.type),
                    $template = $(Confluence.Templates.PlaySQL.columnPropertiesAdvanced({
                        name: column.name,
                        hasFormula: hasFormula,
                        renderer: column.renderer,
                        viewRenderer: column.viewRenderer || column.renderer,
                        coltype: column.type,
                        length: column.length,
                        renderers: renderers,
                        dataTypes: DATA_TYPES,
                        canChangeBackEndCols : spreadsheet.perms.hasPerm('EDIT_COLS_BE')
                    }));
                $("#inline-dialog-col-properties .contents .advanced-properties > tbody").append($template);

                function updateFieldType() {
                    var $select = $("#column-type"),
                        val = $select.val(),
                        $message = $select.parent().children(".description"),
                        $editRendererCombo = $("#column-renderer"),
                        editRenderer = column.renderer;

                    if (val != column.type) {
                        var text = ("The original type is {}. If you save, the column will be transformed to {}. The conversion" +
                        " is up to the database system, not Play SQL. You may or may not be satisfied with the new values.")
                            .replace("{}", column.type).replace("{}", val);
                        $message.text(text);
                    } else {
                        $message.empty();
                    }

                    for (var i in DATA_TYPES) {
                        if (DATA_TYPES[i].name == val) {
                            $(".for-length").toggleClass("hidden", !DATA_TYPES[i].hasLength);
                            break;
                        }
                    }

                    // Update the list of renderers in the combo
                    renderers = getRenderers(val);
                    $editRendererCombo.empty();
                    for (var i in renderers.sameTypeRenderers) {
                        var el = renderers.sameTypeRenderers[i];
                        $("<option>", {
                            value: el.key,
                            text: el.name,
                            selected: editRenderer == el.key
                        }).appendTo($editRendererCombo);
                    }
                    updateRenderer();
                }

                function updateRenderer() {
                    var $select = hasFormula ? $("#view-renderer") : $("#column-renderer"),
                        val = $select.val(),
                        $placeholderTable = $("#inline-dialog-col-properties .advanced-properties > tbody");
                    renderer = Confluence.PlaySQL.Spreadsheet.renderers[val];

                    $("#inline-dialog-col-properties .format-details").remove();

                    renderer.appendRendererDetailsTo(spreadsheet, $placeholderTable, column);
                }

                $("#column-type").change(updateFieldType);
                $(hasFormula ? "#view-renderer" : "#column-renderer").change(updateRenderer);

                updateFieldType();
            }
        } // End of Confluence.PlaySQL.Spreadsheet.ColumnToolbar

        var DATA_TYPES = [{name:'VARCHAR', hasLength: true},
            {name:'TEXT'},
            {name:'INTEGER'},
            {name:'BIG_INTEGER'},
            {name:'BOOLEAN'},
            {name:'DATE'},
            {name:'NUMERIC'},
            {name:'DOUBLE'}];

        function displayColProps(spreadsheet, $cell) {
            if (!spreadsheet.perms.hasPerm('EDIT_COLS_FE')) return;
            var column = spreadsheet.colOf($cell),
                dialog = Confluence.PlaySQL.Spreadsheet.ColumnProperties(spreadsheet, column, spreadsheet.perms.hasPerm('EDIT_COLS_BE'));
            dialog.show();
        }


        var QUICK_RENDERERS = {
            "euro":   { renderer: "currency", formats: {aboveZero: "# ##0,00 €"}},
            "dollar": { renderer: "currency", formats: {aboveZero: "$ #,##0.00"}},
            "integer": { renderer: "integer"},
            "decimal": { renderer: "currency", formats: {aboveZero: "#,##0.00"}, updateType: true},
            "markdown": { renderer: "markdown"},
            "text": { renderer: "as-string"},
            "link": { renderer: "link"},
            "boolean": { renderer: "boolean"},
            "date-picker": { renderer: "date"},
            "user-picker": { renderer: "confluenceuser"},
            "date-picker": { renderer: "date"},
            "page-picker": { renderer: "confluencecontent"}
        };

        function getQuickRendererBestGuess(column) {
            var action;
            for (action in QUICK_RENDERERS) {
                var quickRenderer = QUICK_RENDERERS[action],
                    matches = true;
                if (quickRenderer.renderer != (column.viewRenderer || column.renderer)) continue;
                for (var formatKey in quickRenderer.formats) {
                    if (column.formats == null || column.formats[formatKey] != quickRenderer.formats[formatKey]) {
                        matches = false;
                        break;
                    }
                }
                if (matches) {
                    return action;
                }
            }
            if ((column.viewRenderer || column.renderer) == "currency") {
                action = "integer";
                return action;
            }
            return undefined;
        }

        function getProposedRenderer(action, column) {
            var result = QUICK_RENDERERS[action];

            if (action == "integer") {
                if (column.type == "NUMERIC") {
                    result.renderer = "currency";
                    result.formats = {aboveZero: "#,##0"};
                } else {
                    result.renderer = "integer";
                }
            }
            return result;
        }

        function getRenderers(colType) {
            var sameTypeRenderers = [],
                otherTypeRenderers = [],
                renderers = Confluence.PlaySQL.Spreadsheet.renderers;
            for (var key in renderers) {
                if (renderers[key].isSystemRenderer()) continue;
                var r = {
                    key: key,
                    name: renderers[key].getI18nName()
                }
                if (renderers[key].isSupportedType(colType)) {
                    sameTypeRenderers.push(r);
                } else {
                    otherTypeRenderers.push(r);
                }
            }
            return {
                sameTypeRenderers : sameTypeRenderers,
                otherTypeRenderers : otherTypeRenderers
            }
        }

        Confluence.PlaySQL.Spreadsheet.ColumnToolbar.wizardForeignKey = function(spreadsheet, $header, column, successCallback, failureCallback) {

            var inlineDialog = AJS.InlineDialog($header, "wizard-foreign-key", function($content, trigger, showPopup) {
                $content.css("width", "450px");
                $content.html(Confluence.Templates.PlaySQL.wizardForeignKeyInlineDialog({}));

                $content.find(".btn-submit").click(function(){
                    var formula = $content.find(".preview").attr("data-result"),
                        targetColumnRenderer = $content.find("#targetlabel option:checked").attr("data-renderer");
                    if (!formula) {
                        $content.find(".error").last().text("All fields are mandatory");
                        return;
                    }
                    successCallback && successCallback(formula, targetColumnRenderer);
                    inlineDialog && inlineDialog.close();
                });
                showPopup();

                $content.find("select").change(function(){
                    function quoteIfNecessary(columnname) {
                        if (columnname == undefined) return columnname;
                        if (spreadsheet.defaultCase == "LOWER") {
                            if (columnname.toLowerCase() != columnname) {
                                columnname = '"' + columnname + '"';
                            }
                        } else if (spreadsheet.defaultCase == "UPPER") {
                            if (columnname.toUpperCase() != columnname) {
                                columnname = '"' + columnname + '"';
                            }
                        } else {
                            columnname = '"' + columnname + '"';
                        }
                        return columnname;
                    }
                    var targettable = $content.find("#targettable").val(),
                        targetlabel = $content.find("#targetlabel").val(),
                        targetid = $content.find("#targetid").val(),
                        sourcecolumn = $content.find("#sourcecolumn").val(),
                        formula = "= " + quoteIfNecessary(targettable) + "." + quoteIfNecessary(targetlabel)
                            + " WHERE " + quoteIfNecessary(targettable) + "." + quoteIfNecessary(targetid) + "=" + quoteIfNecessary(sourcecolumn),
                        $previewArea = $content.find(".preview");

                    if (!!targettable && !!targetlabel && !!targetid && !!sourcecolumn) {
                        $previewArea.text(formula).attr("data-result", formula);
                        $content.find(".error").last().text("");
                    } else {
                        $previewArea.html("&nbsp;").attr("data-result", "");
                    }
                });

                var url = AJS.contextPath() + "/rest/playsql/1.0/autocomplete/" + spreadsheet.context,
                    $selectTargetTable = $content.find("#targettable"),
                    $selectTargetLabel = $content.find("#targetlabel"),
                    $selectTargetId = $content.find("#targetid"),
                    $selectSourceColumn = $content.find("#sourcecolumn");

                // Load the table list
                $.ajax({
                    url: url,
                    type: 'GET',
                    contentType: "application/json; charset=utf-8"
                }).done(function(data){
                    var tables = data.tables["current schema"];
                    $selectTargetTable.empty();
                    $("<option>", {
                        value: "",
                        text: "-- Select a table to lookup"
                    }).appendTo($selectTargetTable);
                    for (var i in tables) {
                        $("<option>", {
                            value: tables[i].name,
                            text: tables[i].label
                        }).appendTo($selectTargetTable);
                    }
                }).fail(function(data){
                    failureCallback && failureCallback("Couldn't load the table list");
                });

                // Load the local columns
                $.ajax({
                    url: url + "/" + spreadsheet.id,
                    type: 'GET',
                    contentType: "application/json; charset=utf-8"
                }).done(function(data){
                    var tables = data.tables["current schema"],
                        tableName
                    $selectSourceColumn.empty();
                    for (var i in tables) {
                        if (tables[i].name == spreadsheet.id) {
                            var columns = tables[i].columns;
                            for (var j in columns) {
                                $("<option>", {
                                    value: columns[j].name,
                                    text: tables[i].label + "." + columns[j].label + (column.name == columns[j].name ? " (this column)" : "")
                                }).appendTo($selectSourceColumn);
                            }
                        }
                    }
                    $selectSourceColumn.val(column.name);
                }).fail(function(data){
                    failureCallback && failureCallback("Couldn't load the table list");
                });

                // Load the other columns
                $selectTargetTable.change(function(){
                    $.ajax({
                        url: url + "/" + $selectTargetTable.val(),
                        type: 'GET',
                        contentType: "application/json; charset=utf-8"
                    }).done(function(data){
                        var tables = data.tables["current schema"],
                            tableName
                        $selectTargetLabel.empty();
                        $selectTargetId.empty();
                        $("<option>", {
                            value: "",
                            text: "-- Select the column to display"
                        }).appendTo($selectTargetLabel);
                        $("<option>", {
                            value: "",
                            text: "-- Select the key"
                        }).appendTo($selectTargetId);
                        for (var i in tables) {
                            if (tables[i].name == $selectTargetTable.val()) {
                                var columns = tables[i].columns;
                                for (var j in columns) {
                                    $("<option>", {
                                        value: columns[j].name,
                                        text: tables[i].label + "." + columns[j].label,
                                        "data-renderer": columns[j].renderer
                                    }).appendTo($selectTargetLabel)
                                        .clone().appendTo($selectTargetId)
                                }
                            }
                        }
                        $selectTargetId.val("ID");
                    }).fail(function(data){
                        failureCallback && failureCallback("Couldn't load the table list");
                    });
                });
            }, {
                noBind: true,
                hideCallback: function(){
                    inlineDialog && inlineDialog.close();
                },
                hideDelay: 10000000
            });
            inlineDialog.close = function() {
                inlineDialog && inlineDialog.remove();
            };
            inlineDialog.show();
        }

        Confluence.PlaySQL.Spreadsheet.ColumnToolbar.wizardGMaps = function(spreadsheet, $header, column, successCallback, failureCallback) {
            if (column.type != 'VARCHAR' && column.type != 'TEXT') {
                var message = "Please change the type of the column to Link before proceeding.";
                failureCallback && failureCallback(message) || alert(message);
                return;
            }
            var inlineDialog = AJS.InlineDialog($header, "wizard-gmaps", function($content, trigger, showPopup) {
                $content.css("width", "450px");
                $content.html(Confluence.Templates.PlaySQL.wizardGMapsInlineDialog({
                    columns: spreadsheet.columns,
                    currentcolumn: column.name
                }));
                $content.find(".btn-submit").click(function(){
                    var sourceColumn = $content.find("#sourcecolumn").val(),
                        url = $content.find("#url").val(),
                        display = $content.find("input[name='display']:checked").attr("id"),
                        convertToSql = $content.find("#convert-to-sql").is(":checked"),
                        formula = "= TO_URL('" + url + "', " + sourceColumn;
                    if (display == "display-link") {
                        formula += ")";
                    } else if (display == "display-field") {
                        formula += ", " + sourceColumn + ")";
                    } else if (display == "display-label") {
                        formula += ", 'Link')";
                    }
                    successCallback(formula, 'link');
                    inlineDialog && inlineDialog.close();
                });
                showPopup();
            }, {
                noBind: true,
                hideCallback: function(){
                    inlineDialog && inlineDialog.close();
                },
                hideDelay: 10000000
            });
            inlineDialog.close = function() {
                inlineDialog && inlineDialog.remove();
            };
            inlineDialog.show();
        }
    });
})(AJS.$);