(function(){
window.Confluence = window.Confluence || {};
window.Confluence.PlaySQL = window.Confluence.PlaySQL || {};
window.Confluence.PlaySQL.Spreadsheet = window.Confluence.PlaySQL.Spreadsheet || {};
window.Confluence.PlaySQL.Spreadsheet.FormulaEditor = window.Confluence.PlaySQL.Spreadsheet.FormulaEditor || {

    /**
     * Displays the formula editor for this cell.
     * If the cell was in edit mode, cancel it first.
     */
    edit : function(spreadsheet, $cell, renderer, setValue) {
        var inlinedialog;
        if (!spreadsheet.perms.hasPerm('EDIT_ROWS_METADATA')) {
            // Should never reach there because checkBeforeEdit() should return 'no'
            return false;
        }
        if ($cell.hasClass("editing") && !$cell.hasClass("formula")) {
            renderer.cancelEdit();
        }

        // Keep the old value for cancel
        if (!$cell.hasClass("saving")) {
            $cell.attr('data-rendered', $cell.html());
        }

        $cell.addClass("editing formula");

        // Define the value. If the cell was in error, the last correct value is the text of the cell.
        // Otherwise it is provided through the parameters and it comes from the cellMetadata.
        if ($cell.hasClass("invalid") || $cell.hasClass("saving")) {
            // The cell is re-edited after an invalid state
            setValue = $cell.text();
            if (setValue.charAt(0) == "=") {
                setValue = setValue.substr(1);
            }
            $cell.removeClass("saving");
        } else {
            // The cell is edited for the first time.
            if (setValue == undefined) {
                setValue = "";
            } else if (setValue.charAt(0) == "=") {
                setValue = setValue.substr(1);
            }
            $cell.attr('data-cell-metadata-formula', "=" + setValue);
        }

        var selectionManager = PlaySQL.Selection.forTable(spreadsheet.$placeholder);

        // Add the input field
        var $input = $("<div>", {
                        contenteditable: true,
                        "class": 'formula-editor playsql-formula',
                        text: setValue
                    });
        $input.appendTo($cell.empty().append("=&nbsp;"));
        // Key movements while editing the formula
        $input.on("keydown", function(e){
            // This time it's Play SQL's bindings that we want to skip
            e.stopPropagation();
            Confluence.PlaySQL.Spreadsheet.FormulaEditor.saveSelection($input);
            // And we want to map a few keys
            var keyCombination = Confluence.PlaySQL.Helpers.getKeyCombinationDetail(e);
            if (keyCombination.key == "backspace") {
                var caret = Confluence.PlaySQL.Helpers.caretSelection($input[0]);
                if (caret == undefined) return;
                if (caret.position == 0 && caret.length == 0) {
                    e.preventDefault();
                    Confluence.PlaySQL.Spreadsheet.FormulaEditor.removeFormula(spreadsheet, $cell, renderer);
                    renderer.focus();
                }
            } else if (keyCombination.canonical == "up"
                || keyCombination.canonical == "down"
                || keyCombination.canonical == "enter") {
                e.preventDefault();
            }
            // We need to wait for the character to be inserted
            window.setTimeout(function() {
                var result = Confluence.PlaySQL.Spreadsheet.FormulaEditor.triggerKeyboardCellSelection(spreadsheet, $cell, $input, inlinedialog, e);
                if (result) {
                    Confluence.PlaySQL.Spreadsheet.FormulaEditor.triggerAutocompletion(spreadsheet, $cell, $input, inlinedialog, e);
                }
            }, 0);
        }).on("keypress", function(e) {
            // Note: Keypress is only generated on actual characters after keydown, not on ESC/ENTER
            // We want to dodgeskip Confluence's keyboard shortcuts while we're in there, especially 'c' for Create
            e.stopPropagation();
            Confluence.PlaySQL.Spreadsheet.FormulaEditor.saveSelection($input);
        }).on("input, keyup", function(e){
            Confluence.PlaySQL.Spreadsheet.FormulaEditor.saveSelection($input);
            Confluence.PlaySQL.Spreadsheet.FormulaEditor.highlightCellsAtCaret(spreadsheet, $input);
            debouncedTriggerSyntaxColoration();
            spreadsheet.inlinedialog && spreadsheet.inlinedialog.refresh();
        }).on("click", function(e) {
            Confluence.PlaySQL.Spreadsheet.FormulaEditor.saveSelection($input);
            Confluence.PlaySQL.Spreadsheet.FormulaEditor.highlightCellsAtCaret(spreadsheet, $input);
        });
        var debouncedTriggerSyntaxColoration = $.debounce(function(){
            if (!$input.is(":visible")) return;
            Confluence.PlaySQL.Spreadsheet.FormulaEditor.triggerSyntaxColoration($input, spreadsheet, inlinedialog);
        }, 500);

        // Create the inline dialog
        spreadsheet.inlinedialog && spreadsheet.inlinedialog.close && spreadsheet.inlinedialog.close();
        var javadoc = null;
        inlinedialog = AJS.InlineDialog($cell, "formula-editor",
            function ($content, trigger, showPopup) {
                $content.css("width", "");
                $content.closest("#inline-dialog-formula-editor").addClass("small-inline");
                $content.html(Confluence.Templates.PlaySQL.formulaEditor({
                    javadoc : javadoc
                }));
                showPopup();
                $(document).on("mouseenter.playsql-inlinedialog", "#inline-dialog-formula-editor .autocomplete-item", function(){
                    $(this).addClass("active").siblings().removeClass("active");
                }).on("click.playsql-inlinedialog", "#inline-dialog-formula-editor .autocomplete-item", function(){
                    var $item = $(this);
                    $input.focus();
                    Confluence.PlaySQL.Spreadsheet.FormulaEditor.insertAutocomplete(spreadsheet, $item);
                });
                $("#inline-dialog-formula-editor .btn-cancel").click(function(){
                    Confluence.PlaySQL.Spreadsheet.FormulaEditor.closeDialog(spreadsheet, true);
                });
                $("#inline-dialog-formula-editor .btn-submit").click(function(){
                    // Save the cell
                    var cell = spreadsheet.getCell($cell);
                    if (cell != null) {
                        cell.renderer.saveAndView();
                        spreadsheet.sprayRequestPipe();
                    }
                });

                return false;
            }, {
                noBind: true,
                persistent: true,
                hideCallback: function () {
                    if (spreadsheet.inlinedialog) {
                        if (!spreadsheet.inlinedialog.redisplay || !spreadsheet.inlinedialog.redisplay.isRedisplaying) {
                            spreadsheet.inlinedialog && spreadsheet.inlinedialog.close();
                        }
                    }
                },
                hideDelay: 10000000,
                gravity: "e"
            }
        );
        spreadsheet.inlinedialog = inlinedialog;
        inlinedialog.name = "formula-editor";
        inlinedialog.$input = $input;
        inlinedialog.append = function(formulaAddition) {
            var $field = $cell.find(".formula-editor").first();
            if (formulaAddition instanceof $) {
                var caret = Confluence.PlaySQL.Helpers.caretSelection($field[0], $field.data("saved-selection")),
                    range = caret.range;
                // Either the caret is in the middle of a subrule-value-cell, either something is selected,
                // either we insert at the end. We position the caret at the end of the insert.
                if (caret.length === 0) {
                    // If the cursor is over a cell reference, select the whole <span>
                    if (caret.position > 0) {
                        Confluence.PlaySQL.Helpers.setSelection(range, $field[0], caret.position - 1, "setStart");
                    }
                    var $parentRef = $(range.startContainer).closest(".subrule-value-cell");
                    if ($parentRef.size() == 1) {
                        //range = document.createRange();
                        range.setEndAfter($parentRef[0]);
                        // If the last character before the cursor is not alphanum or *,
                        // we'll consider the user has added a separator and we won't replace the last element
                        var lastChar = range.toString();
                        if (/[^a-zA-Z0-9*:\[\]]/.test(lastChar)) {
                            range.collapse(false);
                        } else {
                            range.setStartBefore($parentRef[0]);
                        }
                    } else if (!$(range.startContainer).closest("td").hasClass("editing")) {
                        // The caret isn't in the formula cell. Move it in.
                        range = document.createRange();
                        var node = $field[0].childNodes.length > 0 ? $field[0].childNodes[$field[0].childNodes.length - 1] : $field[0];
                        range.setStart(node, node.length);
                        range.setEnd(node, node.length);
                    } else {
                        // Cancel the selection modification we've made earlier
                        range.collapse(false);
                    }
                }
                range.deleteContents();
                range.insertNode(formulaAddition[0]);
                // Refocus
                $field.focus();
                var selection = window.getSelection();
                selection.removeAllRanges();
                range.collapse(false);
                selection.addRange(range);
                $field.data("saved-selection", range);
            } else if (typeof formulaAddition == "string") {
                $field.val($field.val() + formulaAddition);
                // Refocus
                $field.focus();
            } else {
                throw "inlinedialog.append(): Unknown object: " + formulaAddition;
            }
        }
        inlinedialog.setJavadoc = function(data) {
            javadoc = Confluence.Templates.PlaySQL.javadoc(data);
            inlinedialog.redisplay && inlinedialog.redisplay();
        }
        inlinedialog.close = function(){
            inlinedialog.remove();
            if (spreadsheet.inlinedialog == inlinedialog) spreadsheet.inlinedialog = undefined;
            selectionManager.options.cssClassMap = undefined;
            Confluence.PlaySQL.Spreadsheet.FormulaEditor.closeDialog(spreadsheet, false);
        }
        inlinedialog.redisplay = function() {
            inlinedialog.redisplay.isRedisplaying = true;
            inlinedialog.hide();
            inlinedialog.show();
            inlinedialog.redisplay.isRedisplaying = false;
        }
        var onClickOutsideInlineDialog = function (e) {
            if (!spreadsheet.inlinedialog || !spreadsheet.inlinedialog.is(":visible")) return false;
            var target = e.target || this;
            var $this = $(target),
                $targetCell = $this.closest("td,th");
            // Abort if the user clicks in the current cell
            if ($targetCell.is(inlinedialog.$input.closest("td, th"))) return;
            // Abort if the user clicks in an inline dialog
            if ($this.closest("#inline-dialog-formula-editor" //+
                //"#null-sorting-dropdown, " +
                //"#select2-drop, #select2-drop-mask, #select2-drop-mask"
            ).size() > 0) return;
            // If it's not in the DOM, it was detached, therefore it's probably button of the Inline Dialog
            // which was removed.
            if (!$.contains(document.documentElement, target)) return;

            // Ok, so now we have a click on a cell. Wait for playsql-selection.js to complete, then
            // look at what cell-ref-active brings back.
            window.setTimeout(function(){
                if (!spreadsheet.inlinedialog || spreadsheet.inlinedialog.name != "formula-editor") return;
                var $selection = spreadsheet.$placeholder.find(".cell-ref-active")
                                    .not(spreadsheet.inlinedialog.$input.closest("td"));
                if ($selection.size() == 0) {
                    // console.log("A click for a formula was lost");
                } else if ($selection.is("thead th")) {
                    $selection = $selection.filter("thead th").first();
                    var colName = $selection.attr("data-name");
                    var $el = $("<span>", {"class": "subrule-value-range subrule-range-all"})
                        .append($("<span>", {
                                    "class": "subrule-column-identifier",
                                    text: colName
                                }))
                        .append($("<span>", {
                                    "class": "subrule-range-all",
                                    text: ":*"
                                }));
                    inlinedialog.append($el);
                } else {
                    var $selection = $selection.first("td");
                    var cell = spreadsheet.getCell($selection),
                        currentCell = spreadsheet.getCell(inlinedialog.$input.closest("td")),
                        $el;
                    if (!cell) return;
                    if (!!currentCell && cell.rowId == currentCell.rowId) {
                        $el = $("<span>", {
                                "class": "subrule-value-cell subrule-cell-same-row subrule-column-identifier",
                                text: cell.column.name
                            });
                    } else {
                        $el = $("<span>", {"class": "subrule-value-cell subrule-cell-row-id"})
                            .append($("<span>", {
                                "class": "subrule-column-identifier",
                                text: cell.column.name
                            }))
                            .append(":")
                            .append($("<span>", {
                                "class": "subrule-row-id-digits",
                                text: cell.rowId
                            }));
                    }
                    inlinedialog.append($el);
                }
            }, 10);
            return;

            // Check whether we should append the formula instead
            var cell = spreadsheet.getCell($targetCell);
            if (cell) {
                if ($targetCell[0] == $cell[0]){
                    // Let something else take action: The user probably selected text in the formula
                    return true;
                }
                // Build a range definition
                var row = ":" + (cell.isColumnHeader ? "*" : $cell.closest("tr").attr("data-row-id"));
                var columnName = cell.column.name;
                var formulaAddition = columnName + row;
                inlinedialog.append(formulaAddition);
                // Here add colors for the formula
            }
            // If we don't prevent bubbling, each of the parents will receive the same click
            return false;
        };
        spreadsheet.$placeholder.on("mouseup.playsql-inlinedialog", "td, th", onClickOutsideInlineDialog);

        selectionManager.options.cssClassMap = { "selected": "cell-ref-active", "focused": "cell-ref-focused" };
        inlinedialog.show();
        inlinedialog.find(".remove-formula-link").click(function(e){
            e.preventDefault();
            Confluence.PlaySQL.Spreadsheet.FormulaEditor.removeFormula(spreadsheet, $cell, renderer);
            return false;
        });
        debouncedTriggerSyntaxColoration();
        return inlinedialog;
    },
    triggerAutocompletion: function(spreadsheet, $cell, $input, inlinedialog, event) {
        var text = $input.text(),
            caret = Confluence.PlaySQL.Helpers.caretSelection($input[0]);
        if (text.length == 0 || caret.position == undefined) return;
        var textBeforeCaret = text.substr(0, caret.position),
            parsed = Confluence.PlaySQL.Helpers.parseFunctionAtCaret(textBeforeCaret),
            targetTable = parsed.currentTable || spreadsheet.id;
        Confluence.PlaySQL.Helpers.getFunctionList(spreadsheet.context, targetTable, function (functionList) {
            if (!functionList || functionList.length == 0) return;
            var options = {},
                current = parsed.current && parsed.current.toLowerCase();
            function highlight(substring, text) {
                // If 'text' is part of HTML tags used here, don't take the risk
                if (new RegExp("<.*" + substring + ".*>").test(text)) return text;
                return text.replace(new RegExp("(" + substring + ")", 'i'), "<b>$1</b>");
            }
            if (current) {
                var candidates = [];
                for (var i in functionList.functions) {
                    var item = functionList.functions[i],
                        matches = item.name.toLowerCase().indexOf(current) == 0,
                        description = item.description,
                        priority = 1;
                    if (!matches && item.keywords) {
                        for (var i in item.keywords)
                            if (item.keywords[i].toLowerCase().indexOf(current) == 0) {
                                priority = 2;
                                matches = true;
                                description += " (keyword: " + highlight(current, item.keywords[i]) + ")";
                                break;
                            }
                    }
                    if (!matches && description) {
                        if (description.toLowerCase().indexOf(current) != -1) {
                            description = highlight(current, description);
                            priority = 20;
                            matches = true;
                        }
                    }
                    if (matches) {
                        candidates.push({
                            name: priority < 2 ? highlight(current, item.name) : item.name,
                            autocomplete: item.name,
                            type: "function",
                            // Override the description
                            detail: $.extend({}, item, {description:description}),
                            priority: priority
                        });
                    }
                }
                for (var i in functionList.columnNames) {
                    var item = functionList.columnNames[i];
                    if (item.name.toLowerCase().indexOf(current) == 0) {
                        candidates.push({
                            name: highlight(current, item.name),
                            autocomplete: item.name,
                            type: "column",
                            detail: item,
                            priority : 3
                        });
                    }
                }
                for (var i in functionList.tables) {
                    var item = functionList.tables[i];
                    if (item.name.toLowerCase().indexOf(current) == 0) {
                        candidates.push({
                            name: highlight(current, item.name),
                            autocomplete: item.name,
                            type: "table",
                            detail: item,
                            priority : 4
                        });
                    }
                }
                for (var i in functionList.constants) {
                    var item = functionList.constants[i];
                    if (item.name.toLowerCase().indexOf(current) == 0) {
                        candidates.push({
                            name: highlight(current, item.name),
                            autocomplete: item.name,
                            type: item.type,
                            detail: item,
                            priority : 5
                        });
                    }
                }
                // Sort by priority
                candidates.sort(function(a, b) {return a && a.priority && b && b.priority ? a.priority - b.priority : 0 });
                options.candidates = candidates;
            }
            if (parsed.parentFn) {
                var functionDef = functionList.functions.filter(function(f){ return f.name == parsed.parentFn })[0];
                if (!functionDef) {
                    options.parentFnNotFound = parsed.parentFn;
                } else {
                    options.parentFn = functionDef;
                }
            }
            inlinedialog.setJavadoc(options);
        });
    },
    triggerSyntaxColoration : function($input, spreadsheet, inlinedialog) {
        // Replace the contents of the cell with syntax-highlighted contents
        var text = $input.text(),
            url = AJS.contextPath() + "/rest/playsql/1.0/autocomplete/" + spreadsheet.context + "/" + spreadsheet.id + "/checkFormula",
            caret = Confluence.PlaySQL.Helpers.caretSelection($input[0]);
        if (caret.length != 0 || text.length < 3) {
            spreadsheet.$placeholder
                .find(".cell-ref-inactive, .cell-ref-active")
                .removeClass("cell-ref-inactive cell-ref-active");
            return;
        }

        url += "?" + $.param({
            formula: text,
            column: "col1",
            caretPosition: caret.position
        });

        $.ajax({
            url: url,
            type: 'GET',
            contentType: "application/json; charset=utf-8"
        }).done(function(data){
            if (!$input.is(":visible")) return;
            var caret = Confluence.PlaySQL.Helpers.caretSelection($input[0]),
                inputText = $input.text();
            if (caret.length != 0) return;
            if (data.formula != inputText) {
                declassifyFormulaPartsAtCaretPosition();
                return;
            }
            // The markup is good and matches the input. We can update the formula, reposition the caret,
            // and update the cell-ref-active and cell-ref-inactive.
            var markupHtml = data.markup,
                markupText = $("<div>").html(markupHtml).text();
            if (markupText != inputText) {
                if (inputText.indexOf(markupText) == 0) {
                    // The user has just added a few letters. We can still recover by appended the added
                    // text.
                    markupHtml += AJS.escapeHtml(inputText.substr(markupText.length));
                } else {
                    declassifyFormulaPartsAtCaretPosition();
                    return;
                }
            }

            $input.html(markupHtml);
            //inlinedialog.setJavadoc(data.currentFunctions.length > 0 && data.currentFunctions[data.currentFunctions.length - 1]);

            var range = document.createRange();
            Confluence.PlaySQL.Helpers.setSelection(range, $input[0], caret.position, "setEnd");
            range.collapse(false); // to end
            window.getSelection().removeAllRanges();
            window.getSelection().addRange(range);
            $input.data("saved-selection", range);

            spreadsheet.$placeholder
                .find(".cell-ref-inactive, .cell-ref-active")
                .removeClass("cell-ref-inactive cell-ref-active");

            var $current = getCurrentFormulaElement();
            $input.find(".subrule-value-cell, .subrule-value-range").not($current).each(function(){
                highlightOriginCells(spreadsheet, $input, $(this), "cell-ref-inactive", "");
            });
            highlightOriginCells(spreadsheet, $input, $current, "cell-ref-active", "");

        }).fail(function(jqXHR, error, reason){
            AJS.log("Can't get syntax highlighting for formula: " + text);
            declassifyFormulaPartsAtCaretPosition();
        });
    },
    highlightCellsAtCaret: function(spreadsheet, $input) {
        // If the caret in a formula is above a cell reference, we need to highlight it is active
        // and highlight other references as inactive
        spreadsheet.$placeholder.find(".cell-ref-active").addClass("cell-ref-inactive").removeClass("cell-ref-active");
        var $current = getCurrentFormulaElement();
        highlightOriginCells(spreadsheet, $input, $current, "cell-ref-active", "cell-ref-inactive");
    },
    triggerKeyboardCellSelection: function(spreadsheet, $cell, $input, inlinedialog, e){
        var keyCombinationDetail =  Confluence.PlaySQL.Helpers.getKeyCombinationDetail(e),
            $li = $("#inline-dialog-formula-editor .autocomplete-item.active");
        switch (keyCombinationDetail.canonical) {
            case "up":
                //e.preventDefault();
                $li = $li.removeClass("active").prev(".autocomplete-item").addClass("active");
                if ($li.size() == 0) {
                    $("#inline-dialog-formula-editor .autocomplete-item").last().addClass("active");
                }
                return false;
            case "down":
                //e.preventDefault();
                $li = $li.removeClass("active").next(".autocomplete-item").addClass("active");
                if ($li.size() == 0) {
                    $("#inline-dialog-formula-editor .autocomplete-item").first().addClass("active");
                }
                return false;
            case "enter":
                //e.preventDefault();
                var success = Confluence.PlaySQL.Spreadsheet.FormulaEditor.insertAutocomplete(spreadsheet, $li);
                if (!success) {
                    // Save the cell
                    var cell = spreadsheet.getCell($cell);
                    if (cell != null) {
                        cell.renderer.saveAndView();
                        spreadsheet.sprayRequestPipe();
                    }
                }
                return false;
            case "esc":
                Confluence.PlaySQL.Spreadsheet.FormulaEditor.cancelEdit(spreadsheet, $cell);
                break;
        }
        return true;
    },
    /** Keeps the selection in the data of the $input */
    saveSelection: function($input) {
        var selection = window.getSelection();
        if (selection.rangeCount > 0) {
            var range = selection.getRangeAt(0).cloneRange();
            $input.data("saved-selection", range);
        }
    },
    /* Inserts the autocompleted value into the formula
    * Syntax: insertAutocomplete(spreadsheet, $(".autocomplete-item.active"))
    * @return true if something was autocompleted, false otherwise (if false, the user prolly wanted to save the cell)
    * */
    insertAutocomplete: function(spreadsheet, $li){
        // Check if the active dialog is still the formula editor
        var inlinedialog = spreadsheet.inlinedialog;
        if (!inlinedialog || !inlinedialog.name == "formula-editor") return false;
        var $input = inlinedialog.$input;
        var autocompleteText, autocompleteType;
        if ($li instanceof $) {
            if ($li.size() == 0) return;
            autocompleteText = $li.attr("data-autocomplete");
            autocompleteType = $li.attr("data-autocomplete-type");
        } else {
            autocompleteText = arguments[1];
            autocompleteType = arguments[2];
        }
        var autocompleteText2 = autocompleteType == "function" ? autocompleteText + "(" : autocompleteText;
        var regex = /^[a-zA-Z0-9_:]+\(?$/;

        function selectTextMatchingRegex(range, regex, el, method) {
            var startingFromEnd = (method == "setEnd");
            if (el.nodeType == 3) {
                // Loop from the beginning or from the end of the text
                for (var i = startingFromEnd ? el.length : 0 ; i <= el.length && i >= 0 ; startingFromEnd ? i-- : i++) {
                    // We clone because there's a risk we collapse it
                    var range2 = range.cloneRange();
                    range2[method](el, i);
                    if (range2.collapsed) return false;
                    var text = range2.toString();
                    if (regex.test(text)) {
                        // The selection is set. Stop here.
                        range[method](el, i);
                        return true;
                    }
                }
            } else if (el.childNodes) {
                if (!startingFromEnd) {
                    for (var i in el.childNodes) {
                        var result = selectTextMatchingRegex(range, regex, el.childNodes[i], method);
                        // If the result is true or false, return it
                        if (typeof result != "undefined") return result;
                    }
                } else {
                    // Same loop, but starting from the end
                    for (var i = el.childNodes.length - 1 ; i >= 0 ; i--) {
                        var result = selectTextMatchingRegex(range, regex, el.childNodes[i], method);
                        // If the result is true or false, return it
                        if (typeof result != "undefined") return result;
                    }
                }
            }
        }

        var selection = window.getSelection();
        if (!selection.rangeCount) {
            $input.append(document.createTextNode(autocompleteText2));
            return true;
        }
        var range = selection.getRangeAt(0);
        var root = $(range.startContainer).closest("[contenteditable]")[0];
        if (!root) return true;
        if (selectTextMatchingRegex(range, regex, root, "setStart")) {
            range.deleteContents();
        }
        if (selectTextMatchingRegex(range, regex, root, "setEnd")) {
            range.deleteContents();
        }
        var newNode = document.createTextNode(autocompleteText2);
        range.insertNode(newNode);
        range.collapse(false);
        selection.removeAllRanges();
        selection.addRange(range);

        // Now find the functionDef and reset the javadoc
        spreadsheet.inlinedialog.setJavadoc("");
        Confluence.PlaySQL.Helpers.getFunctionList(spreadsheet.context, spreadsheet.id, function (functionList) {
            if (!functionList || functionList.length == 0 || !functionList.functions) return;
            var functionDef = functionList.functions.filter(function(f){ return f.name == autocompleteText; })[0];
            // The function may be empty if 'spreadsheet.id' is not the same as parsed.currentTable.
            if (!functionDef) return;
            spreadsheet.inlinedialog.setJavadoc({parentFn: functionDef});
        });
        Confluence.PlaySQL.Spreadsheet.FormulaEditor.triggerSyntaxColoration($input, spreadsheet, inlinedialog);
        return true;
    },
    validate: function(spreadsheet, $cell, renderer) {
        // Formulas can always be stored
        return true;
    },
    saveAndView: function(spreadsheet, $cell, renderer) {
        spreadsheet.inlinedialog && spreadsheet.inlinedialog.close();
        $cell.removeClass("invalid");
        var formula = "=" + $cell.find(".formula-editor").text(),
            previousValue = $cell.attr("data-cell-metadata-formula");
        if (formula == previousValue) {
            this.cancelEdit(spreadsheet, $cell, renderer);
            return;
        }
        if (formula.trim() == "=") {
            // Just delete the contents
            $cell.removeAttr('data-rendered');
            this.cancelEdit(spreadsheet, $cell, renderer);
            if (spreadsheet.saveCellMetadata($cell, "formula", null))
                spreadsheet.sprayRequestPipe();
            return;
        }

        $cell.empty()
            .text(formula)
            .addClass("saving").removeClass("editing");

        if (spreadsheet.saveCellMetadata($cell, "formula", formula)) {
            spreadsheet.sprayRequestPipe();
        } else {
            $cell.html($cell.attr("data-rendered"));
        }
    },
    cancelEdit: function(spreadsheet, $cell) {
        spreadsheet.inlinedialog && spreadsheet.inlinedialog.close();
        $cell.tipsy("disable");
        $(".tipsy").remove();
        var rendered = $cell.attr('data-rendered');
        if (rendered == undefined) {
            $cell.empty().removeClass("editing saving invalid");
        } else {
            $cell.html(rendered).removeClass("editing saving invalid");
        }
    },
    /**
     * Turns the cell back to a normally-rendered cell and sets the anti-formula class
     */
    removeFormula: function(spreadsheet, $cell, renderer) {
        var value = "=" + $cell.find(".formula-editor").text();
        Confluence.PlaySQL.Spreadsheet.FormulaEditor.cancelEdit(spreadsheet, $cell, renderer);
        $cell.removeClass("formula").addClass("anti-formula");
        renderer.edit(value);
    },
    checkAndSaveFormulaRemoval: function(spreadsheet, $cell) {
        if (spreadsheet.perms.hasPerm('EDIT_ROWS_METADATA')) {
            spreadsheet.saveCellMetadata($cell, "formula", null);
        }
    },
    /**
     * Feel free to trigger this function every time you're not sure whether the inline dialog / formula were closed.
     **/
    closeDialog: function(spreadsheet, cancelAllEditingFormulas){
        $(".tipsy.tipsy-n").remove();
        $(document).off(".playsql-inlinedialog");
        spreadsheet.$placeholder.off(".playsql-inlinedialog");
        spreadsheet.$placeholder.find(".cell-ref-inactive, .cell-ref-active").removeClass("cell-ref-inactive cell-ref-active");
        if (spreadsheet.inlinedialog && spreadsheet.inlinedialog.name == "formula-editor") {
            spreadsheet.inlinedialog.close();
        }
        var selectionManager = PlaySQL.Selection.forTable(spreadsheet.$placeholder);
        selectionManager.options.cssClassMap = undefined;
        if (cancelAllEditingFormulas) {
            var $editingCells = spreadsheet.$placeholder.find(".editing.formula");
            if ($editingCells.size() > 0) {
                Confluence.PlaySQL.Spreadsheet.FormulaEditor.cancelEdit(spreadsheet, $editingCells);
            }
        }
        // Should there still be a reluctant dialog, kill it forcefully
        window.setTimeout(function(){ $("#inline-dialog-formula-editor").remove(); }, 0);
    }
};

function declassifyFormulaPartsAtCaretPosition() {
    var range = window.getSelection().rangeCount > 0 && window.getSelection().getRangeAt(0);
    if (!range || !range.collapsed) return;
    var $caretContainer = $(range.endContainer);
    // It only acts if the caret is within the formula
    $caretContainer.parents(".playsql-formula *").attr("class", "");
    return;
}

function getCurrentFormulaElement() {
    var sel = window.getSelection();
    if (!sel.rangeCount > 0) return $([]);
    var range = sel.getRangeAt(0),
        $startNode = $(range.startContainer.parentNode);
    return $startNode.closest(".subrule-value-cell, .subrule-value-range");
}

function highlightOriginCells(spreadsheet, $input, $formulaCellOrRangeElement, cssClass, cssClassToRemove) {
    var $this = $formulaCellOrRangeElement,
        columnName = $this.is(".subrule-column-identifier") ? $this.text() : $this.find(".subrule-column-identifier").first().text(),
        rowNum = $this.find(".subrule-row-id-digits").text(),
        isRange = $this.is(".subrule-value-range");
    var column = spreadsheet.colOf(columnName);
    if (!column) return;
    rowNum = rowNum.replace(/[^0-9]/g, "");

    if (isRange) {
        spreadsheet.$findColumnCells(column, true).addClass(cssClass);

    } else {
        if (!rowNum) {
            var $formulaCell = $input.closest("th,td");
            if ($formulaCell.size() == 0) return;
            rowNum = spreadsheet.getCell($formulaCell).rowId;
        } else {
            if (spreadsheet.$placeholder.find("tr[data-row-id=" + rowNum + "]").size() == 0) return;
        }
        var cell = spreadsheet.getCell(rowNum, columnName);
        if (cell) cell.$cell.addClass(cssClass).removeClass(cssClassToRemove);
    }
}
})();