(function($){
    AJS.bind("playsql-spreadsheet-init", function(event, api) {

        function outerHtml($value) {
            return $(document.createElement("div")).html($value).html();
        }

        var PageRenderer = function(){
            var parent = new Confluence.PlaySQL.Spreadsheet.RendererBase();

            return $.extend({}, parent, {
                key: "confluencecontent",
                getI18nName : function() {
                    return AJS.I18n.getText("com.playsql.playsql-base-plugin.renderers.page-picker")
                },
                isSupportedType : function(type) {
                    return type == "VARCHAR" || type == "TEXT";
                },
                getDefaultType : function() {
                    return "TEXT";
                },
                isSystemRenderer : function() {
                    return false;
                },
                renderCellContents: function(value, spreadsheet, column) {
                    if (value == null) {
                        return "<span class='not-a-value'>&nbsp;</span>";
                    }
                    var parts = value.split(";");
                    if (parts.length >= 3) {
                        var contentId = parts[0],
                            spaceKey = parts[1],
                            title = parts.slice(2).join(";");
                        var href = AJS.contextPath() + "/pages/viewpage.action?pageId=" + AJS.escapeHtml(contentId),
                            $link = $("<a>", {
                                href : href,
                                text: title
                            });
                        return outerHtml($link);
                    } else {
                        return AJS.escapeHtml(value).replace(/\n/g, "<br>\n");
                    }
                },
                createInputBox : function(value) {
                    // The value we display to the user should be user-friendly. It's when the autocomplete is triggered
                    // that the correct value is inserted.
                    var parts = !!value && value.split(";");
                    if (parts && parts.length > 2 && !!parts[2]) {
                        value = parts[2];
                    }
                    var $input = parent.createInputBox(value);
                    // https://developer.atlassian.com/display/CONFDEV/JavaScript+Components
                    $input.addClass("autocomplete-confluence-content").attr("data-template", "{id};{space.key};{title}");
                    window.setTimeout(function(){
                        $input.closest("td").addClass("no-autocommit pagepicker no-move");
                        AJS.Confluence.Binder.autocompleteConfluenceContent($input.parent());
                        $input.on("selected.autocomplete-content", function() {
                            $input.parent().removeClass("no-autocommit");
                        });
                    }, 0);
                    return $input;
                },
                forCell: function(spreadsheet, $cell) {
                    var result = parent.forCell.apply(this, arguments);
                    return $.extend(result, {
                        saveAndView: function() {
                            if (!$cell.hasClass("editing") || $cell.hasClass("saving")) return;
                            if (!this.validate()) {
                                return;
                            }
                            var rowId = $cell.closest("tr").attr("data-row-id");
                            if (rowId == undefined || rowId == "") {
                                return;
                            }
                            var column = spreadsheet.colOf($cell);
                            var $input = $cell.find("input, textarea"),
                                val = $input.val();
                            if (!spreadsheet.perms.hasPerm('EDIT_CELLS') && !spreadsheet.perms.hasPerm('Q_VIEW_EDIT', column)) return;
                            if ($cell.hasClass("formula") || !$cell.hasClass("anti-formula") && val.charAt(0) == "=") {
                                /* TODO Formula activation/deactivation here */
                                Confluence.PlaySQL.Spreadsheet.FormulaEditor.saveAndView(spreadsheet, $cell, this);
                                return;
                            }
                            var previousValue = $cell.attr('data-value'),
                                rendered = this.parent.renderCellContents(val, spreadsheet, column);
                            if (val == '') {
                                val = null;
                            }
                            if (previousValue == val) {
                                $cell.html($cell.attr('data-rendered')).removeClass('editing');
                                return;
                            }

                            $cell.attr('data-value', val);
                            // Keep the value for cancel
                            $cell.attr('data-rendered', rendered);

                            $cell.html(rendered).addClass("saving").removeClass("editing");
                            spreadsheet.saveCell($cell, val);
                            Confluence.PlaySQL.Spreadsheet.FormulaEditor.checkAndSaveFormulaRemoval(spreadsheet, $cell);
                        }
                    });
                }
            });
        };
        api.register(new PageRenderer());


        var UserRenderer = function(){
            var parent = new Confluence.PlaySQL.Spreadsheet.RendererBase();

            return $.extend({}, parent, {
                key: "confluenceuser",
                getI18nName : function() {
                    return AJS.I18n.getText("com.playsql.playsql-base-plugin.renderers.user-picker")
                },
                isSupportedType : function(type) {
                    return type == "VARCHAR" || type == "TEXT";
                },
                getDefaultType : function() {
                    return "TEXT";
                },
                isSystemRenderer : function() {
                    return false;
                },
                renderCellContents: function(value, spreadsheet, column) {
                    if (value == null) {
                        return "<span class='not-a-value'>&nbsp;</span>";
                    }
                    var parts = value.split(";");
                    if (parts.length >= 4) {
                        var username = parts[0],
                            type = parts[1],
                            id = parts[2],
                            title = parts.slice(3).join(";");
                        var href = AJS.contextPath() + "/display/~" + AJS.escapeHtml(username),
                            $link = $("<a>", {
                                href : href,
                                text: title,
                                "class": "confluence-userlink user-mention", // userlink-0
                                "data-linked-resource-id": id,
                                "data-linked-resource-type": "user",
                                "data-username": username,
                                "data-base-url": AJS.Meta.get("base-url")
                            });
                        window.setTimeout(function(){
                            // to ensure user hover over user links work
                            Confluence.Binder.userHover();
                        }, 0);
                        return outerHtml($link);
                    } else {
                        return AJS.escapeHtml(value).replace(/\n/g, "<br>\n");
                    }
                },
                createInputBox : function(value) {
                    // The value we display to the user should be user-friendly. If it's saved,
                    // the back-end will rebuild the value with the user key and full name.
                    var parts = !!value && value.split(";");
                    if (parts && parts.length > 3 && !!parts[0] && !!parts[3]) {
                        value = parts[3] + (!!parts[0] ? " (" + parts[0] + ")" : "");
                    } else if (parts && parts.length > 0 && !!parts[0]) {
                        value = parts[0];
                    }
                    var $input = parent.createInputBox(value);
                    // https://developer.atlassian.com/display/CONFDEV/JavaScript+Components
                    // Note: It looks like the data-template attribute doesn't work.
                    $input.addClass("autocomplete-user").attr("data-template", "{username}");
                    window.setTimeout(function(){
                        $input.closest("td").addClass("no-autocommit userpicker no-move");
                        if (AJS.Confluence.Binder.autocompleteUserOrGroup) {
                            // The method searches in the whole page. We give a scope in the first argument.
                            AJS.Confluence.Binder.autocompleteUserOrGroup($input.parent());
                        } else if (AJS.Confluence.Binder.autocompleteUserOrGroup) {
                            AJS.Confluence.Binder.autocompleteUser($input.parent());
                        }
                        $input.on("selected.autocomplete-user", function() {
                            $input.parent().removeClass("no-autocommit");
                        });
                    }, 0);
                    return $input;
                },
                forCell: function(spreadsheet, $cell) {
                    var result = parent.forCell.apply(this, arguments);
                    return $.extend(result, {
                        saveAndView: function() {
                            if (!$cell.hasClass("editing") || $cell.hasClass("saving")) return;
                            if (!this.validate()) {
                                return;
                            }
                            var rowId = $cell.closest("tr").attr("data-row-id");
                            if (rowId == undefined || rowId == "") {
                                return;
                            }
                            var column = spreadsheet.colOf($cell);
                            var $input = $cell.find("input, textarea"),
                                val = $input.val();
                            if (!spreadsheet.perms.hasPerm('EDIT_CELLS') && !spreadsheet.perms.hasPerm('Q_VIEW_EDIT', column)) return;
                            if ($cell.hasClass("formula") || !$cell.hasClass("anti-formula") && val.charAt(0) == "=") {
                                /* TODO Formula activation/deactivation here */
                                Confluence.PlaySQL.Spreadsheet.FormulaEditor.saveAndView(spreadsheet, $cell, this);
                                return;
                            }
                            var previousValue = $cell.attr('data-value'),
                                rendered = this.parent.renderCellContents(val, spreadsheet, column);
                            if (val == '') {
                                val = null;
                            }
                            if (previousValue == val) {
                                $cell.html($cell.attr('data-rendered')).removeClass('editing');
                                return;
                            }

                            $cell.attr('data-value', val);
                            // Keep the value for cancel
                            $cell.attr('data-rendered', rendered);

                            $cell.html(rendered).addClass("saving").removeClass("editing");
                            spreadsheet.saveCell($cell, val);
                            Confluence.PlaySQL.Spreadsheet.FormulaEditor.checkAndSaveFormulaRemoval(spreadsheet, $cell);
                        }
                    });
                }
            });
        };
        api.register(new UserRenderer());

        function checkNotNull(value, message) {
            if (value == undefined) {
                throw "Unexpectedly undefined value: " + value + "(" + message + ")";
            }
            if (value instanceof $) {
                if (value.size() == 0)
                    throw "Unexpectedly empty jquery element: " + value + "(" + message + ")";
            }
            return value;
        }
    });
})(AJS.$);