(function($){
    AJS.bind("playsql-spreadsheet-init", function(event, api) {

        function outerHtml($value) {
            return $(document.createElement("div")).html($value).html();
        }

        var LinkRenderer = function(){
            var base = new Confluence.PlaySQL.Spreadsheet.RendererBase(),
                url = AJS.contextPath() + "/rest/playsql/1.0/jiralinks?";
            Confluence.PlaySQL.Spreadsheet.Applinks = {
                latch: 0,
                projects: {}
            };

            function getJIRAProject(name) {
                var projectKey = name && name.split("-")[0];
                if (!Confluence.PlaySQL.Spreadsheet.Applinks.projects[projectKey] || Confluence.PlaySQL.Spreadsheet.Applinks.projects[projectKey].status == 'UNKNOWN') {
                    Confluence.PlaySQL.Spreadsheet.Applinks.projects[projectKey] = { status: "SEARCHING" };

                    var url = AJS.contextPath() + "/rest/playsql/1.0/jiralinks?" + $.param({issue: projectKey});
                    $.ajax({
                        url: url,
                        type: 'GET',
                        contentType: "application/json; charset=utf-8"
                    }).done(function (data, status, request) {
                        Confluence.PlaySQL.Spreadsheet.Applinks.projects[projectKey] = {
                            status: "FOUND",
                            baseUrl: data.substr(0, data.length - projectKey.length - 1)
                        };
                        ajaxDone();
                    }).fail(function (jqXHR, error, reason) {
                        Confluence.PlaySQL.Spreadsheet.Applinks.projects[projectKey] = { status: "UNKNOWN" };
                        AJS.log("Can't find JIRA project for " + projectKey);
                    });
                }
                return Confluence.PlaySQL.Spreadsheet.Applinks.projects[projectKey];
            }

            function ajaxDone() {
                var $toRender = $("a.jira-issue-to-render");
                $toRender.each(function(){
                    var $this = $(this),
                        key = $this.attr('data-jira-key'),
                        project = getJIRAProject(key);
                    if (project.status == "FOUND") {
                        $this.attr('href', project.baseUrl + key).removeClass("jira-issue-to-render");
                    }
                });
            }

            function render(value) {
                if (/^[A-Z-]+[0-9]+$/.test(value)) {
                    var project = getJIRAProject(value);

                    if (project.status == "FOUND") {
                        return outerHtml($("<a></a>", {
                            href: project.baseUrl + value,
                            "data-jira-key": value,
                            text: value
                        }));
                    }
                    return outerHtml($("<a></a>", {
                        href: undefined,
                        "data-jira-key": value,
                        "class": "jira-issue-to-render",
                        text: value
                    }));
                }
                return AJS.escapeHTML(value);
            }

            return $.extend({}, base, {
                key: "link",
                getI18nName : function() {
                    return AJS.I18n.getText("com.playsql.playsql-base-plugin.renderers.link")
                },
                isSupportedType : function(type) {
                    return type == "VARCHAR" || type == "TEXT";
                },
                getDefaultType : function() {
                    return "TEXT";
                },
                isSystemRenderer : function() {
                    return false;
                },
                renderCellContents: function(value, spreadsheet, column) {
                    if (value == null) {
                        return "<span class='not-a-value'>&nbsp;</span>";
                    }
                    // URL is at the beginning
                    if (value.indexOf(":") != -1 || value.indexOf("@") != -1) {
                        var protocolIndexOf = value.indexOf(":"),
                            spaceIndexOf = value.indexOf(" "),
                            linkText = value,
                            linkHref = value;
                        if (spaceIndexOf != -1) {
                            if (spaceIndexOf > protocolIndexOf) {
                                linkText = value.substring(spaceIndexOf + 1);
                                linkHref = value.substring(0, spaceIndexOf);
                            } else {
                                // Link in the middle of a text
                                value = AJS.escapeHtml(value);
                                var values = value.split(" ");
                                for (var i in values) {
                                    var valueI = values[i];
                                    if (/^[a-z][a-z0-9+.-]+:[^ ]/i.test(valueI)) {
                                        values[i] = "<a href='" + valueI + "'>" + valueI + "</a>";
                                    } else if (valueI.indexOf("@") != -1) {
                                        values[i] = "<a href='mailto:" + valueI + "'>" + valueI + "</a>";
                                    }
                                }
                                return values.join(" ");
                            }
                        }
                        return outerHtml($("<a></a>", {
                            href: linkHref.indexOf("@") != -1 && linkHref.indexOf("mailto:") == -1 ? "mailto:" + linkHref : linkHref,
                            text: linkText
                        })).replace(/\n/g, "<br>\n");
                    }
                    var jiraIssue = /^[A-Z]+-[0-9]+$/
                    if (jiraIssue.test(value)) {
                        return render(value);
                    }
                    return AJS.escapeHtml(value).replace(/\n/g, "<br>\n");
                },
                createInputBox : function(value) {
                    return this.createTextareaBox(value);
                },
                forCell: function(spreadsheet, $cell) {
                    var result = base.forCell.apply(this, arguments);
                    return $.extend(result, {
                        // Nothing to extend here
                    });
                }
            });
        };
        api.register(new LinkRenderer());

        function checkNotNull(value, message) {
            if (value == undefined) {
                throw "Unexpectedly undefined value: " + value + "(" + message + ")";
            }
            if (value instanceof $) {
                if (value.size() == 0)
                    throw "Unexpectedly empty jquery element: " + value + "(" + message + ")";
            }
            return value;
        }
    });
})(AJS.$);