(function($){

    var checkNotNull = Confluence.PlaySQL.Helpers.checkNotNull;

    /**
     * @deprecated this id is not unique on the page
     * @param id
     * @returns {*}
     */
    function byId(id) {
        if (Confluence.PlaySQL.Spreadsheet.spreadsheets[id]) {
            return Confluence.PlaySQL.Spreadsheet.spreadsheets[id];
        }

        if (id.hasClass && id.hasClass("playsql-spreadsheet-placeholder")) {
            // The guy substituted a jquery element for the id. Swap that.
            return byPlaceholder(id);
        } else {
            var $placeholder = $(".playsql-spreadsheet-placeholder[data-macro-spreadsheet-id='" + id + "'");
            if ($placeholder.size() != 1) throw "No placeholder with ID " + id;
            return byPlaceholder($placeholder);
        }
    }
    function byPlaceholder($placeholder) {
        var spreadsheet = $placeholder.data('playsql-spreadsheet');
        if (spreadsheet) return spreadsheet;
        spreadsheet = {
            $placeholder: $placeholder,
            id : $placeholder.attr('data-macro-spreadsheet-id'),
            context : $placeholder.attr('data-macro-context'),
            label: undefined,
            columns: undefined,
            whereClause : $placeholder.attr('data-macro-where-clause'),
            whereClauseColumnFilters : undefined,
            whereClauseConsolidated : undefined,
            orderbyClause : $placeholder.attr('data-macro-orderby-clause'),
            displayColumns : $placeholder.attr('data-macro-display-columns'),
            pagination: undefined,
            permlevel : $placeholder.attr('data-macro-permlevel'),
            perms : Confluence.PlaySQL.Helpers.parsePerms($placeholder.attr('data-macro-permlevel')),
            customEmptyMessage : $placeholder.attr('data-macro-custom-empty-message'),
            decimalSeparator : '.',
            defaultCase: undefined,
            dialect: undefined,
            mode: undefined, // Mode: READ_ONLY or READ_WRITE
            systemColumnsMode: undefined, // JUST_USE or SYNTHETIC
            // When we want to pipe the requests to the server, we just create an array here.
            requestPipe : undefined,
            totals: undefined,
            /* If provided, this definition is used instead of the model
             * saved in the database for the time of this request. Optional.
             */
            queryModelOverride: undefined,
            /* The list of user parameters. To be displayed before the spreadsheet. */
            userParameters: [],
            /* If user parameters are present, whether the form should be displayed */
            useForm: undefined,
            options : {
                onDisplay: undefined,
                onSelectionSet: undefined,
                onChangeTotals: undefined,
                saveTotals: undefined,
                saveColumn: undefined
            },
            /* Disables focusing on cells while the js is editing the spreadsheet (pasteHandler) */
            disableFocusManagement: false
        };

        if (!spreadsheet.id) throw "No ID on placeholder " + $placeholder.html();
        $placeholder.data('playsql-spreadsheet', spreadsheet);

        /**
         * Builds and returns the base url for a rest request
         * @param endpoint the resource name, such as "/data". The last path can
         *          be an object, in this case it's considered as the query params
         */
        var baseUrl = function(endpoint, paths, queryParams) {
            // If you think you could reuse baseUrlForContext, don't forget that V8 engines can't perform
            // optimizations on sliced arguments, thus requiring you to write just as much code.
            var suffix = "";
            for (var i = 0 ; i < arguments.length ; i++) {
                var arg = arguments[i];
                if (i == arguments.length -1 && (typeof arg == "object")) {
                    suffix += "?" + $.param(arg);
                } else {
                    suffix += "/" + arg.replace(/^\/|\/$/g, "");
                }
            }

            return AJS.contextPath() + "/rest/playsql-spreadsheets/1.0/" + spreadsheet.context + "/" + spreadsheet.id + suffix;
        }

        var baseUrlForContext = function(endpoint, paths, queryParams) {
            var suffix = "";
            for (var i = 0 ; i < arguments.length ; i++) {
                var arg = arguments[i];
                if (i == arguments.length -1 && (typeof arg == "object")) {
                    suffix += "?" + $.param(arg);
                } else {
                    suffix += "/" + arg.replace(/^\/|\/$/g, "");
                }
            }

            return AJS.contextPath() + "/rest/playsql-spreadsheets/1.0/" + spreadsheet.context + suffix;
        }

        /** Loads the spreadsheet.
         */
        var load = function(urlParameters, requestOptions) {
            if($.browser.msie && $.browser.version <= 8) {
                checkNotNull($placeholder.find(".playsql-message-inner")).text("Internet Explorer 8 is not supported.");
                return;
            }
            spreadsheet.whereClauseConsolidated = Confluence.PlaySQL.Spreadsheet.ColumnToolbar.buildWhereClause(spreadsheet.whereClause, spreadsheet.whereClauseColumnFilters);
            notifyMacroBrowserIfExists(spreadsheet);
            var urlOptions = {
                whereClause : spreadsheet.whereClauseConsolidated,
                orderby : spreadsheet.orderbyClause,
                columns: spreadsheet.displayColumns
            };
            if (!!spreadsheet.userParameters && spreadsheet.userParameters.length > 0) {
                // At first load, there's no param because we haven't fetched the spreadsheet definition yet
                for (var i = 0; i < spreadsheet.userParameters.length ; i++) {
                    var param = spreadsheet.userParameters[i];
                    if (!param.builtIn) {
                        var paramName = "user-parameter-" + param.cleanName;
                        if (typeof param.value != "undefined") {
                            urlOptions[paramName] = param.value;
                        }
                    }
                }
            } else {
                // At first load however, we can send the provided user parameters
                $.each($placeholder[0].attributes, function(){
                    if (this.specified && this.name.indexOf("data-user-parameter-") == 0) {
                        var attrName = this.name.substr("data-".length);
                        urlOptions[attrName] = this.value;
                    }
                });
            }
            $.extend(urlOptions, urlParameters);

            if (!spreadsheet.id) {
                AJS.log("Placeholder has no spreadsheet id: ", $placeholder);
                return;
            }
            var displayCallback = function(data, status, request){
                var colNames = request.getResponseHeader("X-COL-NAMES").split(","),
                    tableDef = data.tableDef;
                /*colLabels = request.getResponseHeader("X-COL-LABELS").split(","),
                 colRenderers = request.getResponseHeader("X-COL-RENDERERS").split(","),
                 colTypes = request.getResponseHeader("X-COL-TYPES").split(","),*/
                function findCol(name) {
                    for (var i in tableDef.columns)
                        if (tableDef.columns[i].name == name)
                            return tableDef.columns[i];
                    throw "Column not found: " + name;
                }
                var columns = colNames.map(function(name, i){
                    var colDefinition = findCol(name);
                    return readColumnDefinition(i, colDefinition, $placeholder);
                });
                spreadsheet.columns = columns;
                spreadsheet.totals = tableDef.totals || [];
                spreadsheet.perms = Confluence.PlaySQL.Helpers.intersect(Confluence.PlaySQL.Helpers.parsePerms(tableDef.perms), spreadsheet.perms);
                spreadsheet.customEmptyMessage = tableDef.customEmptyMessage;
                spreadsheet.userParameters = filterBuiltInParameters(data.userParameters || []);
                spreadsheet.useForm = data.useForm;
                spreadsheet.orderbyClause = data.orderBy || "";

                spreadsheet.defaultCase = request.getResponseHeader("X-DIALECT_DEFAULT_CASE");
                spreadsheet.dialect = request.getResponseHeader("X-DIALECT");
                if (spreadsheet.defaultCase != "LOWER" && spreadsheet.defaultCase != "UPPER") {
                    $("#create-query-button, #create-spreadsheet-button").hide();
                    throw "Invalid default case: " + defaultCase;
                }
                spreadsheet.mode = request.getResponseHeader("X-MODE"); // READ_ONLY or READ_WRITE
                spreadsheet.systemColumnsMode = data.systemColumnsMode;
                spreadsheet.label = tableDef.label;
                spreadsheet.pagination = Confluence.PlaySQL.Helpers.parsePagination(data.pagination);

                renderCells(columns, data.data);

                var errorMessage = data.errorMessage != undefined && data.errorMessage.length > 1 ? (data.errorMessage.charAt(0).toUpperCase() + data.errorMessage.slice(1)) : "";

                var html = Confluence.Templates.PlaySQL.fullTable({
                    name: spreadsheet.id,
                    columns: columns,
                    data: data.data,
                    totals: data.totals,
                    pagination: spreadsheet.pagination,
                    recoverableErrorMessage: errorMessage,
                    customEmptyMessage: spreadsheet.customEmptyMessage,
                    editorUrl: $placeholder.attr("data-editor-url"),
                    userParameters: spreadsheet.userParameters,
                    useForm: spreadsheet.useForm,
                    useCache: data.useCache
                });

                // Remove the Position column
                columns.splice(1, 1);
                for (var index = 0 ; index < columns.length ; index++) {
                    columns[index].index = index;
                }

                $placeholder.html(html);
                bindSpreadsheet();
                triggerDisplayEvents();
                if (errorMessage) {
                    // Success without formulas
                    spreadsheet.options.onDisplay && spreadsheet.options.onDisplay(spreadsheet, "PARTIAL_SUCCESS", data);
                } else {
                    // Total success
                    spreadsheet.options.onDisplay && spreadsheet.options.onDisplay(spreadsheet, "SUCCESS", data);
                }
            };
            var failCallback = function(jqXHR, error, reason){
                if (requestOptions && requestOptions.failureCallback) {
                    if (requestOptions.failureCallback(jqXHR) === false) {
                        return;
                    }
                }
                // Is it a 4xx error?
                if (Math.round(jqXHR.status/100) * 100 == 400) {
                    try {
                        // If the BAD_REQUEST is specially formatted by Play SQL
                        var data = $.parseJSON(jqXHR.responseText);
                        var message;
                        if (data.type == "SQL_SYNTAX") {
                            message = "There is an error in the SQL syntax: " + data.message;// + "\n" + data.sql;
                        } else {
                            message = data.message || "Error";
                        }
                        if (data.userParameters) {
                            spreadsheet.userParameters = filterBuiltInParameters(data.userParameters);
                        }

                        var html = Confluence.Templates.PlaySQL.fullTable({
                            name: spreadsheet.id,
                            columns: [],
                            data: [],
                            totals: undefined,
                            recoverableErrorMessage: message,
                            customEmptyMessage: spreadsheet.customEmptyMessage,
                            editorUrl: $placeholder.attr("data-editor-url"),
                            userParameters: spreadsheet.userParameters
                        });

                        $placeholder.html(html);
                        bindSpreadsheet();
                        triggerDisplayEvents();
                        // The error can be recovered because it comes from the user, and probably the userParameters
                        // were provided
                        spreadsheet.options.onDisplay && spreadsheet.options.onDisplay(spreadsheet, "PARTIAL_FAILURE", data);
                        return;
                    } catch (err) {
                        // Swallow and proceed to the next section
                        AJS.log(err);
                    }
                }
                var message = "Sorry, the spreadsheet couldn't be displayed: " + reason
                        + " (HTTP status code: " + jqXHR.status + ")",
                    $area = $placeholder.find(".playsql-message-inner");
                AJS.logError("PlaySQL error: " + message + "\n" + jqXHR.responseText);
                if ($area.size() > 0) {
                    $area.text(message);
                    $placeholder.find(".hidden").text(jqXHR.responseText);
                } else {
                    Confluence.PlaySQL.Messages.addMessage("error", message, "save-spreadsheet-bean")
                }
                // The response couldn't even be parsed and/or the error isn't 4xx.
                spreadsheet.options.onDisplay && spreadsheet.options.onDisplay(spreadsheet, "FAILURE");
            }

            if (typeof spreadsheet.queryModelOverride != "undefined") {
                // "Preview" mode, where the queryModel is posted with the data
                var url = baseUrl("/data/preview", urlOptions);
                $.ajax({
                    url: url,
                    type: 'POST',
                    contentType: "application/json; charset=utf-8",
                    timeout: 180000,
                    cache: false,
                    contentType: "application/json; charset=utf-8",
                    dataType: "json",
                    data: JSON.stringify(spreadsheet.queryModelOverride)
                }).done(displayCallback).fail(failCallback);

            } else {
                // Normal mode
                var url = baseUrl("/data", urlOptions);
                $.ajax({
                    url: url,
                    type: 'GET',
                    contentType: "application/json; charset=utf-8",
                    timeout: 180000,
                    cache: false
                }).done(displayCallback).fail(failCallback);
            }
            return false;
        }
        var notifyMacroBrowserIfExists = function(){
            var macroFilterInput = window.parent &&
                window.parent.document &&
                window.parent.document.getElementById("macro-param-where-clause");
            if (macroFilterInput) {
                $(macroFilterInput).val(PlaySQL.spreadsheet.whereClauseConsolidated);
            }
        }
        var filterBuiltInParameters = function(userParameters) {
            if (!userParameters) return userParameters;
            var newList = [];
            for (var i = 0 ; i < userParameters.length ; i++) {
                if (!userParameters[i].builtIn) {
                    userParameters[i].cleanName = userParameters[i].name.toLowerCase().replace(/[^a-z0-9]/g, "");
                    newList.push(userParameters[i]);
                }
            }
            return newList;
        }

        // Calls each renderer for the cells
        var renderCells = function(cols, data, totals) {
            for (var rowIndex in data) {
                var row = data[rowIndex],
                    rowMetadataText = row[1][0],
                    rowMetadata;
                try {
                    rowMetadata = AJS.$.parseJSON(rowMetadataText);
                } catch (err) {}
                for (var cellIndex in row) {
                    var cell = row[cellIndex],
                        col = cols[cellIndex],
                        renderer = Confluence.PlaySQL.Spreadsheet.renderers[col.viewRenderer || col.renderer],
                        cellMetadata = rowMetadata && rowMetadata["c_" + col.name];
                    // Cell already contains [FE value, BE value].
                    // Let's add the rendered content
                    cell.push(renderer.renderCellContents(cell[0], spreadsheet, col));
                    // Let's add the css classes
                    cell.push(buildCssClassesForCell(cell[0], cellMetadata));
                    // And the styles
                    cell.push(buildCssStylesForCell(cell[0], cellMetadata));
                }
            }
        }

        /** This is where I bind all events on buttons and get the spreadsheet alive */
        var bindSpreadsheet = function() {
            var editorUrl = AJS.contextPath() + "/playsql-base-plugin/spreadsheet-editor.action?" +
                $.param({key: spreadsheet.context.split(":")[1], spreadsheetId: spreadsheet.id});
            if (!spreadsheet.perms.hasPerm('ADD_COLS')) {
                spreadsheet.$placeholder.find(".add-column a").addClass("hidden");
            }
            if (!spreadsheet.perms.hasPerm('ADD_ROWS')) {
                spreadsheet.$placeholder.find(".totals-row a").addClass("hidden");
            }

            $placeholder.find("thead tr.playsql-col-headers th:first-child").html("<a href='" + editorUrl + "'>&nbsp;</a>");
            var sSelection = Confluence.PlaySQL.Selection.forTable($placeholder),
                alreadyInitialized = sSelection.alreadyInitialized;

            sSelection.init({
                onInit: function(){
                    Confluence.PlaySQL.Selection.attach($placeholder, sSelection, {
                        "click thead tr.playsql-col-headers th.add-column a": addColumn,
                        "click th.add-row a": function($this, e){
                            e.preventDefault();
                            addRow(undefined);
                            return false;
                        },
                        "key f2": displayCellProperties,
                        "key backspace": removeCellContents,
                        "key delete": removeCellContents
                    });
                },
                canSelectFooter: spreadsheet.perms.hasPerm('EDIT_TOTALS'),
                columnSecondClick : columnSecondClick,
                rowSecondClick : rowSecondClick,
                saveCells: function($cells) {
                    // We use a timeout because we need to leave enough time for the DatePicker and the PagePicker to
                    // register the [enter] event, put their value in the $input, and save.
                    window.setTimeout(function(){
                        var totalsModified = false;
                        spreadsheet.requestPipe = spreadsheet.requestPipe || [];
                        $cells.each(function(){
                            var $cell = $(this);
                            if ($cell.is("tr.totals-row > *.editing")) {
                                if (spreadsheet.perms.hasPerm('EDIT_TOTALS')) {
                                    Confluence.PlaySQL.Totals.saveTotal(spreadsheet.totals, $cell);
                                    totalsModified = true;
                                }
                            } else {
                                if (spreadsheet.perms.hasPerm('EDIT_CELLS') || spreadsheet.perms.hasPerm('Q_VIEW_EDIT')) {
                                    var cell = getCell($(this));
                                    cell.renderer.saveAndView();
                                }
                            }
                        });
                        spreadsheet.sprayRequestPipe();
                        if (totalsModified) {
                            saveTotals();
                        }
                    }, 0);
                },
                editCell: function($cell) {
                    if ($cell.parent().is("tr.totals-row") && !$cell.is(":last-child")) {
                        if (!$cell.hasClass("editing")) {
                            if (!spreadsheet.perms.hasPerm('EDIT_TOTALS')) return;
                            if (!$cell.hasClass("aui-dropdown2-trigger") && spreadsheet.totals.length > 0) {
                                var totals = spreadsheet.totals,
                                    menu = {
                                        "edit": ["Edit", Confluence.PlaySQL.Totals.editTotal, totals, $cell],
                                        "addLeft": ["Split cell (on the left)", Confluence.PlaySQL.Totals.addTotal, totals, $cell, -1],
                                        "addRight": ["Split cell (on the right)", Confluence.PlaySQL.Totals.addTotal, totals, $cell, 1],
                                        "remove": ["Remove this total", Confluence.PlaySQL.Totals.removeTotal, totals, $cell, saveTotals]
                                    };
                                if (!(parseInt($cell.attr('colspan')) >= 2))
                                    delete menu["addRight"];
                                if (!(parseInt($cell.attr('colspan')) >= 3))
                                    delete menu["addLeft"];
                                Confluence.PlaySQL.Selection.bindMenu($cell, menu);
                            } else {
                                // Zyva, edit his gddm cell...
                                $cell.removeClass("aui-dropdown2-trigger")
                                    .removeAttr('aria-controls aria-haspopup aria-owns');
                                Confluence.PlaySQL.Totals.editTotal(spreadsheet.totals, $cell);
                            }
                        }
                    } else {
                        // Body cells
                        if (!spreadsheet.perms.hasPerm('EDIT_CELLS') && !spreadsheet.perms.hasPerm('Q_VIEW_EDIT')) return;
                        cellSecondClick($cell);
                    }
                },
                cancelEdit : function($cell) {
                    if ($cell.parent().is("tr.totals-row")) {
                        var rendered = $cell.attr("data-rendered");
                        if ($cell.hasClass("editing") && rendered !== undefined) {
                            Confluence.PlaySQL.Selection.inlineDialog && Confluence.PlaySQL.Selection.inlineDialog.hide();
                            $cell.html(rendered).removeClass("editing");
                        }
                    } else {
                        var cell = getCell($cell);
                        cell.renderer.cancelEdit();
                    }
                },
                getCell : getCell,
                colOf : colOf,
                popupTrigger: ".totals-row .totals-calculation",
                popupContentFn: function($content, trigger, showPopup) {
                    var fields = {},
                        exampleTemplates = ['SUM(%s)', 'ROUND(AVG(%s))', 'MAX(%s)'],
                        exampleTemplatesVarchar = ['MAX(%s)'],
                        examples = [],
                        examplesVarchar = [],
                        fieldName, type, template,
                        isSystemField,
                        tableName = spreadsheet.id;
                    for (var i in spreadsheet.columns) {
                        fieldName = spreadsheet.columns[i].name;
                        isSystemField = spreadsheet.columns[i].statusPK || spreadsheet.columns[i].statusPosition;
                        type = spreadsheet.columns[i].type;
                        if (typeof type == "undefined") {
                            type = "\"Undefined\"";
                        }
                        if (spreadsheet.defaultCase == "LOWER" && fieldName.toLowerCase() != fieldName
                            || spreadsheet.defaultCase == "UPPER" && fieldName.toUpperCase() != fieldName) {
                            fieldName = '"' + fieldName + '"';
                        }
                        fields[type] = (fields[type] || []).concat(fieldName);
                        if (!isSystemField) {
                            if (type == 'INTEGER' || type == 'NUMERIC') {
                                template = exampleTemplates[examples.length % exampleTemplates.length];
                                examples.push(template.replace('%s', fieldName));
                            } else if ((type == 'VARCHAR' || type == 'TEXT') && examplesVarchar.length < exampleTemplatesVarchar.length) {
                                template = exampleTemplatesVarchar[examplesVarchar.length % exampleTemplatesVarchar.length];
                                examplesVarchar.push(template.replace('%s', fieldName));
                            }
                        }
                    }
                    if (spreadsheet.defaultCase == "LOWER" && tableName.toLowerCase() != tableName
                        || spreadsheet.defaultCase == "UPPER" && tableName.toUpperCase() != tableName)
                        tableName = '"' + tableName + '"';

                    $content.unbind()
                        .html(Confluence.Templates.PlaySQL.formulaHelp({fields:fields, examples:examples.concat(examplesVarchar), tableName: tableName}));
                    $content.on("click", ".examples a", function() {
                        $(trigger).val($(this).text()).focus();
                        Confluence.PlaySQL.Selection.inlineDialog && Confluence.PlaySQL.Selection.inlineDialog.hide();
                        return false;
                    });
                    showPopup();
                },
                pasteHandler: pasteHandler,
                addRow : addRow,
                updateColumnWidth : updateColumnWidth
            });

            if (!alreadyInitialized) {
                $placeholder.on("click", ".pagination-number", function (e) {
                    e.preventDefault();
                    var offset = parseInt($(this).data("offset"));
                    if (!isNaN(offset)) {
                        // (...) As simple as that...
                        // So much tame weitin afore plementin't...
                        spreadsheet.load({offset: offset});
                    }
                });
            }

            // Also bind the user parameters
            $placeholder.find(".playsql-user-parameters-form").on("click", ".submit", function(e){
                e.preventDefault();
                var $form = $(this).closest(".playsql-user-parameters-form");
                if (!!spreadsheet.userParameters) {
                    for (var i = 0 ; i < spreadsheet.userParameters.length ; i++) {
                        var param = spreadsheet.userParameters[i];
                        param.value = $form.find("#user-parameter-" + param.cleanName).val();
                    }
                }
                spreadsheet.load();
            });
            $placeholder.find("#user-parameter-refresh-query").on("click", function(e){
                e.preventDefault();
                spreadsheet.load({"refreshCache": "true"});
            });

            // Trigger event, so function handlers can be overridden
            // We pass the sSelection object rather than just override handlers because they may need the original
            // handlers
            spreadsheet.options.onSelectionSet && spreadsheet.options.onSelectionSet(sSelection);
        };

        var saveTotals = function() {
            var totals = spreadsheet.totals,
                urlOptions = {};

            if (!!spreadsheet.userParameters) {
                for (var i = 0; i < spreadsheet.userParameters.length ; i++) {
                    var param = spreadsheet.userParameters[i];
                    if (!param.builtIn && typeof param.value != "undefined") {
                        var paramName = "user-parameter-" + param.cleanName;
                        urlOptions[paramName] = param.value;
                    }
                }
            }

            var url = baseUrl("/totals", urlOptions);

            Confluence.PlaySQL.Totals.updateColspans(totals, spreadsheet.$placeholder);

            var successCallback = function(data, status, request){
                if (!data.message) {
                    var $footer = $(Confluence.Templates.PlaySQL.totalsRow({
                        totals: data.totals,
                        pagination: spreadsheet.pagination,
                        columnCount: spreadsheet.columns.length - 1
                    }));
                    $placeholder.find("tfoot").html($footer.contents());
                    if (!spreadsheet.perms.hasPerm("ADD_ROWS")) {
                        $placeholder.find(".totals-row a").addClass("hidden");
                    }
                    spreadsheet.options.onChangeTotals && spreadsheet.options.onChangeTotals(spreadsheet);
                } else {
                    var message = data.errorMessage;
                    message = message.replace("ERROR: ", "");
                    $placeholder.find("tfoot td").first().text(message + ". Edit the total again to fix the formula.");
                }
            };
            var failureCallback = function(jqXHR, error, reason){
                if (jqXHR.status == 400) {
                    try {
                        var data = JSON.parse(jqXHR.responseText);
                        if (data != null) {
                            var message = data.message || "Error";
                            message = message.replace("ERROR: ", "");
                            if (data.type == "SQL_SYNTAX") {
                                message += " - Edit the total again to fix the formula.";
                            }
                            $placeholder.find("tfoot td").first().text(message);
                            return;
                        }
                    } catch (err) {
                        // Can't manage the JSON. Proceed to the next section
                    }
                }
                Confluence.PlaySQL.Messages.addMessage("error", "The settings for the totals couldn't be saved",
                    "change-totals", jqXHR.responseText);
            };

            if (spreadsheet.options.saveTotals) {
                spreadsheet.options.saveTotals(spreadsheet, urlOptions, successCallback, failureCallback);
            } else {
                $.ajax({
                    url: url,
                    type: 'PUT',
                    contentType: "application/json; charset=utf-8",
                    dataType: "json",
                    data: JSON.stringify(totals)
                }).done(successCallback)
                    .fail(failureCallback);
            }
        }

        var updateTotals = function() {
            if (!spreadsheet.perms.hasPerm('EDIT_TOTALS')) return;
            var url = baseUrl("", { fields: "totals" });
            $.ajax({
                url: url,
                type: 'GET',
                contentType: "application/json; charset=utf-8",
                cache: false
            }).done(function(data, status, request){
                var $footer = $(Confluence.Templates.PlaySQL.totalsRow({
                    totals: data.totals,
                    pagination: spreadsheet.pagination,
                    columnCount: spreadsheet.columns.length - 1
                }));
                $placeholder.find("tfoot").html($footer.contents());
            }).fail(function(jqXHR, error, reason){
                Confluence.PlaySQL.Messages.addMessage("error", "The totals couldn't be retrieved",
                    "change-totals", jqXHR.responseText);
            });
        }

        var triggerDisplayEvents = function() {
            for (var i in spreadsheet.columns) {
                var column = spreadsheet.columns[i],
                    renderer = checkNotNull(Confluence.PlaySQL.Spreadsheet.renderers[column.viewRenderer || column.renderer]);
                renderer.onColumnDisplay(spreadsheet, column);
            }

            if (spreadsheet.selectedRecord) {
                var $row = $placeholder.find("tr[data-row-id=" + spreadsheet.selectedRecord + "] > *:not(.right-padding)");
                $row.addClass("selected");
                $row.first().addClass("focused");
                spreadsheet.selectedRecord = undefined;
            }
        };

        var cellSecondClick = function ($cell) {
            var cell = getCell($cell);
            if (cell) {
                if (!$cell.hasClass("editing")) {
                    if (!spreadsheet.perms.hasPerm('EDIT_CELLS') && !spreadsheet.perms.hasPerm('Q_VIEW_EDIT', cell.column)) return;
                    cell.renderer.edit();
                    cell.renderer.focus();
                } else if ($cell.find("*:focus").length == 0) {
                    cell.renderer.focus();
                }
            }
        }

        var columnSecondClick = function($header) {
            if ($header.hasClass("editing"))
                return;
            var column = colOf($header);
            spreadsheet.inlinedialog = Confluence.PlaySQL.Spreadsheet.ColumnToolbar.create(spreadsheet, column, $header);
            spreadsheet.inlinedialog.show();
        }

        var updateColumnWidth = function($header, width) {
            var column = colOf($header),
                newColumn = {
                    name: column.name,
                    label: column.label,
                    type: column.type,
                    formula: column.formula,
                    renderer: column.renderer,
                    length: column.length,
                    formats: column.formats,
                    width: width
                };
            // Attempt to save, don't tell anything in failure/success
            saveColumn($header, column, newColumn, ["WIDTH"]);
        }

        var rowSecondClick = function($header) {
            var rowIds = $placeholder
                    .find(".selected")
                    .closest("tr")
                    .map(function(index, tr){return $(tr).attr("data-row-id");}),
                menu = {
                    "delete": ["Delete", deleteRows, rowIds]
                };
            checkNotNull($header.parent().attr('data-row-id'));
            if (!spreadsheet.perms.hasPerm('REMOVE_ROWS')) delete menu["delete"];
            if (Object.keys(menu).length > 0)
                Confluence.PlaySQL.Selection.bindMenu($header, menu);
        }

        /**
         * Finds a <tr/>
         * @param position the position
         * @returns $row
         */
        var findRowByPosition = function(position) {
            var $row;
            if (position != undefined) {
                $row = $placeholder.find("tr[data-row-position='" + position + "']");
            } else {
                throw "Not implemented";
            }
            if ($row.size() == 0)
                throw "Couldn't find the row: " + position;
            return $row;
        }

        var findRowByPk = function(jointSpreadsheetKey, pk) {
            if (arguments.length == 1) {
                pk = jointSpreadsheetKey;
                jointSpreadsheetKey = undefined;
            }
            if (jointSpreadsheetKey == undefined && typeof pk === "string" && pk.indexOf("-") != -1) {
                jointSpreadsheetKey = pk.substr(0, pk.indexOf("-"));
                pk = pk.substr(pk.indexOf("-") + 1);
            }
            if (pk instanceof $ && pk.is("tr")) {
                return pk;
            } else if (jointSpreadsheetKey != undefined) {
                for (var i in spreadsheet.columns) {
                    var formats = spreadsheet.columns[i].formats;
                    if (formats["joint-table"] == jointSpreadsheetKey) {
                        var nthChild = spreadsheet.columns[i].index + 1, // the nth-child is 1-based
                            $jointIdCell = $placeholder.find("td:nth-child(" + nthChild + ")[data-value='" + checkNotNull(pk) + "']");
                        return checkNotNull($jointIdCell, pk).parent();
                    }
                }
                throw "Can't find an ID column for " + spreadsheetKey;
            } else {
                var $row = $placeholder.find("tr[data-row-id='" + checkNotNull(pk) + "']");
                return checkNotNull($row, pk);
            }
        }

        var getRecordIds = function($cell, useJointId) {
            var id = $cell.closest("tr").attr("data-row-id"),
                result = {id:id};
            if (useJointId) {
                var cell = spreadsheet.getCell($cell),
                    jointId = cell.column.formats["joint-id"],
                    jointTable = cell.column.formats["joint-table"];
                if (jointTable) {
                    // This cell is the joint id
                    result.jointId = cell.$cell.attr("data-value");
                    result.jointIdName = cell.column.name;
                    result.jointIdCell = cell;
                    result.jointTable = jointTable;
                    result.jointFk = cell.column.formats["joint-fk"];
                } else if (jointId) {
                    var idCell = spreadsheet.getCell(id, jointId);
                    if (!idCell) throw "There is no cell for the column " + jointId + " on the row " + $cell.closest("tr").attr("data-row-id");
                    var idValue = idCell.$cell.attr("data-value");
                    result.jointId = idValue;
                    result.jointIdName = jointId;
                    result.jointIdCell = idCell;
                    result.jointTable = idCell.column.formats["joint-table"];
                    result.jointFk = idCell.column.formats["joint-fk"];
                }
            }
            return result;
        }

        /** Returns a cell handler.
         * Syntax: getCell($td) returns {cell} - Please pass the td, not an inner element.
         * Syntax: getCell($row, colName) returns {cell}
         * Syntax: getCell($row, colNumber) returns {cell}
         * Syntax: getCell([jointSpreadsheetKey,] rowPrimaryKey, colName) returns {cell}
         * Syntax: getCell([jointSpreadsheetKey,] rowPrimaryKey, colNumber) returns {cell}
         */
        var getCell = function(jointSpreadsheetKey, row, col) {
            var $cell,
                $row,
                column,
                isColumnHeader,
                renderer,
                rendererInstance,
                viewRenderer,
                viewRendererInstance,
                rowId;
            if (arguments.length == 2) {
                col = arguments[1];
                row = arguments[0];
                jointSpreadsheetKey = undefined;
                if (typeof row === "string" && row.indexOf("-") != -1) {
                    jointSpreadsheetKey = row.substr(0, row.indexOf("-"));
                    $row = findRowByPk(row);
                }
            } else if (arguments.length == 1) {
                row = arguments[0];
                jointSpreadsheetKey = undefined;
            }
            if (col == undefined) {
                if (row == undefined) return undefined;
                $cell = row;
                if (!$cell instanceof $)
                    throw "Cell isn't a jQuery object: " + $cell;
                if (!$cell.is("td,th")) {
                    AJS.log("Cell isn't a td/th: " + $cell + " (it's a " + $cell.prop("tagName") + ")");
                    return undefined;
                }
                if (!$cell.closest(".playsql-spreadsheet-placeholder").is(spreadsheet.$placeholder)) {
                    AJS.log("Cell isn't on the current spreadsheet");
                    return undefined;
                }
            } else if (jointSpreadsheetKey != undefined) {
                var $row = findRowByPk(jointSpreadsheetKey, row);

                if ($.isNumeric(col)) {
                    $cell = $row.children().eq(col);
                } else {
                    // The 'col' name is the name of the joint column
                    var colDefinition = colOf(jointSpreadsheetKey, col);
                    if (colDefinition == null) return undefined;
                    $cell = $row.children().eq(colDefinition.index);
                }
            } else {
                $row = findRowByPk(row);
                if ($.isNumeric(col)) {
                    $cell = $row.children().eq(col);
                } else {
                    var col2 = colOf(col);
                    if (col2 == undefined) return undefined;
                    $cell = $row.children().eq(col2.index);
                }
            }
            if ($cell.index() >= spreadsheet.columns.length) {
                return undefined;
            } else if ($cell.find(".playsql-message-inner").size() > 0) {
                return undefined;
            }
            column = spreadsheet.columns[$cell.index()];
            isColumnHeader = $cell.is("thead > tr.playsql-col-headers > th");
            renderer = isColumnHeader
                ? Confluence.PlaySQL.Spreadsheet.renderers["header"]
                : Confluence.PlaySQL.Spreadsheet.renderers[column.renderer];
            rendererInstance = renderer.forCell(spreadsheet, $cell);
            if (!isColumnHeader && !!column.formula && column.viewRenderer) {
                viewRenderer = Confluence.PlaySQL.Spreadsheet.renderers[column.viewRenderer];
                viewRendererInstance = viewRenderer.forCell(spreadsheet, $cell);
            }
            rowId = $cell.closest("tr").attr("data-row-id");
            return {
                $cell: $cell,
                isColumnHeader: isColumnHeader,
                column : column,
                renderer : checkNotNull(rendererInstance, column.renderer),
                viewRenderer : viewRendererInstance,
                rowId: rowId
            }
        }

        var getRowMetadata = function($cell) {
            return $cell.closest("tr").data("row-metadata");
        }
        var getCellMetadata = function($cell, path) {
            var rowMetadata = getRowMetadata($cell),
                column = spreadsheet.columns[$cell.index()];
            var cellMetadata = rowMetadata && rowMetadata["c_" + column.name];
            if (cellMetadata && path) {
                return cellMetadata[path];
            } else {
                return cellMetadata;
            }
        }

        /** Given a cell (or header cell), find the column.
         * Usage: colOf($cell)
         * Usage: colOf("firstcolumn")
         * Usage: colOf(3)
         * Usage: colOf("spreadsheetKey", "firstcolumn")
         * */
        var colOf = function(spreadsheetKey, $cell) {
            if ($cell == undefined) {
                $cell = spreadsheetKey;
                spreadsheetKey = null;
            }
            if ($cell instanceof $)
                return spreadsheet.columns[$cell.index()];
            else if (typeof $cell == "string") {
                var colName = $cell;
                if (spreadsheetKey) {
                    var jointIdColumn = {};
                    for (var i in spreadsheet.columns) {
                        var formats = spreadsheet.columns[i].formats;
                        if (formats["joint-table"] == spreadsheetKey) {
                            jointIdColumn = spreadsheet.columns[i];
                            break;
                        }
                    }
                    for (var i in spreadsheet.columns) {
                        var formats = spreadsheet.columns[i].formats,
                            jointColumnName = formats["joint-column"] || spreadsheet.columns[i].name;
                        if (formats["joint-id"] == jointIdColumn.name && jointColumnName == colName) {
                            return spreadsheet.columns[i];
                        }
                    }
                    return undefined;
                } else {
                    for (var i in spreadsheet.columns) {
                        if (spreadsheet.columns[i].name == colName)
                            return spreadsheet.columns[i];
                    }
                }
                return undefined;
            } else if (typeof $cell == "number") {
                var colIndex = $cell;
                return spreadsheet.columns[colIndex];
            }
            throw "colOf: Couldn't parse the argument: " + $cell;
        }

        /** Returns all cells for the provided column definition */
        var $findColumnCells = function(column, includeHeader) {
            if (!column) return $();
            var $allCells = this.$placeholder.find("tbody td:nth-child(" + (column.index+1) + ")");
            if (includeHeader) {
                $allCells = $allCells.add(this.$placeholder.find("thead th[data-name='" + column.name + "']"));
            }
            return $allCells;
        }

        var addRow = function(beforePosition) {
            addRows(1, beforePosition);
        }

        var addRows = function(count, beforePosition) {
            if (!spreadsheet.perms.hasPerm('ADD_ROWS')) return false;
            // Remove the message placeholder if present
            var $messageRow = $placeholder.find("tr.playsql-message");
            if ($messageRow.size() != 0) $messageRow.remove();

            var data = [],
                $newRows = [],
                colNames = spreadsheet.columns.map(function(column){return column.name;});
            colNames.splice(1, 0, "POSITION");

            for (var j = 0; j < count; j++) {
                var row = emptyRow(colNames.length);

                // Make the row
                var template = Confluence.Templates.PlaySQL.oneRow({row:row}),
                    $row;
                // Insert the row
                if (beforePosition)
                    $row = $(template).insertBefore(findRowByPosition(beforePosition));
                else
                    $row = $(template).appendTo($placeholder.find("tbody"));
                // Edit all cells
                $row.children("td").first().each(function(i) {
                    var $cell = $(this),
                        cell = getCell($cell);
                    cell.renderer.edit();
                    cell.renderer.focus();
                    $placeholder.find(".focused").removeClass("focused selected");
                    $cell.addClass("focused selected");
                });
                // Spin the row header
                Confluence.PlaySQL.Messages.spin($row.find("th:first")).on(undefined, "The row is being created on the server side. You cannot submit cells until the row is ready.");
                // Keep'em until we get the answer from the REST servlquest
                data[j] = row.splice(1, 1); // We don't send much data because we're just creating rows
                $newRows[j] = $row;
            }
            colNames = ["POSITION"];

            // Ajax to get the ID
            var url = baseUrl("/data");
            $.ajax({
                url: url,
                type: 'POST',
                contentType: "application/json; charset=utf-8",
                dataType: "json",
                data: JSON.stringify(data),
                headers: {
                    "X-COL-NAMES": colNames
                }
            }).done(function(data, status, request){
                for (var i = 0 ; i < data.insertedData.length && i < $newRows.length ; i++) {
                    var idCell = getCell($newRows[i], 0);
                    idCell.renderer.setOriginalValue(data.insertedData[i], data.insertedData[i]);
                }
                dispatchCellValues(data.data, data.tableDef);
            }).fail(function(jqXHR, error, reason){
                Confluence.PlaySQL.Messages.addMessage("error", "The row couldn't be created.", "add-row", jqXHR.responseText);
                for (var i = 0 ; i < $newRows.length ; i++) {
                    var $row = $newRows[i];
                    if ($row.text().length < 10) {
                        $row.remove();
                    }
                    $row.children("td:not(:last-child)").addClass("invalid").attr("title", "The row couldn't be created.");
                }
            });
            return false;
        };

        var deleteRows = function (rowIds){
            rowIds.each(function(index, rowId) {
                deleteRow(rowId);
            });
        };

        var deleteRow = function (rowId){
            if (!spreadsheet.perms.hasPerm('REMOVE_ROWS')) return;
            var $row = findRowByPk(rowId),
                url = baseUrl("/data", checkNotNull(rowId));

            var spinner = Confluence.PlaySQL.Messages.spin($row.children().first())
                .on(500, "Deleting the row");
            $.ajax({
                url: url,
                type: 'DELETE'
            }).done(function(data, status, request){
                $row.remove();
                dispatchCellValues(data.data, data.tableDef);
            }).fail(function(jqXHR, error, reason){
                spinner.off();
                Confluence.PlaySQL.Messages.addMessage("error", "The row couldn't be deleted because of a " +
                "communication error with the server. ", "save-spreadsheet-bean", jqXHR.responseText);
            });
        };

        var addJointRow = function(argument, callback) {
            if (!spreadsheet.perms.hasPerm("Q_VIEW_EDIT")) return;
            var $cell, cell;
            if (argument instanceof $) {
                $cell = argument;
                cell = spreadsheet.getCell($cell);
            } else if (argument && argument.$cell) {
                cell = argument;
                $cell = cell.$cell;
            } else {
                throw "Not implemented: addJointRow with argument " + argument;
            }
            Confluence.PlaySQL.Messages.spin($cell).on(500, "The row is being created in the table on the server side. You cannot submit values until the row is ready.");
            if (!cell || !cell.column.formats) {
                Confluence.PlaySQL.Messages.addMessage("error", "It seems like the properties of the joint table are incorrect. Please run the Joint Table wizard again.", "add-row");
                return;
            }
            var jointTable = cell.column.formats["joint-table"],
                jointFk = cell.column.formats["joint-fk"],
                pk = cell.rowId;

            // Ajax to get the ID
            var url = AJS.contextPath() + "/rest/playsql-spreadsheets/1.0/" + spreadsheet.context + "/" + jointTable + "/data",
                colNames = "POSITION," + jointFk,
                data = [[[], [pk]]];
            $.ajax({
                url: url,
                type: 'POST',
                contentType: "application/json; charset=utf-8",
                dataType: "json",
                data: JSON.stringify(data),
                headers: {
                    "X-COL-NAMES": colNames
                }
            }).done(function(data, status, request){
                Confluence.PlaySQL.Messages.spin($cell).off();
                if (data.insertedData.length != 1) {
                    Confluence.PlaySQL.Messages.addMessage("error", "The row couldn't be created.", "add-row", jqXHR.responseText);
                } else {
                    cell.renderer.setOriginalValue(data.insertedData[0], data.insertedData[0]);
                }
                callback && callback();
                spreadsheet.dispatchCellValues(data.data, data.tableDef);
            }).fail(function(jqXHR, error, reason){
                Confluence.PlaySQL.Messages.spin($cell).off();
                Confluence.PlaySQL.Messages.addMessage("error", "The row couldn't be created.", "add-row", jqXHR.responseText);
            });
        }

        var removeJointRow = function($cell) {
            if (!spreadsheet.perms.hasPerm("Q_VIEW_EDIT")) return;
            var spinner = Confluence.PlaySQL.Messages.spin($cell).on(500, "Deleting the row");
            var cell = spreadsheet.getCell($cell);
            if (!cell || !cell.column.formats) {
                spinner.off();
                Confluence.PlaySQL.Messages.addMessage("error", "It seems like the properties of the joint table are incorrect. Please run the Joint Table wizard again.", "remove-row");
                return;
            }
            var jointTable = cell.column.formats["joint-table"],
                jointId = cell.column.name,
                pk = $cell.attr("data-value");
            if (!pk) return;

            // Ajax to remove the ID
            var url = AJS.contextPath() + "/rest/playsql-spreadsheets/1.0/" + spreadsheet.context + "/" + jointTable + "/data/" + pk;

            $.ajax({
                url: url,
                type: 'DELETE'
            }).done(function(data, status, request){
                spinner.off();
                cell.renderer.setOriginalValue(null);
                for (var i in spreadsheet.columns) {
                    var column = spreadsheet.columns[i];
                    if (column.formats["joint-id"] == jointId) {
                        var cellToDelete = spreadsheet.getCell($cell.parent(), column.index);
                        cellToDelete.$cell.empty().removeAttr("data-value").removeClass("editing saving invalid");
                    }
                }
                spreadsheet.dispatchCellValues(data.data, data.tableDef);
            }).fail(function(jqXHR, error, reason){
                spinner.off();
                Confluence.PlaySQL.Messages.addMessage("error", "The row couldn't be deleted.", "remove-row", jqXHR.responseText);
            });
        }

        var recalculateFormulas = function($cells) {
            var rowId = $cells.size() == 1 ? $cells.closest("tr").attr("data-row-id") : undefined,
                options = { reindex: true };
            if (rowId != undefined) options.pkvalue = rowId;
            var url = spreadsheet.baseUrl("/metadata", options);
            $.ajax({
                url: url,
                type: 'PUT',
                dataType: "json"
            }).done(function(data, status, request){
                dispatchCellValues(data.data, data.tableDef);
            }).fail(function(jqXHR, error, reason){
                var data,
                    message;

                if (jqXHR.status == 400 || (jqXHR.status == 500 && jqXHR.responseText.charAt(0) == '{')) {
                    try {
                        data = JSON.parse(jqXHR.responseText);
                        if (data != null) {
                            message = data.message || "Error";
                        }
                    } catch (err) {
                        // Can't understand the json. No worries, keep going.
                    }
                }

                if (data && data.type == "FORMULA_ERROR") {
                    // If the cell couldn't be found, display the message with a popup
                    Confluence.PlaySQL.Messages.addMessage("error", message, jqXHR.responseText);
                    return;
                }
            });
        }
        /**
         * Saves some cell's metadata, such as the formula
         * @param $cell
         * @param name name of the property
         * @param value the value, or null if the property must be deleted
         * @return true if some data was changed/saved, false otherwise
         */
        var saveCellMetadata = function($cell, name, value) {
            if (!$cell.is("td") || $cell.hasClass("right-padding")) return;
            if (!spreadsheet.perms.hasPerm('EDIT_ROWS_METADATA')) return;
            var rowId = $cell.parent().attr("data-row-id"),
                column = checkNotNull(colOf($cell), $cell),
                // We make a deep copy because we don't want to modify the original metadata until it's
                // saved on the server side
                rowMetadata = $.extend(true, {}, getRowMetadata($cell)) || {},
                colName2 = "c_" + column.name;
            // If an update request was already piped, start with the rowMetadata of that request
            spreadsheet.requestPipe = spreadsheet.requestPipe || [];
            var indexPreviousRequest = undefined;
            for (var i in spreadsheet.requestPipe || []) {
                var req = spreadsheet.requestPipe[i];
                if (req.column == "POSITION" && req.rowId == rowId) {
                    rowMetadata = req.value ? $.parseJSON(req.value) : {};
                    indexPreviousRequest = i;
                    break;
                }
            }
            var property = {};
            property[name] = value;
            var properties = rowMetadata[colName2];
            if (!properties) {
                if (value == undefined) {
                    // No data to modify, there are no cellMetadata and the value is null.
                    return false;
                }
                rowMetadata[colName2] = property;
            } else {
                if (value == undefined || value == null) {
                    if (properties[name] == undefined) {
                        // No data to modify, there are cellMetadata but none for this property
                        return false;
                    }
                    delete properties[name];
                    if (Object.keys(properties).length == 0) {
                        delete rowMetadata[colName2];
                    }
                } else if (properties[name] == value) {
                    // No data to modify
                    return false;
                } else {
                    properties[name] = value;
                }
            }
            if (Object.keys(rowMetadata).length == 0) {
                rowMetadata = null;
            }

            // Now update the value in the POSITION column.
            if (indexPreviousRequest != undefined) {
                spreadsheet.requestPipe.splice(indexPreviousRequest, 1);
            }
            spreadsheet.requestPipe.push({
                spreadsheet: undefined,
                rowId : rowId,
                column : "POSITION",
                value: rowMetadata != null ? JSON.stringify(rowMetadata) : null
            });
            return true;
        }
        var saveCell = function($cell, value) {
            var rowId = $cell.parent().attr("data-row-id"),
                column = checkNotNull(colOf($cell), $cell),
                jointIdColumnName = column.formats && column.formats["joint-id"],
                data = [[value]],
                spreadsheetKey = null,
                columnName = column.name;
            if (jointIdColumnName) {
                columnName = column.formats["joint-column"] || columnName;
                var jointIdColumn = checkNotNull(colOf(jointIdColumnName), jointIdColumnName);
                spreadsheetKey = jointIdColumn.formats && jointIdColumn.formats["joint-table"];
                rowId = $cell.parent().children().eq(jointIdColumn.index).attr("data-value");
                if (rowId == null) {
                    Confluence.PlaySQL.Messages.addMessage("error", "There is no row for the cell which is being saved.");
                    return;
                }
            }

            spreadsheet.requestPipe = spreadsheet.requestPipe || [];
            spreadsheet.requestPipe.push({
                spreadsheet: spreadsheetKey,
                rowId: rowId,
                column: columnName,
                value: value
            });
        }

        /** Processes the spreadsheet.requestPipe queue. It's in the form:
         * - spreadsheet (can be undefined)
         * - rowId
         * - column
         * - value (the value itself, no subarray)
         *
         * In return, call dispatchCellValues() or mark the cells in error.
         */
        var sprayRequestPipe = function() {
            if (spreadsheet.requestPipe == undefined) return;
            var requestPipe = spreadsheet.requestPipe,
                rows = [],// Rows' index is "spreadsheetid-rowid"
                req, row, url, rowId,
                colNames, data;
            spreadsheet.requestPipe = undefined;
            // Group by row
            for (var i = 0; i < requestPipe.length ; i++) {
                req = requestPipe[i];
                if (req == null) continue;
                // req.spreadsheet may be null or undefined
                rowId = (req.spreadsheet ? req.spreadsheet + "-" : "-") + req.rowId;
                rows[rowId] = rows[rowId] || [];
                rows[rowId].push(req);
            }

            // Submit... well, submit all rows... yeah, yeah separately, because
            // Postgres doesn't like recovering from an exception in a transaction, so we wouldn't know which
            // cells fail... that's gonna flush...
            for (var i in rows) {
                //!\ this 'i' in the foreach loop takes the list of keys, it's very different from 0..rows.length because
                // the updates rows are not continuous.
                row = rows[i];
                url = !row[0].spreadsheet ? baseUrl("/data", checkNotNull(row[0].rowId))
                                          : baseUrlForContext(row[0].spreadsheet, "/data", checkNotNull(row[0].rowId)),
                    colNames = row.map(function(cell) {return cell.column;}).join(",");
                data = row.map(function(cell) {return [cell.value];});
                $.ajax({
                    url: url,
                    type: 'PUT',
                    contentType: "application/json; charset=utf-8",
                    dataType: "json",
                    data: JSON.stringify(data),
                    headers: {
                        "X-COL-NAMES" : colNames
                    }
                }).done(function(data, status, request){
                    dispatchCellValues(data.data, data.tableDef);
                }).fail(function(jqXHR, error, reason){
                    var colNames = this.headers["X-COL-NAMES"].split(","),
                        urlParts = this.url.split("/"),
                        spreadsheetKey = urlParts[urlParts.length - 3],
                        spreadsheetKey2 = spreadsheetKey != spreadsheet.id ? spreadsheetKey : null,
                        rowId = urlParts[urlParts.length - 1],
                        rowId2 = spreadsheetKey2 && (spreadsheetKey2 + "-" + rowId) || rowId,
                        colName, cell,
                        data,
                        message;

                    if (jqXHR.status == 400 || (jqXHR.status == 500 && jqXHR.responseText.charAt(0) == '{')) {
                        try {
                            data = JSON.parse(jqXHR.responseText);
                            if (data != null) {
                                message = data.message || "Error";
                            }
                        } catch (err) {
                            // Can't understand the json. No worries, keep going.
                        }
                    }

                    for (var i = 0; i < colNames.length; i++) {
                        colName = $.trim(colNames[i]);
                        if (colName == "POSITION") {
                            // If the POSITION column was changed, it's a bit complicated.
                        } else {
                            cell = getCell(spreadsheetKey2, rowId, colName);
                            cell.renderer.markInError(message);
                        }
                    }

                    if (data && data.type == "FORMULA_ERROR") {
                        if (colName == "POSITION") {
                            var $row = findRowByPk(spreadsheetKey2, rowId),
                                $cellsWithFormula = $row && $row.find(".saving.formula"),
                                cell = $cellsWithFormula && getCell($cellsWithFormula),
                                errorList = data.errorList,
                                firstError = errorList && errorList.length > 0 && errorList[0],
                                firstErrorPosition = firstError && (firstError.line == 1) && firstError.position;
                            if (cell) {
                                cell.renderer.markInError(message);
                                cell.renderer.focus(firstErrorPosition);
                                return;
                            }
                        }
                        // If the cell couldn't be found, display the message with a popup
                        Confluence.PlaySQL.Messages.addMessage("error", message, jqXHR.responseText);
                        return;
                    }
                });
            }
        };

        /** Shared method between saveCell, sprayRequestPipe and insert/remove rows:
         * It reads the data and the tableDef and updates them all. It's useful in order
         * to update all dependencies of a formula, whether you're inserting, deleting or
         * editing a row.
         * @param data the data - usually the data.data property of the JSON response
         * @param tableDef the table definition - usually the data.tableDef property.
         */
        function dispatchCellValues(data, tableDef) {
            for (var rowNum in data) {
                var id,
                    row = data[rowNum],
                    jointSpreadsheetKey = tableDef.name != spreadsheet.id ? tableDef.name : undefined,
                    rowMetadata = undefined;
                for (var i in tableDef.columns) {
                    var col = tableDef.columns[i];
                    if (col.statusPK) {
                        id = row[i][0];
                    } else if (id != undefined && col.statusPosition) {
                        rowMetadata = row[i][0];
                        try {
                            rowMetadata = $.parseJSON(rowMetadata) || {};
                        } catch (err) {
                            rowMetadata = {};
                        }
                        var $row = findRowByPk(jointSpreadsheetKey, id);
                        $row.data("row-metadata", rowMetadata);
                    } else if (id != undefined && !col.statusPosition) {
                        var cell = getCell(jointSpreadsheetKey, id, col.name),
                            cellMetadata = cell && getCellMetadata(cell.$cell);
                        if (cell == undefined) {
                            // Because of formulas and joint cells, it could happen that we get cells which aren't on the current page
                            continue;
                        } else if (!cell.$cell.hasClass("editing")) {
                            cell.renderer.setOriginalValue(row[i][0], row[i][1]);
                        }
                        cell.$cell.attr("style", buildCssStylesForCell(undefined, cellMetadata) || "");
                        if (cellMetadata && cellMetadata["formula"]) {
                            cell.$cell.addClass("formula");
                        } else {
                            cell.$cell.removeClass("formula");
                        }
                    }
                }
            }
            AJS.$.debounce(updateTotals, 500);
        }

        var saveColumn = function($header, column, newColumn, changes, successCallback, failureCallback) {
            function innerSuccessCallback(oldName) {
                if (changes.indexOf("RENAME") != -1 || changes.indexOf("CHLBL") != -1) {
                    // $header exists when called from the inline dialog, and doesn't when called
                    // from the "renderer-specific" sub-inline dialog, which doesn' matter because
                    // we don't modify the label there and the caller redraws the whole column
                    // anyway.
                    if ($header) {
                        var cell = checkNotNull(getCell($header));
                        cell.renderer.view();
                    }
                }
                // When renaming, if any row uses this column, then the full spreadsheet needs to be reloaded for
                // the metadata.
                if (changes.indexOf("RENAME") != -1) {
                    var rowsWithMetadata = $placeholder.find("tr[data-row-metadata]").filter(function() {
                        var metadata = $(this).data("row-metadata");
                        if (!metadata) return false;
                        if (!oldName || JSON.stringify(metadata).indexOf(oldName) == -1) return false;
                        return true;
                    });
                    if (rowsWithMetadata.length > 0) {
                        spreadsheet.load();
                    }
                }
                if (changes.indexOf("CHFMA") != -1) {
                    testSpreadsheet(column.name, function (success, value) {
                        if (success) {
                            spreadsheet.load();
                            successCallback && successCallback();
                        } else {
                            failureCallback && failureCallback(value);
                        }
                    });
                } else {
                    // in 'else' because only a CHFMA would be asynchronous
                    successCallback && successCallback();
                }
            }
            if (spreadsheet.options.saveColumn) {
                spreadsheet.options.saveColumn(spreadsheet, column, newColumn, changes,
                    innerSuccessCallback, failureCallback);
            } else {
                var url = spreadsheet.baseUrl("/cols", checkNotNull(column.name), {
                        confirm: changes.join(",")
                    }),
                    data = newColumn;
                $.ajax({
                    url: url,
                    type: 'PUT',
                    contentType: "application/json; charset=utf-8",
                    dataType: "json",
                    data: JSON.stringify(data)
                }).done(function (data, status, request) {
                    var columnDef = Confluence.PlaySQL.Spreadsheet.readColumnDefinition(column.index, data),
                        oldName = column.name;
                    column.name = columnDef.name;
                    column.label = columnDef.label;
                    column.type = columnDef.type;
                    column.length = columnDef.length;
                    column.formula = columnDef.formula;
                    column.renderer = columnDef.renderer;
                    column.viewRenderer = columnDef.viewRenderer;
                    column.formats = columnDef.formats;
                    column.width = columnDef.width;
                    innerSuccessCallback(oldName);
                }).fail(function (jqXHR, error, reason) {
                    try {
                        // If the BAD_REQUEST is specially formatted by Play SQL
                        var data = $.parseJSON(jqXHR.responseText);
                        failureCallback && failureCallback(data.message);
                        return;
                    } catch (err) {
                        // Swallow and proceed to the next section
                        AJS.log(err);
                    }
                    failureCallback && failureCallback(jqXHR.responseText);
                });
            }
        }
        /** Do create the column in the back-end
         * At the time of calling, the column shouldn't be among the spreadsheet.columns.
         */
        var createColumn = function($cell, column) {
            if (!spreadsheet.perms.hasPerm('ADD_COLS')) return;
            var url = baseUrl("/cols", checkNotNull(column.name)),
                data = {
                    name: checkNotNull(column.name),
                    type: checkNotNull(column.type),
                    label: checkNotNull(column.label),
                    renderer: checkNotNull(column.renderer),
                    viewRenderer: column.viewRenderer,
                    options: undefined
                };
            $.ajax({
                url: url,
                type: 'PUT',
                contentType: "application/json; charset=utf-8",
                dataType: "json",
                data: JSON.stringify(data)
            }).done(function(data, status, request){
                var cell = getCell($cell),
                    col = colOf($cell);
                col.name = column.name;
                col.type = column.type;
                col.label = column.label;
                col.renderer = column.renderer;
                col.viewRenderer = column.viewRenderer;
                cell.renderer.view();
            }).fail(function(jqXHR, error, reason){
                Confluence.PlaySQL.Messages.addMessage("error", "The column couldn't be created: " + column.label,
                    "change-col", jqXHR.responseText);
                $cell.removeClass("saving").addClass("invalid").attr("title", "The column couldn't be created");
            });
        }

        var emptyRow = function(length) {
            var result = [];
            for (var i = 0 ; i < length; i++) {
                result[i] = ["", "", ""]; // The third value is usually provided by renderCells()
            }
            return result;
        }

        function incrementColSpan($tds) {
            if ($tds == null || $tds.size() < 1) {
                return;
            }
            $tds.each(function(i, $td) {
                $td = $($td);
                var span = parseInt($td.attr('colspan')) || 1;
                $td.attr('colspan', span + 1);
            });
        }

        /** Manages adding a column in the HTML */
        var addColumnUI = function(column) {
            Confluence.PlaySQL.Selection.forTable(column.$placeholder).unselectAll();

            // Enlarge TBODY
            column.$placeholder.find("tbody tr td:last-child").each(function() {
                var $this = $(this);
                if ($this.parent().is(".playsql-message")) {
                    incrementColSpan($this);
                } else {
                    var $td = $("<td></td>", {
                        "data-value": "",
                        "data-rendered": ""
                    }).insertBefore($this);
                }
            });

            // Enlarge THEAD

            var $th = $("<th></th>", {
                "data-name": column.name,
                "data-type": column.type,
                "data-renderer": column.renderer,
                "data-view-renderer": column.viewRenderer,
                "data-status-pk": column.statusPK ? "true" : undefined,
                "data-status-position": column.statusPosition ? "true" : undefined,
                "text": column.label
            }).insertBefore(column.$placeholder.find("thead tr.playsql-col-headers th:last-child"));

            incrementColSpan(column.$placeholder.find("tr.pagination-row td:not(.right-padding)"));
            incrementColSpan(column.$placeholder.find("tr.playsql-message td:not(.right-padding)"));

            // Enlarge TBOTTOM
            incrementColSpan(column.$placeholder.find("tfoot tr td:not(:last-child)").last());

            return $th;
        }

        // Add a column on the front-end
        var addColumn = function($this, e) {
            if (!spreadsheet.perms.hasPerm('ADD_COLS')) return;

            var column = {
                name: "",
                label: "New column",
                type: "TEXT",
                renderer: "link",
                viewRenderer: undefined,
                formats: {},
                width: null,
                $placeholder: $placeholder,
                index: spreadsheet.columns.length
            };

            var $th = addColumnUI(column);

            spreadsheet.columns.push(column);

            // Move the header to 'edit'
            var cell = getCell($th).renderer;
            cell.edit();
            cell.focus();

            saveTotals();

            e.preventDefault();
            return false;
        }

        var deleteColumn = function($cell) {
            if (!spreadsheet.perms.hasPerm('REMOVE_COLS')) return;
            Confluence.PlaySQL.Selection.forTable($placeholder).unselectAll();
            var column = colOf($cell);

            // The column name could be empty if the user was still creating the column
            // and pressed Esc. Otherwise we ask him to confirm.
            if (column.name != "" && !confirm(("You are about to delete the column '$1' (label: $2). The operation cannot be cancelled " +
                "and the contents of the cells will be lost. Are you sure?")
                    .replace("$1", column.name).replace("$2", column.label))) {
                return false;
            }

            spreadsheet.columns.splice(column.index, 1);
            // Reindex the columns
            for (var index = 0 ; index < spreadsheet.columns.length ; index++) {
                spreadsheet.columns[index].index = index;
            }

            checkNotNull($placeholder.find("thead > tr.playsql-col-headers > *").eq(column.index)).remove();
            $placeholder.find("tbody td:nth-child(" + (column.index+1) + ")").remove();
            removeTotalAtColumn(spreadsheet.totals, column.index - 1);

            // We don't need to remove it server-side if we were just editing the column
            if (column.name) {
                var url = baseUrl("/cols", checkNotNull(column.name)),
                    data = {
                        name: checkNotNull(column.name),
                        type: checkNotNull(column.type),
                        label: checkNotNull(column.label),
                        options: undefined
                    };
                $.ajax({
                    url: url + "?" + $.param({ confirm: 'DELETE' }),
                    type: 'DELETE',
                    contentType: "application/json; charset=utf-8"
                }).done(function (data, status, request) {
                    // Do nothing, show nothing.
                    saveTotals();
                }).fail(function (jqXHR, error, reason) {
                    Confluence.PlaySQL.Messages.addMessage("error", "The column couldn't be deleted: " + column.label,
                        "change-col", jqXHR.responseText);
                });
            } else {
                saveTotals();
            }

            return false;
        }

        /**
         * Remove the total for the right column
         * @param totals the totals array (the model)
         * @param columnIndex the column index, starting at 0 = the first column / the 0 index in totals (Column ID = -1)
         */
        var removeTotalAtColumn = function(totals, columnIndex) {
            if (!totals || totals.length == 0) return;
            var position = 0;
            $placeholder.find("tfoot > tr > *:not(:first-child):not(.right-padding)").each(function(index){
                var $th = $(this),
                    span = parseInt($th.attr("colspan")) || 1,
                    totalDef = totals[index];
                if (columnIndex >= position && columnIndex < position + span) {
                    // That's where we remove something
                    if (span >= 3) {
                        $th.attr("colspan", --span);
                        totalDef.colspan = span;
                        return false;
                    } else if (span == 2) {
                        $th.removeAttr("colspan");
                        totalDef.colspan = 1;
                        return false;
                    } else {
                        $th.remove();
                        totals.splice(index, 1);
                        return false;
                    }
                } else {
                    position += span;
                }
            });
        }

        /*var startRenameColumn = function($header) {
         var renderer = getCell($header).renderer;
         renderer.edit();
         renderer.focus();
         }*/

        var renderAllColumn = function(column) {
            var renderer = Confluence.PlaySQL.Spreadsheet.renderers[column.viewRenderer || column.renderer];
            $placeholder.find("tbody > tr > td:nth-child(" + (1+column.index) + ")").each(function(){
                var $cell = $(this),
                    value = $cell.attr('data-value');
                if (!$cell.hasClass("editing saving")) {
                    $cell.html(renderer.renderCellContents(value, spreadsheet, column));
                }
            });
        }

        var sortCol = function($header, way, nulls) {
            // Parse spreadsheet's order by clause
            var clause = Confluence.PlaySQL.Helpers.parseOrderByClause(spreadsheet.orderbyClause),
                col = checkNotNull(colOf($header)),
                colname = '"' + col.name + '"',
                perform = true;
            way = way || 'REVERSE';
            for (var i = 0 ; i < clause.length ; i++) {
                if (clause[i].field.indexOf(colname) == 0) {
                    if (i == 0) {
                        if (way == clause[i].way && nulls == clause[i].nulls) return;
                        if (way == 'REVERSE') {
                            way = clause[i].way.toUpperCase() == 'ASC' ? 'DESC' : 'ASC';
                        } else if (way == 'NO-SORTING') {
                            clause.splice(i, 1);
                            perform = false;
                            break;
                        }
                        clause[i].way = way;
                        clause[i].nulls = nulls;
                        perform = false;
                        break;
                    } else {
                        // Delete and make it the primary sorting
                        clause.splice(i, 1);
                        break;
                    }
                }
            }
            if (perform && way != 'NO-SORTING') {
                if (way == 'REVERSE') way = 'ASC';
                var cast = col.type == 'VARCHAR' || col.type == 'TEXT' ? '::text'
                    : col.type == 'INTEGER' || col.type == 'BIG_INTEGER' || col.type == 'BOOLEAN' || col.type == 'DATE' || col.type == 'TIME' || col.type == 'NUMERIC' || col.type == 'DOUBLE' ? '::' + col.type
                    : '';
                clause.unshift({
                    field: colname,// + cast,
                    way: way,
                    nulls: nulls
                });
            }
            var clauseString = Confluence.PlaySQL.Helpers.buildOrderByClause(clause);
            spreadsheet.orderbyClause = clauseString;
            AJS.log("ORDER BY " + clauseString);
            spreadsheet.load();
        }

        var pasteHandler = function($contenteditable) {
            // Make a big array of the values: { rows -> { cells -> [ value, formula ] } }
            var allValues = [],
                $placeholder = this.$placeholder,
                $focused = $placeholder.find(".focused"),
                $selected = $placeholder.find(".selected"),
                isPlaySQLPasteSource = $contenteditable.find("table > thead").length != 0,
                columnHeaders = [];
            if (isPlaySQLPasteSource) {
                columnHeaders = $contenteditable.find("table > thead th").map(function(i, el){return $(el).text();});
                // Remove the columns which contain a th, i.e. the ID column
                $contenteditable.find("table > tbody > tr:first > th").each(function(index, th){
                    columnHeaders.splice($(th).index(), 1);
                });
                // Chop the head off, keep the body, Vive la Revolution!
                $contenteditable.find("table > thead, table.playsql > tbody > tr > th").remove();
            }
            var $row, $cell, cell;
            $contenteditable.find("table tr").each(function(){
                var $tr = $(this),
                    $cells = $tr.find("th,td"),
                    row = [];
                $cells.each(function(){
                    var $cell = $(this),
                        value = $cell.attr("data-value"),
                        formula = $cell.attr("data-cell-metadata-formula");
                    if (typeof value == 'undefined') {
                        value = $cell.text();
                    }
                    row.push([value, formula]);
                });
                allValues.push(row);
            });
            if (allValues.length == 0) {
                // Did the person start typing?
                if ($.trim($contenteditable.html()).length > 0 && $contenteditable.html() != "&#x200B;" && $contenteditable.html() != "\u200B") {
                    var walker = document.createTreeWalker($contenteditable[0], NodeFilter.SHOW_ALL, null, false),
                        text = "",
                        blockNodes = ['DIV', 'BR', 'P', 'LI', 'DL', 'DD', 'DT', 'FORM', 'PRE', 'TR'];
                    while (node = walker.nextNode()) {
                        if (blockNodes.indexOf(node.nodeName) != -1) {
                            text += "\n";
                        } else if (node.nodeType == Node.TEXT_NODE) {
                            text += node.nodeValue;
                        }
                    }
                    var rows = $.trim(text).split(/\r?\n\r?/);
                    for (var i in rows) {
                        // The map() function returns the [value, formula] pair
                        allValues.push(rows[i].split(/\t/)
                            .map(function(e){return [e, undefined]}));
                    }

                    if (catchTextShortcuts(allValues)) {
                        $contenteditable.empty();
                        return;
                    }
                } else {
                    return;
                }
            }
            // Don't let foreign content in the table...
            $contenteditable.empty();
            if ($focused.is("tbody th")) {
                $focused = $focused.next(":not(.right-padding)");
            } else if ($focused.is("thead th")) {
                alert("You are attempting to paste into a full column. The contents will only be pasted into the visible rows.");
            }
            if ($focused.length == 0)
                return;
            var originCol = $focused.index(),
                originRow = $focused.closest("tr").index(),
                maxRows = $placeholder.find("> table > tbody tr").length - originRow,
                missingRows = Math.max(0, allValues.length - maxRows),
                spreadsheetWidth = spreadsheet.columns.length,
                maxCols = spreadsheetWidth - originCol,
                missingCols = allValues[0].length - maxCols,
                pasteAtOrigin = $selected.size() == 1,
                pasteWithRepeat = allValues.length == 1,
                pasteWithinSelection = $selected.size() > 1;
            spreadsheet.disableFocusManagement = true;

            if (missingRows > 0) {
                this.addRows(missingRows);
            }
            var column;
            if (missingCols > 0) {
                var createMoreCols = spreadsheet.perms.hasPerm('ADD_COLS') &&
                    confirm("The data you're attempting to paste is wider than the number of columns. Do you want Play SQL to kindly create columns for you?");
                if (createMoreCols) {
                    var columnsToCreate = [],
                        columnNamesToCreate = [];
                    for (var i = spreadsheetWidth; i < spreadsheetWidth + missingCols; i++) {
                        column = {
                            name: spreadsheet.defaultCase == 'LOWER' ? "newcolumn" : "NEWCOLUMN",
                            label: "New column",
                            type: "TEXT",
                            length: undefined,
                            renderer: "link",
                            viewRenderer: undefined,
                            formats: {},
                            width: undefined,
                            $placeholder: $placeholder,
                            index: spreadsheet.columns.length
                        };
                        if (columnHeaders[i - originCol]) {
                            column.label = columnHeaders[i - originCol];
                            column.name = Confluence.PlaySQL.Helpers.generateKey(column.label, spreadsheet.defaultCase);
                        }
                        column.name = getUniqueColumnName(spreadsheet.columns, column.name);
                        //column.name = getUniqueColumnName(columnsToCreate, column.name);

                        addColumnUI(column);
                        spreadsheet.columns.push(column);
                        columnsToCreate.push({
                            name: checkNotNull(column.name),
                            type: checkNotNull(column.type),
                            label: checkNotNull(column.label),
                            renderer: checkNotNull(column.renderer),
                            viewRenderer: undefined,
                            options: undefined
                        });
                        columnNamesToCreate.push(column.name);
                        maxCols++;
                    }

                    var url = baseUrl("/cols");
                    $.ajax({
                        url: url,
                        type: 'POST',
                        contentType: "application/json; charset=utf-8",
                        dataType: "json",
                        data: JSON.stringify(columnsToCreate)
                    }).done(function(data, status, request){
                        var columns = data.columns;
                        for (var i in columns) {
                            var column = columns[i];
                            if (columnNamesToCreate.indexOf(column.name) == -1) {
                                continue;
                            }
                            var col = colOf(column.name),
                                cell = getCell($cell);
                            col.name = column.name;
                            col.type = column.type;
                            col.label = column.label;
                            col.renderer = column.renderer;
                            col.viewRenderer = column.viewRenderer;
                            cell.renderer.view();
                        }
                    }).fail(function(jqXHR, error, reason){
                        Confluence.PlaySQL.Messages.addMessage("error", "The column couldn't be created: " + column.label,
                            "change-col", jqXHR.responseText);
                        $cell.addClass("invalid").attr("title", "The column couldn't be created");
                    });
                }
            }

            function paste(cell, value) {
                var formula = value[1],
                    value = value[0];
                if (cell) {
                    if (formula) {
                        spreadsheet.saveCellMetadata($cell, "formula", formula);
                    } else {
                        cell.renderer.edit(value);
                    }
                    cell.$cell.addClass("selected");
                }
            }
            // Get going, paste values in cells
            if (pasteWithRepeat) {
                $selected.closest("tr").each(function(){
                    var $row = $(this);
                    for (var col in allValues[0]) {
                        if (parseInt(col) >= maxCols) break;
                        $cell = $row.find("th,td").eq(parseInt(col) + originCol);
                        if (!$cell.hasClass("selected")) continue;
                        cell = getCell($cell);
                        paste(cell, allValues[0][col]);
                    }
                });
            } else if (pasteAtOrigin || pasteWithinSelection) {
                for (var row in allValues) {
                    $row = $placeholder.find("> table.playsql > tbody tr").eq(parseInt(row) + originRow);
                    for (var col in allValues[row]) {
                        if (parseInt(col) >= maxCols) break;
                        $cell = $row.find("th,td").eq(parseInt(col) + originCol);
                        if (pasteWithinSelection && !$cell.hasClass("selected")) continue;
                        cell = getCell($cell);
                        paste(cell, allValues[row][col]);
                    }
                }
            }

            Confluence.PlaySQL.Selection.forTable($placeholder).unselectAll();
            spreadsheet.disableFocusManagement = false;
            cell = cell || getCell($focused)
            if (cell) {
                cell.$cell.addClass("selected focused");
                cell.renderer.focus(true);
            }
            spreadsheet.sprayRequestPipe();
        }

        function catchTextShortcuts(allValues) {
            // It used to be that we'd catch '=' and display the column's formula editor. Not anymore:
            // now we have cell-local formulas and they're managed seamlessly by the cell renderer.
            // var text = allValues.toString()
            /*if (text == "=") {
                var $cell = $placeholder.find(".focused").first(),
                    cell = getCell($cell);
                if ($cell.index() > 0 && cell != undefined) {
                    cell.renderer.editFormula();
                    return true;
                }
            }*/
            return false;
        }

        /** Fetches 1 row of the spreadsheet, to check whether there's an error
         * @param colname the name of the column for which we want to return the sample value
         */
        var testSpreadsheet = function(colname, callback) {
            var url = baseUrl("/data", {
                limit: 1
            });

            $.ajax({
                url: url,
                type: 'GET',
                contentType: "application/json; charset=utf-8",
                timeout: 180000,
                cache: false
            }).done(function(data, status, request){
                var colNames = request.getResponseHeader("X-COL-NAMES").split(","),
                    tableDef = data.tableDef,
                    colIndex, message;
                for (var i in tableDef.columns)
                    if (tableDef.columns[i].name == colname) {
                        colIndex = i;
                        break;
                    }
                if (colIndex == undefined) {
                    return callback(false, "Column " + colname + " doesn't exist");
                }
                var column = readColumnDefinition(colIndex, tableDef.columns[i], undefined),
                    value = data.data.length > 0 ? data.data[0][colIndex][0] : undefined;

                message = data.errorMessage;
                if (data.errorMessage) {
                    message = message.replace("Error with a calculated column:", "");
                    message = message.replace("ERROR: ", "");
                    message = message.replace(/: syntax error at or near ""[a-zA-Z0-9]+""/, "");
                    message = message.replace(/ Position: [\d]+/, "");
                    return callback(false, message);
                } else {
                    return callback(true, value);
                }
            }).fail(function(jqXHR, error, reason) {
                var message = "The test of the spreadsheet was unsuccessful: " + jqXHR.responseText;
                callback(false, message);
            });
        }

        function displayCellProperties(e) {
            Confluence.PlaySQL.Spreadsheet.CellProperties.display(spreadsheet, $(".selected"));
        }

        function removeCellContents(e) {
            var $cells = $placeholder.find("td.selected:not(.right-padding)");
            if ($cells.size() == 0 || $cells.filter(".editing").size() > 0) return;
            // We don't want to let the browser go to the previous page (because of the Backspace key) if a
            // cell is selected
            e.preventDefault();

            if (spreadsheet.perms.hasPerm("EDIT_CELLS") || spreadsheet.perms.hasPerm("Q_VIEW_EDIT")) {
                if ($cells.size() > 1) {
                    if (!confirm("Do you want to delete the contents of " + $cells.size() + " cells?")) return;
                }
                $cells.each(function () {
                    var $cell = $(this);
                    if ($cell.hasClass("joint-id-renderer")) {
                        if ($cell.find(".joint-id-remove").size() > 0) {
                            if (!confirm("This cell is a foreign key to a joint table. Do you want to delete the full row of the joint table?")) return;
                            spreadsheet.removeJointRow($cell);
                        }
                    } else if (spreadsheet.perms.hasPerm("EDIT_CELLS") || spreadsheet.perms.hasPerm('Q_VIEW_EDIT', spreadsheet.colOf($cell))) {
                        spreadsheet.saveCell($cell, null);
                        if (spreadsheet.perms.hasPerm('EDIT_ROWS_METADATA')) {
                            spreadsheet.saveCellMetadata($cell, "formula", null);
                        }
                    }
                });
                spreadsheet.sprayRequestPipe();
            }
        }

        function getUniqueColumnName(columns, name) {
            for (var i = 0 ; i < columns.length ; i++) {
                var suggestedName = name + (i != 0 ? i : ""),
                    found = false;
                for (var j in columns) {
                    if (columns[j].name == suggestedName) {
                        found = true;
                        break;
                    }
                }
                if (!found) return suggestedName;
            }
            return "";
        }

        function buildCssClassesForCell(value, cellMetadata) {
            if (cellMetadata && cellMetadata.formula) {
                return "formula";
            }
            if (value != undefined && value.charAt(0) == "=") {
                return "anti-formula";
            }
            return undefined;
        }
        function buildCssStylesForCell(value, cellMetadata) {
            if (cellMetadata && cellMetadata["background-color"]) {
                return "background-color: " + cellMetadata["background-color"] + ";";
            }
            return undefined;
        }

        /** Reset the properties of the spreadsheet, as if never been touched by a user on this page */
        function reset() {
            // Reset the Order By and the Filters
            spreadsheet.orderbyClause = $placeholder.attr('data-macro-orderby-clause');
            spreadsheet.whereClauseColumnFilters = undefined;
            // Reset the User Parameters
            spreadsheet.userParameters = undefined;
        }

        // Append all methods to 'this' and return it
        var result = $.extend(spreadsheet, {
            load: load,
            reset: reset,
            // Finding elements
            colOf: colOf,
            getRecordIds: getRecordIds,
            getCell: getCell,
            getCellMetadata : getCellMetadata,
            getRowMetadata : getRowMetadata,
            findRowByPosition: findRowByPosition,
            findRowByPk: findRowByPk,
            $findColumnCells : $findColumnCells,
            // Actions on elements
            addRow: addRow,
            addRows: addRows,
            deleteRow: deleteRow,
            deleteRows: deleteRows,
            addJointRow: addJointRow,
            removeJointRow: removeJointRow,
            sprayRequestPipe: sprayRequestPipe,
            saveCell: saveCell,
            saveCellMetadata: saveCellMetadata,
            recalculateFormulas: recalculateFormulas,
            dispatchCellValues: dispatchCellValues,
            //renameColumn: renameColumn,
            saveColumn: saveColumn,
            createColumn: createColumn,
            deleteColumn: deleteColumn,
            renderAllColumn : renderAllColumn,
            sortCol: sortCol,
            pasteHandler: pasteHandler,
            // Helpers
            emptyRow: emptyRow,
            baseUrl: baseUrl
        });
        Confluence.PlaySQL.Spreadsheet.spreadsheets[spreadsheet.id] = result;
        Confluence.PlaySQL.spreadsheet = result;
        return result;
    } // End of byPlaceholder()

    function readColumnDefinition(index, colDefinition, $placeholder) {
        return {
            name: colDefinition.name,
            label: colDefinition.label,
            type: colDefinition.type,
            length: colDefinition.length,
            formula: colDefinition.formula,
            renderer: checkNotNull(colDefinition.renderer, colDefinition),
            viewRenderer: colDefinition.viewRenderer,
            statusPK: colDefinition.statusPK,
            statusPosition: colDefinition.statusPosition,
            formats: colDefinition.formats || {},
            width: colDefinition.width,
            index: index,
            $placeholder: $placeholder
        };
    }

    function registerRenderer(renderer) {
        Confluence.PlaySQL.Spreadsheet.renderers[renderer.key] = renderer;
    }

    function initializeAll() {
        var anchor = window.location.hash,
            recordId;
        if (anchor && anchor.indexOf("#")==0) {
            recordId = parseInt(anchor.substring(1));
            if (isNaN(recordId)) {
                recordId = undefined;
            }
        }

        $(".playsql-spreadsheet-placeholder").each(function(){
            try {
                var spreadsheet = Confluence.PlaySQL.Spreadsheet.byPlaceholder($(this));
                spreadsheet.selectedRecord = recordId;
                spreadsheet.load();
            } catch (err) {
                AJS.log("Error while loading spreadsheet: ", err);
            }
        });
    }

    Confluence.PlaySQL = Confluence.PlaySQL || {};
    window.PlaySQL = Confluence.PlaySQL;
    Confluence.PlaySQL.Spreadsheet = Confluence.PlaySQL.Spreadsheet || {
        byId: byId,
        byPlaceholder: byPlaceholder,
        initializeAll : initializeAll,
        spreadsheets: {},
        renderers: {},
        readColumnDefinition: readColumnDefinition
    };

    $(function () {
        window.setTimeout(function() {
            AJS.$("#license-banner").remove();
        }, 1000);

        if (!Confluence.PlaySQL.Spreadsheet.initialized) {
            AJS.trigger("playsql-spreadsheet-init", { register: registerRenderer });
            AJS.unbind("playsql-spreadsheet-init");
            Confluence.PlaySQL.Spreadsheet.initialized = true;
            Confluence.PlaySQL.Selection.setupCopyHandler({
                pasteHandler: function($contenteditableDiv) {
                    // 'this' refers to the ss
                    return this.pasteHandler($contenteditableDiv);
                }
            });
        }

        // Load spreadsheets by default, except if we're in the Query Editor
        if ($("body.query-editor-mode").size() == 0) {
            Confluence.PlaySQL.Spreadsheet.initializeAll();
        }
    });

})(AJS.$);