/**
 * This file overrides the spreadsheetId field for the Macro Browser
 */
AJS.bind("init.rte", function() {
    var $ = AJS.$;
    var readonly = false;
    function fillValues() {
        var spaceKey = AJS.Meta.get("space-key"),
            url = AJS.contextPath() + "/rest/playsql-spreadsheets/1.0/space:" + spaceKey;
        $.ajax({
            url: url,
            type: 'GET',
            dataType: "json"
        }).done(function(data, status, jqXHR){
            var $select = $("#macro-param-spreadsheetId"),
                originalValue = $select.val(),
                mode = jqXHR.getResponseHeader("X-MODE");
            $select.empty();
            addValue($select, ["undefined", "Select a value"]);
            for (var i in data) {
                addValue($select, data[i]);
            }
            setSpreadsheetId($select, originalValue);

            readonly = (mode != "READ_WRITE");
            updateUserParameters();
        }).fail(function(data){
            AJS.logError("Can't fetch the list of queries for space:" + spaceKey)
        });
    }
    function addValue($select, data) {
        var type = data[2],
            $target = $select;
        // types 'undefined' and 'SPREADSHEET' don't use an optgroup
        if (!!type) {
            $target = $select.find("optgroup[data-key=" + type + "]");
            if ($target.size() == 0) {
                var label = type == "JOINT_TABLE" ? "Joint Tables"
                        : type == "SPREADSHEET" ? "Spreadsheets"
                        : type;
                $select.append($("<optgroup>", {
                    "data-key": type,
                    "label": label
                }));
                $target = $select.find("optgroup[data-key=" + type + "]");
            }
        }

        $target.append(AJS.$("<option/>", {
            value: data[0],
            text: data[1]
        }));
    }
    function setSpreadsheetId($select, id) {
        $select.val(id);
        if (id != undefined && $select.val() !== id) {
            addValue($select, [id, "Spreadsheet"]);
            $select.val(id);
        }
    }
    function updateUserParameters() {
        var spreadsheetIdField = AJS.MacroBrowser.fields["spreadsheetId"],
            spreadsheetId = spreadsheetIdField.getValue(),
            oldSpreadsheetId = spreadsheetIdField.loadedSpreadsheetId,
            spaceKey = AJS.Meta.get("space-key");
        if (spreadsheetId === oldSpreadsheetId) {
            return;
        }
        spreadsheetIdField.loadedSpreadsheetId = spreadsheetId;
        // In IE8, make sure the MacroBrowser knows required fields have values
        AJS.MacroBrowser.paramChanged();

        /*for (var fieldName in AJS.MacroBrowser.fields) {
         if (fieldName.indexOf("user-parameter-") == 0) {
         AJS.MacroBrowser.fields[fieldName].remove();
         }
         }*/

        if (readonly) {
            var $replaceAllParams = $("#macro-param-div-spreadsheetId");
            $replaceAllParams.empty().html("This space is read-only. You cannot have Spreadsheets in this space.");
            $replaceAllParams.parent().find(".macro-param-div:not(#macro-param-div-spreadsheetId)").remove();
        } else if (spreadsheetId == undefined || spreadsheetId == "undefined") {
            var $description = $("#macro-param-div-spreadsheetId .macro-param-desc"),
                $createLink = $("<a></a>", {
                    href : AJS.contextPath() + "/playsql-base-plugin/index.action?" + $.param({key: spaceKey}) + "#create-spreadsheet",
                    text : "or create a new spreadsheet...",
                    target: "_blank"
                });
            if ($description.size() == 0) {
                $description = $("#macro-param-div-spreadsheetId")
                    .append("<div class='macro-param-desc'></div>")
                    .find("div");
            }
            $description.empty().html($createLink);
        } else {
            var $description = $("#macro-param-div-spreadsheetId .macro-param-desc"),
                $macroFieldColumns = $("#macro-param-div-display-columns"),
                $editLink = $("<a></a>", {
                    href : AJS.contextPath() + "/playsql-base-plugin/spreadsheet-editor.action?key=" + spaceKey + "&spreadsheetId=" + spreadsheetId,
                    text : "Edit the spreadsheet...",
                    target : "_blank"
                });
            if ($description.size() == 0) {
                $description = $("#macro-param-div-spreadsheetId")
                    .append("<div class='macro-param-desc'></div>")
                    .find("div");
            }
            $description.unbind().empty().html($editLink);
            $macroFieldColumns.removeClass("managed").find(".managed-col").remove();

            var url = AJS.contextPath() + "/rest/playsql-spreadsheets/1.0/space:" + AJS.Meta.get("space-key") + "/" + spreadsheetId
            $.ajax({
                url: url,
                type: 'GET',
                dataType: "json"
            }).done(function(data){
                if (!data || !data.tableDef)
                {
                    return;
                }
                var spreadsheetName = data.tableDef.label || data.tableDef.name,
                    $macroFieldArea = $("#macro-insert-container .macro-input-fields");
                // Update the spreadsheet name, if required
                $macroFieldArea.find("#macro-param-spreadsheetId option[value='" + spreadsheetId + "']").text(spreadsheetName);

                // If known, list fields as tickboxes and let the user choose some more fields
                $macroFieldColumns.addClass("managed");

                function addColumn(name, label, custom, ticked, hidden) {
                    var $colDesc = $("<div></div>", {
                            "class": "checkbox",
                            html: "<input type='checkbox' class='managed-col checkbox'>"
                            + "<label class='managed-col'></label>"
                        }),
                        $input = $colDesc.find("input");
                    $label = $colDesc.find("label");

                    $input.attr("id", "display-" + name);
                    $label.attr("for", "display-" + name).text(label);
                    if (custom) {
                        $input.addClass("custom");
                        $label.addClass("custom");
                    }
                    if (ticked) {
                        $input.attr("checked", "on");
                    }
                    if (hidden) {
                        $input.addClass("hidden");
                        $label.addClass("hidden");
                    }
                    $macroFieldColumns.append($colDesc);
                }

                for (var i in data.tableDef.columns) {
                    var col = data.tableDef.columns[i],
                        hidden = col.name == "ID" || col.name == "POSITION";
                    addColumn(col.name, col.label, false, true, hidden);
                }
                // We'll deactivate that for now
                // addColumn("add-field", "Add a calculated column...", true, false, false);

                if (!$macroFieldColumns.attr("data-managed-events")) {
                    $macroFieldColumns.attr("data-managed-events", "true");
                    function editCol(colId) {
                        var $label = $("label[for='" + colId + "']");
                        var $newInput = $("<input>", {
                            id: "for-" + colId,
                            "class": "custom-col-label",
                            value: $label.text()
                        });
                        $label.replaceWith($newInput);
                        $newInput.focus().select();
                    }
                    $macroFieldColumns.on("change", "input.custom", function(){
                        var $checkbox = $(this),
                            $label = $checkbox.next(),
                            id = $checkbox.attr("id");
                        if (id == "display-add-field") {
                            // Create an ID and assign it
                            id = "display-" + $macroFieldColumns.find("input").size();
                            $checkbox.attr("id", id);
                            $label.attr("for", id).text("Write SQL formula");
                            addColumn("add-field", "Add a calculated column...", true, false);
                            editCol(id);
                        }
                    }).on("dblclick", "label.custom", function(){
                        var id = $(this).attr("for");
                        editCol(id);
                    }).on("blur", "input.custom-col-label", function(){
                        var id = $(this).attr("id").substr(4);
                        $(this).replaceWith($("<label>", {
                            "for": id,
                            text: $(this).val(),
                            "class": "managed-col custom"
                        }));
                    }).on("change", "input", function() {
                        var $checkboxes = $("#macro-param-div-display-columns input.checkbox"),
                            ticked = [],
                            unticked = [],
                            fields;
                        $checkboxes.each(function () {
                            var $cb = $(this),
                                isTicked = $cb.is(":checked"),
                                id = $cb.attr("id").substr(8);
                            if (isTicked) {
                                ticked.push(id);
                            } else {
                                unticked.push(id);
                            }
                        });
                        if (unticked.length == 0) {
                            fields = "*";
                        } else {
                            fields = "";
                            $.each(ticked, function(i, el) {
                                if (i > 0) fields += ", ";
                                fields += el;
                            });
                        }
                        AJS.MacroBrowser.fields["display-columns"].setValue(fields);
                    });
                }
            }).fail(function(data){
                AJS.logError("Can't fetch the spreadsheet for space:" + spaceKey)
            });
        }
        $("#macro-browser-preview-link").click();
    }

    var jsOverrides = {
        "fields" : {
            "string" : function(params, options){
                options = options || {};
                if (params && params.name == "spreadsheetId") {
                    var $paramDiv = $(Confluence.Templates.MacroBrowser.macroParameterSelect());
                    var $select = $("select", $paramDiv);

                    options.setValue = function(value) {
                        setSpreadsheetId($("#macro-param-spreadsheetId"), value);
                    }

                    fillValues();

                    options.onchange = updateUserParameters;

                    var field = AJS.MacroBrowser.Field($paramDiv, $select, options);
                    field.updateSpreadsheet = updateUserParameters;
                    return field;
                } else if (params && params.name == "display-columns") {
                    var $paramDiv = $(Confluence.Templates.MacroBrowser.macroParameter());
                    var $input = $("input", $paramDiv);

                    options.setValue = function(value) {
                        if ($input.data("working")) return;
                        $input.val(value);
                        $input.data("working", true);
                        var fields = value.split(/[", ]+/);
                        for (var i in fields) {
                            if ($.trim(fields[i]).length > 0) {
                                var $el = $paramDiv.find("input.checkbox[id='display-" + $.trim(fields[i]) + "']");
                                if ($el.size() > 0) {
                                    if (!$el.is(":checked")) {
                                        $el.attr("checked", "on");
                                    }
                                }
                            }
                        }
                        $input.data("working", false);
                    }

                    return AJS.MacroBrowser.Field($paramDiv, $input, options);
                } else {
                    return AJS.MacroBrowser.ParameterFields["string"](params, options);
                }
            }
        },
        "beforeParamsSet": function(paramMap, inserting) {
            if (!paramMap["context"]) {
                paramMap["context"] = "space:" + AJS.Meta.get("space-key");
            }
            return paramMap;
        },
        "beforeParamsRetrieved" : function(paramMap, macro, sharedParamMap) {
            /*for (var fieldName in paramMap)
             {
             if (fieldName.indexOf('user-parameter') == 0) {
             delete paramMap[fieldName];
             }
             }
             for (var fieldName in AJS.MacroBrowser.fields) {
             if (fieldName.indexOf("user-parameter-") == 0) {
             var userField = AJS.MacroBrowser.fields[fieldName],
             val = userField.getValue(),
             name = userField.userParameterName;
             paramMap[name] = val;
             }
             }*/
            return paramMap;
        }
    };
    AJS.MacroBrowser.setMacroJsOverride("playsql-spreadsheet", jsOverrides);

    AJS.Confluence.PropertyPanel.Macro.registerButtonHandler("playsql-edit-spreadsheet", function(e, macroNode) {
        var parameters = $(macroNode).attr("data-macro-parameters");
        if (typeof parameters != "string") return;
        parameters = parameters.split("|");
        var spreadsheetId,
            spaceKey = AJS.Meta.get("space-key"),
            context = "space:" + spaceKey;
        for (var i in parameters) {
            if (parameters[i].indexOf("spreadsheetId=") == 0) {
                spreadsheetId = parameters[i].substr("spreadsheetId=".length);
            } else if (parameters[i].indexOf("context") == 0) {
                context = parameters[i].substr("context=".length);
                if (context.indexOf("space:") == 0) {
                    spaceKey = context.substr("space:".length);
                }
            }
        }
        if (spreadsheetId != undefined) {
            window.open(AJS.contextPath() + "/playsql-base-plugin/spreadsheet-editor.action?" + $.param({
                key : spaceKey,
                spreadsheetId : spreadsheetId
            }));
        }
    });
});


AJS.$(function ($) {
    var $doc = $(document).bind('postPaste', function (e, pl, o) {
        // Find out what is being pasted
        var $pasted = $(o.node),
            $table = $pasted.find("table.playsql");
        if ($table.size() == 0) {
            // That's not a paste of a spreadsheet
            return;
        }
        var context = $table.attr("data-context") || $table.attr("data-context"),
            spreadsheetId = $table.attr("data-macro-spreadsheet-id"),
            queryId = $table.attr("data-macro-query-id"),
            theoreticalDisplayColumns = $table.attr("data-macro-display-columns"),
            columns = $table.find("thead tr.playsql-col-headers th:not(:first-child)").map(function(){return $(this).attr("data-name")}),
            displayColumns = columns.length > 0 ? columns.toArray().join(", ") : theoreticalDisplayColumns,
            theoreticalWhereClause = $table.attr("data-macro-where-clause"),
            rowIds = $table.find("tbody tr").map(function(){return $(this).attr("data-row-id");}),
            whereClause = rowIds.length > 0 ? '"ID" in (' + rowIds.toArray().join(', ') + ")" : theoreticalWhereClause;
        if (spreadsheetId == undefined) {
            if (queryId != undefined) {
                spreadsheetId = "QUERY-" + queryId;
            }
        }

        var macroRenderRequest = {
            contentId: AJS.Meta.get('content-id') || "0",
            macro: {
                name: "playsql-spreadsheet",
                params: {
                    context: context,
                    "display-columns": displayColumns,
                    spreadsheetId: spreadsheetId,
                    "where-clause": whereClause
                },
                defaultParameterValue: undefined
            }
        };

        $table.find("thead tr.playsql-col-headers th").map(function(index, th){return $(th).attr("data-name")})

        // Yeah kill the copy-paste!
        o.node = document.createElement("p");
        o.content="<p></p>";
        AJS.log("A Play SQL spreadsheet paste was recognized");

        window.setTimeout(function(){
            // And paste our macro
            tinymce.confluence.MacroUtils.insertMacro(macroRenderRequest);
        }, 0);
    });

    // Swap the handlers around, so our postPaste gets ahead of the others.
    var currentBindings = $doc.data('events')['postPaste'];
    if ($.isArray(currentBindings)) {
        currentBindings.unshift(currentBindings.pop());
    }
});